/*
 * @(#)ICC_Profile.java	1.12 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/**********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.color;

import sun.awt.color.CMM;
import java.io.*;
import java.util.StringTokenizer;

/**
 *  
 * A representation of color profile data for device
 * independent and device dependent color spaces based on the ICC
 * Profile Format Specification, Version 3.4, August 15, 1997, from
 * the International Color Consortium (see <A href="http://www.color.org">
 * http://www.color.org</A>).
 * <p>
 * An ICC_ColorSpace object can be constructed from an appropriate
 * ICC_Profile.
 * Typically, an ICC_ColorSpace would be associated with an ICC
 * Profile which is either an input, display, or output profile (see
 * the ICC specification).  There are also device link, abstract,
 * color space conversion, and named color profiles.  These are less
 * useful for tagging a color or image, but are useful for other
 * purposes (in particular device link profiles can provide improved
 * performance for converting from one device's color space to
 * another's).
 * <p> 
 * ICC Profiles represent transformations from the color space of
 * the profile (e.g. a monitor) to a Profile Connection Space (PCS).
 * Profiles of interest for tagging images or colors have a PCS
 * which is one of the two specific device independent
 * spaces (one CIEXYZ space and one CIELab space) defined in the
 * ICC Profile Format Specification.  Most profiles of interest
 * either have invertible transformations or explicitly specify
 * transformations going both directions.
 * <p>
 * @version 10 Feb 1997
 * @see ICC_ColorSpace
 */


public class ICC_Profile
extends Object {
    long    ID    = 0;
    private static ICC_Profile sRGBprofile = null;
    private static ICC_Profile XYZprofile = null;
    private static ICC_Profile PYCCprofile = null;
    private static ICC_Profile GRAYprofile = null;
    private static ICC_Profile LINEAR_RGBprofile = null;


    /**
     * Profile class is input.
     */
    public static final int CLASS_INPUT = 0;

    /**
     * Profile class is display.
     */
    public static final int CLASS_DISPLAY = 1;

    /**
     * Profile class is output.
     */
    public static final int CLASS_OUTPUT = 2;

    /**
     * Profile class is device link.
     */
    public static final int CLASS_DEVICELINK = 3;

    /**
     * Profile class is color space conversion.
     */
    public static final int CLASS_COLORSPACECONVERSION = 4;

    /**
     * Profile class is abstract.
     */
    public static final int CLASS_ABSTRACT = 5;

    /**
     * Profile class is named color.
<pre>
&nbsp;
&nbsp;
</pre>
     * <B>ICC Color Space signatures.</B>
     */
    public static final int CLASS_NAMEDCOLOR = 6;


    /* Color Space signatures */
    public static final int icSigXYZData        = 0x58595A20;    /* 'XYZ ' */
    public static final int icSigLabData        = 0x4C616220;    /* 'Lab ' */
    public static final int icSigLuvData        = 0x4C757620;    /* 'Luv ' */
    public static final int icSigYCbCrData        = 0x59436272;    /* 'YCbr' */
    public static final int icSigYxyData        = 0x59787920;    /* 'Yxy ' */
    public static final int icSigRgbData        = 0x52474220;    /* 'RGB ' */
    public static final int icSigGrayData        = 0x47524159;    /* 'GRAY' */
    public static final int icSigHsvData        = 0x48535620;    /* 'HSV ' */
    public static final int icSigHlsData        = 0x484C5320;    /* 'HLS ' */
    public static final int icSigCmykData        = 0x434D594B;    /* 'CMYK' */
    public static final int icSigCmyData        = 0x434D5920;    /* 'CMY ' */
    public static final int icSigSpace2CLR        = 0x32434C52;    /* '2CLR' */
    public static final int icSigSpace3CLR        = 0x33434C52;    /* '3CLR' */
    public static final int icSigSpace4CLR        = 0x34434C52;    /* '4CLR' */
    public static final int icSigSpace5CLR        = 0x35434C52;    /* '5CLR' */
    public static final int icSigSpace6CLR        = 0x36434C52;    /* '6CLR' */
    public static final int icSigSpace7CLR        = 0x37434C52;    /* '7CLR' */
    public static final int icSigSpace8CLR        = 0x38434C52;    /* '8CLR' */
    public static final int icSigSpace9CLR        = 0x39434C52;    /* '9CLR' */
    public static final int icSigSpaceACLR        = 0x41434C52;    /* 'ACLR' */
    public static final int icSigSpaceBCLR        = 0x42434C52;    /* 'BCLR' */
    public static final int icSigSpaceCCLR        = 0x43434C52;    /* 'CCLR' */
    public static final int icSigSpaceDCLR        = 0x44434C52;    /* 'DCLR' */
    public static final int icSigSpaceECLR        = 0x45434C52;    /* 'ECLR' */
    /**
       .
<pre>
&nbsp;
&nbsp;
</pre>
     * <B>ICC Profile Class signatures.</B>
     */
    public static final int icSigSpaceFCLR        = 0x46434C52;    /* 'FCLR' */

    /* Profile Class signatures */
    public static final int icSigInputClass            = 0x73636E72;    /* 'scnr' */
    public static final int icSigDisplayClass        = 0x6D6E7472;    /* 'mntr' */
    public static final int icSigOutputClass        = 0x70727472;    /* 'prtr' */
    public static final int icSigLinkClass            = 0x6C696E6B;    /* 'link' */
    public static final int icSigAbstractClass        = 0x61627374;    /* 'abst' */
    public static final int icSigColorSpaceClass    = 0x73706163;    /* 'spac' */
    /**
       .
<pre>
&nbsp;
&nbsp;
</pre>
     * <B>ICC rendering intents.</B>
     */
    public static final int icSigNamedColorClass    = 0x6e6d636c;    /* 'nmcl' */

    /* rendering intents */
    public static final int icPerceptual            = 0;
    public static final int icRelativeColorimetric    = 1;
    public static final int icSaturation            = 2;
    /**
       .
<pre>
&nbsp;
&nbsp;
</pre>
     * <B>ICC Profile Tag signatures.</B>
     */
    public static final int icAbsoluteColorimetric    = 3;

    /* Profile Tag signatures */
    public static final int icSigHead                    = 0x68656164;    /* 'head' - special */
    public static final int icSigAToB0Tag                = 0x41324230;    /* 'A2B0' */ 
    public static final int icSigAToB1Tag                = 0x41324231;    /* 'A2B1' */
    public static final int icSigAToB2Tag                = 0x41324232;    /* 'A2B2' */ 
    public static final int icSigBlueColorantTag        = 0x6258595A;    /* 'bXYZ' */
    public static final int icSigBlueTRCTag                = 0x62545243;    /* 'bTRC' */
    public static final int icSigBToA0Tag                = 0x42324130;    /* 'B2A0' */
    public static final int icSigBToA1Tag                = 0x42324131;    /* 'B2A1' */
    public static final int icSigBToA2Tag                = 0x42324132;    /* 'B2A2' */
    public static final int icSigCalibrationDateTimeTag    = 0x63616C74;    /* 'calt' */
    public static final int icSigCharTargetTag            = 0x74617267;    /* 'targ' */ 
    public static final int icSigCopyrightTag            = 0x63707274;    /* 'cprt' */
    public static final int icSigDeviceMfgDescTag        = 0x646D6E64;    /* 'dmnd' */
    public static final int icSigDeviceModelDescTag        = 0x646D6464;    /* 'dmdd' */
    public static final int icSigGamutTag                = 0x67616D74;    /* 'gamt ' */
    public static final int icSigGrayTRCTag                = 0x6b545243;    /* 'kTRC' */
    public static final int icSigGreenColorantTag        = 0x6758595A;    /* 'gXYZ' */
    public static final int icSigGreenTRCTag            = 0x67545243;    /* 'gTRC' */
    public static final int icSigLuminanceTag            = 0x6C756d69;    /* 'lumi' */
    public static final int icSigMeasurementTag            = 0x6D656173;    /* 'meas' */
    public static final int icSigMediaBlackPointTag        = 0x626B7074;    /* 'bkpt' */
    public static final int icSigMediaWhitePointTag        = 0x77747074;    /* 'wtpt' */
    public static final int icSigPreview0Tag            = 0x70726530;    /* 'pre0' */
    public static final int icSigPreview1Tag            = 0x70726531;    /* 'pre1' */
    public static final int icSigPreview2Tag            = 0x70726532;    /* 'pre2' */
    public static final int icSigProfileDescriptionTag    = 0x64657363;    /* 'desc' */
    public static final int icSigProfileSequenceDescTag    = 0x70736571;    /* 'pseq' */
    public static final int icSigPs2CRD0Tag                = 0x70736430;    /* 'psd0' */
    public static final int icSigPs2CRD1Tag                = 0x70736431;    /* 'psd1' */
    public static final int icSigPs2CRD2Tag                = 0x70736432;    /* 'psd2' */
    public static final int icSigPs2CRD3Tag                = 0x70736433;    /* 'psd3' */
    public static final int icSigPs2CSATag                = 0x70733273;    /* 'ps2s' */
    public static final int icSigPs2RenderingIntentTag    = 0x70733269;    /* 'ps2i' */
    public static final int icSigRedColorantTag            = 0x7258595A;    /* 'rXYZ' */
    public static final int icSigRedTRCTag                = 0x72545243;    /* 'rTRC' */
    public static final int icSigScreeningDescTag        = 0x73637264;    /* 'scrd' */
    public static final int icSigScreeningTag            = 0x7363726E;    /* 'scrn' */
    public static final int icSigTechnologyTag            = 0x74656368;    /* 'tech' */
    public static final int icSigUcrBgTag                = 0x62666420;    /* 'bfd ' */
    public static final int icSigViewingCondDescTag        = 0x76756564;    /* 'vued' */
    public static final int icSigViewingConditionsTag    = 0x76696577;    /* 'view' */
    /**
       .
<pre>
&nbsp;
&nbsp;
</pre>
     * <B>ICC Profile Header locations.</B>
     */
    public static final int icSigNamedColor2Tag            = 0x6E636C32;    /* 'ncl2' */

    /* Header locations */    
    public static final int icHdrSize                = 0;    /* Profile size in bytes */
    public static final int icHdrCmmId                = 4;    /* CMM for this profile */
    public static final int icHdrVersion            = 8;    /* Format version number */
    public static final int icHdrDeviceClass        = 12;    /* Type of profile */
    public static final int icHdrColorSpace            = 16;    /* Color space of data */
    public static final int icHdrPcs                = 20;    /* PCS, XYZ or Lab only */
    public static final int icHdrDate                = 24;    /* Date profile was created */
    public static final int icHdrMagic                = 36;    /* icMagicNumber */
    public static final int icHdrPlatform            = 40;    /* Primary Platform */
    public static final int icHdrFlags                = 44;    /* Various bit settings */
    public static final int icHdrManufacturer        = 48;    /* Device manufacturer */
    public static final int icHdrModel                = 52;    /* Device model number */
    public static final int icHdrAttributes            = 56;    /* Device attributes */
    public static final int icHdrRenderingIntent    = 64;    /* Rendering intent */
    public static final int icHdrIlluminant            = 68;    /* Profile illuminant */
    /**
       .
<pre>
&nbsp;
&nbsp;
</pre>
     * <B>Other ICC constants.</B>
     */
    public static final int icHdrCreator            = 80;    /* Profile illuminant */

    public static final int icTagType            = 0;    /* tag type signature */
    public static final int icTagReserved        = 4;    /* reserved */
    public static final int icCurveCount        = 8;    /* curveType count */
    public static final int icCurveData            = 12;    /* curveType data */
    public static final int icXYZNumberX        = 8;    /* XYZNumber X */


    /**
     * Constructs an empty ICC_Profile object.
     */
    ICC_Profile ()
    {
    }


    /**
     * Constructs an ICC_Profile object with a given ID.
     */
    ICC_Profile(long ID) {
        this.ID = ID;
    }


    /**
     * Frees the resources associated with an ICC_Profile object.
     */
    protected void finalize () {
        CMM.checkStatus (CMM.cmmFreeProfile (ID));
    }


    /**
     * Constructs an ICC_Profile object corresponding to the data in a byte array.
     */
    public static ICC_Profile getInstance (byte[] data) {
    ICC_Profile    thisProfile;
    int            cmmStatus;

        long[] theID = new long [1];
        
        try {
            CMM.checkStatus(CMM.cmmLoadProfile(data, theID));
        } catch (CMMException c) {
            return null;
        }

        try {
            if ((getColorSpaceType (theID[0]) == ColorSpace.TYPE_GRAY) &&
                (getPCSType (theID[0]) == ColorSpace.TYPE_XYZ) &&
                (getData (theID[0], icSigMediaWhitePointTag) != null) &&
                (getData (theID[0], icSigGrayTRCTag) != null)) {
                thisProfile = new ICC_ProfileGray (theID[0]);
            }
            else if ((getColorSpaceType (theID[0]) == ColorSpace.TYPE_RGB) &&
                /* (getPCSType (theID[0]) == ColorSpace.TYPE_XYZ) && */
                (getData (theID[0], icSigMediaWhitePointTag) != null) &&
                (getData (theID[0], icSigRedColorantTag) != null) &&
                (getData (theID[0], icSigGreenColorantTag) != null) &&
                (getData (theID[0], icSigBlueColorantTag) != null) &&
                (getData (theID[0], icSigRedTRCTag) != null) &&
                (getData (theID[0], icSigGreenTRCTag) != null) &&
                (getData (theID[0], icSigBlueTRCTag) != null)) {
                thisProfile = new ICC_ProfileRGB (theID[0]);
            }
            else {
                thisProfile = new ICC_Profile (theID[0]);
            }
        } catch (CMMException c) {
            thisProfile = new ICC_Profile (theID[0]);
        }
        return thisProfile;
    }



    /**
     * Constructs an ICC_Profile corresponding to one of the specific color
     * spaces defined by the ColorSpace class (for example CS_sRGB).
     */
    public static ICC_Profile getInstance (int cspace)
    {
    ICC_Profile thisProfile = null;
    String fileName;

        switch (cspace) {
        case ColorSpace.CS_sRGB:
            if (sRGBprofile == null) {
                sRGBprofile = getInstance ("sRGB.pf");
            }
            thisProfile = sRGBprofile;

            break;
        
        case ColorSpace.CS_CIEXYZ:
            if (XYZprofile == null) {
                XYZprofile = getInstance ("CIEXYZ.pf");
            }
            thisProfile = XYZprofile;

            break;
        
        case ColorSpace.CS_PYCC:
            if (PYCCprofile == null) {
                PYCCprofile = getInstance ("PYCC.pf");
            }
            thisProfile = PYCCprofile;

            break;
        
        case ColorSpace.CS_GRAY:
            if (GRAYprofile == null) {
                GRAYprofile = getInstance ("GRAY.pf");
            }
            thisProfile = GRAYprofile;

            break;
        
        case ColorSpace.CS_LINEAR_RGB:
            if (LINEAR_RGBprofile == null) {
                LINEAR_RGBprofile = getInstance ("LINEAR_RGB.pf");
            }
            thisProfile = LINEAR_RGBprofile;

            break;
        
        default:
            throw new IllegalArgumentException("Unknown color space");
        }
        
        return thisProfile;
    }
        

    /**
     * Constructs an ICC_Profile corresponding to the data in a file.
     * fileName may be an absolute or a relative file specification.
     * Relative file names are looked for in several places: first, relative
     * to any directories specified by the java.iccprofile.path property;
     * second, relative to any directories specified by the java.class.path
     * property; finally, in a directory used to store profiles always
     * available, such as the profile for sRGB.  Built-in profiles use .pf as
     * the file name extension for profiles, e.g. sRGB.pf.
     */
    public static ICC_Profile getInstance(String fileName)
    {
    ICC_Profile    thisProfile = null;
    FileInputStream    fis;
    int        profileSize = 0, bytesRead;
    byte        profileData[];

        if ((fis = openProfile(fileName)) == null) {
            throw new IllegalArgumentException("Cannot open file " + fileName);
        }

        try {
            profileSize = fis.available ();
            profileData = new byte [profileSize];

            try {
                bytesRead = fis.read (profileData);
                try {
                    fis.close();    /* close the file */

                    thisProfile = getInstance (profileData);
                }
                catch (Exception e) {
                }
            }
            catch (Exception e) {
                }
        }
        catch (Exception e) {
        }

        try {
            fis.close();    /* close the file */
        }
        catch (Exception e) {
        }
    
        return thisProfile;
    }


    /**
     * Constructs an ICC_Profile corresponding to the data in an InputStream.
     */
    public static ICC_Profile
        getInstance (    InputStream    s)
    {
    ICC_Profile    thisProfile = null;
    byte        profileData[];
    int        profileSize = 0, bytesRead;

        try {
            profileSize =  s.available ();
        }
        catch (Exception e) {
            }

        profileData = new byte [profileSize];

        try {
            bytesRead = s.read (profileData);

            thisProfile = getInstance (profileData);
        }
        catch (Exception e) {
            }

        return thisProfile;
    }


    /**
     * Returns profile major version.
     */
    public int getMajorVersion() {
    byte[]    theHeader;
    
        theHeader = getData (icSigHead);

        return (int) theHeader[8];
    }

    /**
     * Returns profile minor version.
     */
    public int getMinorVersion() {
    byte[]    theHeader;
    
        theHeader = getData (icSigHead);

        return (int) theHeader[9];
    }

    /**
     * Returns the profile class.  Returns one of the predefined profile
     * class constants.
     */
    public int getProfileClass() {
    byte[]    theHeader;
    int    theClassSig, theClass;
    
        theHeader = getData (icSigHead);

        theClassSig = intFromBigEndian (theHeader, icHdrDeviceClass);
                
        switch (theClassSig) {
        case icSigInputClass:
            theClass = CLASS_INPUT;
            break;

        case icSigDisplayClass:
            theClass = CLASS_DISPLAY;
            break;

        case icSigOutputClass:
            theClass = CLASS_OUTPUT;
            break;

        case icSigLinkClass:
            theClass = CLASS_DEVICELINK;
            break;

        case icSigColorSpaceClass:
            theClass = CLASS_COLORSPACECONVERSION;
            break;

        case icSigAbstractClass:
            theClass = CLASS_ABSTRACT;
            break;

        case icSigNamedColorClass:
            theClass = CLASS_NAMEDCOLOR;
            break;

        default:
            throw new IllegalArgumentException("Unknown profile class");
        }
        
        return theClass;
    }

    /**
     * Returns the color space type.  Returns one of the color space type
     * constants defined by the ColorSpace class.  This is the
     * "input" color space of the profile.  The type defines the
     * number of components of the color space and the interpretation,
     * e.g. TYPE_RGB identifies a color space with three components - red,
     * green, and blue.  It does not define the particular color
     * characteristics of the space, e.g. the chromaticities of the
     * primaries.
     */
    public int getColorSpaceType()
    {
        return    getColorSpaceType(ID);
    }
    
    static int getColorSpaceType (long    profileID)
    {
    byte[]    theHeader;
    int    theColorSpaceSig, theColorSpace;
    
        theHeader = getData (profileID, icSigHead);

        theColorSpaceSig = intFromBigEndian (theHeader, icHdrColorSpace);
                
        theColorSpace = iccCStoJCS (theColorSpaceSig);
                
        return theColorSpace;
    }

    /**
     * Returns the color space type of the Profile Connection Space (PCS).
     * Returns one of the color space type constants defined by the
     * ColorSpace class.  This is the "output" color space of the
     * profile.  For an input, display, or output profile useful
     * for tagging colors or images, this will be either TYPE_XYZ or
     * TYPE_Lab and should be interpreted as the corresponding specific
     * color space defined in the ICC specification.  For a device
     * link profile, this could be any of the color space type constants.
     */
    public int getPCSType()
    {
        return    getPCSType (ID);
    }
    
    
    static int getPCSType (long    profileID)
    {
    byte[]    theHeader;
    int    theColorSpaceSig, theColorSpace;
    
        theHeader = getData (profileID, icSigHead);

        theColorSpaceSig = intFromBigEndian (theHeader, icHdrPcs);

        theColorSpace = iccCStoJCS (theColorSpaceSig);
                
        return theColorSpace;
    }


    /**
     * Store this ICC_Profile in a File.
     */
    public void
        store (    String    fileName)
    {
    FileOutputStream    outputFile;
    int        profileSize = 0, bytesRead;
    byte        profileData[];

        try {
            outputFile = new FileOutputStream (fileName);

            profileData = getData ();
        
            try {
                outputFile.write (profileData);
                try {
                    outputFile.close ();
                }
                catch (Exception e) {
                    }
            }
            catch (Exception e) {
                }
        }
        catch (Exception e) {
            }
    }


    /**
     * Returns a byte array corresponding to the data of this ICC_Profile.
     */
    public byte[] getData () {
    int[]    profileSize = new int [1];
    byte[]    profileData;

        /* get the number of bytes needed for this profile */
        CMM.checkStatus (CMM.cmmGetProfileSize (ID, profileSize));

        profileData = new byte [profileSize[0]];

        /* get the data for the profile */
        CMM.checkStatus (CMM.cmmGetProfileData (ID, profileData));

        return profileData;
    }


    /**
     * Returns a particular tagged data element from the profile as
     * a byte array.  Elements are identified by signatures
     * as defined in the ICC specification.  The signature
     * icSigHead can be used to get the header.  This method is useful
     * for advanced applets or applications which need to access
     * profile data directly.  Returns null if no such tag exists.
     */
    public byte[] getData(int tagSignature) {
        return getData (ID, tagSignature);
    }
    

    static byte[] getData(long profileID, int tagSignature) {
    int[]    tagSize = new int [1];
    byte[]    tagData;
    int    cmmStatus;

	try {
            /* get the number of bytes needed for this tag */
            CMM.checkStatus(CMM.cmmGetTagSize(profileID, tagSignature,
                                              tagSize));

            tagData = new byte [tagSize[0]];    /* get an array for the tag */

            /* get the tag's data */
            CMM.checkStatus(CMM.cmmGetTagData(profileID, tagSignature,
                                              tagData));
	} catch(CMMException c) {
	    tagData = null;
	}

        return tagData;
    }

    /**
     * Sets a particular tagged data element in the profile from
     * a byte array.  This method is useful
     * for advanced applets or applications which need to access
     * profile data directly.
     */
    public void
        setData (    int        tagSignature,
                    byte[]    tagData)
    {
    int    cmmStatus;

        CMM.checkStatus (CMM.cmmSetTagData (ID, tagSignature, tagData));
    }

    /**
     * Sets the rendering intent of the profile.
     * This is used to select the proper transform from a profile that
     * has multiple transforms.
     */
    void
        setRenderingIntent (    int    renderingIntent)
    {
        byte[] theHeader = getData (icSigHead);
        
        intToBigEndian (renderingIntent, theHeader, icHdrRenderingIntent);    /* set the rendering intent */
        
        setData (icSigHead, theHeader);
    }


    /**
     * Returns the rendering intent of the profile.
     * This is used to select the proper transform from a profile that
     * has multiple transforms.  It is typically set in a source profile to select
     * a transform from an output profile.
     */
    int
        getRenderingIntent ()
    {
        byte[] theHeader = getData (icSigHead);
        
        int    renderingIntent = intFromBigEndian (theHeader, icHdrRenderingIntent);    /* set the rendering intent */
        
        return renderingIntent;
    }


    /**
     * Returns the number of color components in the "input" color
     * space of this profile.  For example if the color space type
     * of this profile is TYPE_RGB, then this method will return 3.
     */
    public int getNumComponents()
    {
    byte[]    theHeader;
    int    theColorSpaceSig, theNumComponents;
    
        theHeader = getData (icSigHead);

        theColorSpaceSig = intFromBigEndian (theHeader, icHdrColorSpace);

        switch (theColorSpaceSig) {
        case icSigGrayData:
            theNumComponents = 1;
            break;

        case icSigSpace2CLR:
            theNumComponents = 2;
            break;

        case icSigXYZData:
        case icSigLabData:
        case icSigLuvData:
        case icSigYCbCrData:
        case icSigYxyData:
        case icSigRgbData:
        case icSigHsvData:
        case icSigHlsData:
        case icSigCmyData:
        case icSigSpace3CLR:
            theNumComponents = 3;
            break;

        case icSigCmykData:
        case icSigSpace4CLR:
            theNumComponents = 4;
            break;

        case icSigSpace5CLR:
            theNumComponents = 5;
            break;

        case icSigSpace6CLR:
            theNumComponents = 6;
            break;

        case icSigSpace7CLR:
            theNumComponents = 7;
            break;

        case icSigSpace8CLR:
            theNumComponents = 8;
            break;

        case icSigSpace9CLR:
            theNumComponents = 9;
            break;

        case icSigSpaceACLR:
            theNumComponents = 10;
            break;

        case icSigSpaceBCLR:
            theNumComponents = 11;
            break;

        case icSigSpaceCCLR:
            theNumComponents = 12;
            break;

        case icSigSpaceDCLR:
            theNumComponents = 13;
            break;

        case icSigSpaceECLR:
            theNumComponents = 14;
            break;

        case icSigSpaceFCLR:
            theNumComponents = 15;
            break;

        default:
            throw new ProfileDataException ("invalid ICC color space");
        }
        
        return theNumComponents;
    }


    /**
     * Returns a float array of length 3 containing the X, Y, and Z
     * components of the mediaWhitePointTag in the ICC profile.
     */
    float[] getMediaWhitePoint()
    {
        return getXYZTag (icSigMediaWhitePointTag);    /* get the media white point tag */
    }


    /**
     * Returns a float array of length 3 containing the X, Y, and Z
     * components encoded in an XYZType tag.
     */
    float[]
        getXYZTag (    int    theTagSignature)
    {
    byte[]    theData;
    float[]    theXYZNumber;
    int    i1, i2, theS15Fixed16;
    
        theData = getData (theTagSignature);    /* get the tag data */
        
        theXYZNumber = new float [3];        /* array to return */

        /* convert s15Fixed16Number to float */
        for (i1 = 0, i2 = icXYZNumberX; i1 < 3; i1++, i2 += 4) {
            theS15Fixed16 = intFromBigEndian (theData, i2);
            
            theXYZNumber [i1] = (float) (theS15Fixed16 / 65536.0);
        }
        
        return theXYZNumber;
    }


    /**
     * Returns a gamma value representing a tone reproduction
     * curve (TRC).  If the profile represents the TRC as a table rather
     * than a single gamma value, then an exception is thrown.  In this
     * case the actual table can be obtained via getTRC().  When
     * using a gamma value, the PCS Y component is computed as follows:
<pre>
&nbsp;                         gamma
&nbsp;        PCSY = deviceGray
</pre>
     * @return the gamma value as a float.
     * @exception ProfileDataException if the profile does not specify
     *            the TRC as a single gamma value.
     */
    float getGamma (    int    theTagSignature)
    {
    byte[]    theTRCData;
    float    theGamma;
    int    theU8Fixed8;

            /* get the media white point tag */
        theTRCData = getData (theTagSignature);
        
        if (intFromBigEndian (theTRCData, icCurveCount) != 1) {
            throw new ProfileDataException ("TRC is not a gamma");
        }

            /* convert u8Fixed8 to float */
        theU8Fixed8 = shortFromBigEndian (theTRCData, icCurveData);
            
        theGamma = (float) (theU8Fixed8 / 256.0);
        
        return theGamma;
    }


    /**
     * Returns the TRC as an array of shorts.  If the profile has
     * specified the TRC as linear (gamma = 1.0) or as a simple gamma
     * value, this method throws an exception, and the getGamma() method
     * should be used to get the gamma value.  Otherwise the short array
     * returned here represents a lookup table where the input Gray value
     * is conceptually in the range [0.0, 1.0].  Value 0.0 maps
     * to array index 0 and value 1.0 maps to array index length-1.
     * Interpolation may be used to generate output values for
     * input values which do not map exactly to an index in the
     * array.  Output values also map linearly to the range [0.0, 1.0].
     * Value 0.0 is represented by an array value of 0x0000 and
     * value 1.0 by 0xFFFF, i.e. the values are really unsigned
     * short values, although they are returned in a short array.
     * @return a short array representing the TRC.
     * @exception ProfileDataException if the profile does not specify
     *            the TRC as a table.
     */
    short[] getTRC (    int    theTagSignature)
    {
    byte[]    theTRCData;
    short[]    theTRC;
    int    i1, i2, nElements, theU8Fixed8;

            /* get the media white point tag */
        theTRCData = getData (theTagSignature);
        
        nElements = intFromBigEndian (theTRCData, icCurveCount);

        if (nElements == 1) {
            throw new ProfileDataException ("TRC is not a table");
        }

            /* make the short array */
        theTRC = new short [nElements];

        for (i1 = 0, i2 = icCurveData; i1 < nElements; i1++, i2 += 2) {
            theTRC[i1] = shortFromBigEndian (theTRCData, i2);
        }
        
        return theTRC;
    }


    /* convert an ICC color space signature into a Java color space type */
    static int
        iccCStoJCS (    int    theColorSpaceSig)
    {
    int    theColorSpace;

        switch (theColorSpaceSig) {
        case icSigXYZData:
            theColorSpace = ColorSpace.TYPE_XYZ;
            break;

        case icSigLabData:
            theColorSpace = ColorSpace.TYPE_Lab;
            break;

        case icSigLuvData:
            theColorSpace = ColorSpace.TYPE_Luv;
            break;

        case icSigYCbCrData:
            theColorSpace = ColorSpace.TYPE_YCbCr;
            break;

        case icSigYxyData:
            theColorSpace = ColorSpace.TYPE_Yxy;
            break;

        case icSigRgbData:
            theColorSpace = ColorSpace.TYPE_RGB;
            break;

        case icSigGrayData:
            theColorSpace = ColorSpace.TYPE_GRAY;
            break;

        case icSigHsvData:
            theColorSpace = ColorSpace.TYPE_HSV;
            break;

        case icSigHlsData:
            theColorSpace = ColorSpace.TYPE_HLS;
            break;

        case icSigCmykData:
            theColorSpace = ColorSpace.TYPE_CMYK;
            break;

        case icSigCmyData:
            theColorSpace = ColorSpace.TYPE_CMY;
            break;

        case icSigSpace2CLR:
            theColorSpace = ColorSpace.TYPE_2CLR;
            break;

        case icSigSpace3CLR:
            theColorSpace = ColorSpace.TYPE_3CLR;
            break;

        case icSigSpace4CLR:
            theColorSpace = ColorSpace.TYPE_4CLR;
            break;

        case icSigSpace5CLR:
            theColorSpace = ColorSpace.TYPE_5CLR;
            break;

        case icSigSpace6CLR:
            theColorSpace = ColorSpace.TYPE_6CLR;
            break;

        case icSigSpace7CLR:
            theColorSpace = ColorSpace.TYPE_7CLR;
            break;

        case icSigSpace8CLR:
            theColorSpace = ColorSpace.TYPE_8CLR;
            break;

        case icSigSpace9CLR:
            theColorSpace = ColorSpace.TYPE_9CLR;
            break;

        case icSigSpaceACLR:
            theColorSpace = ColorSpace.TYPE_ACLR;
            break;

        case icSigSpaceBCLR:
            theColorSpace = ColorSpace.TYPE_BCLR;
            break;

        case icSigSpaceCCLR:
            theColorSpace = ColorSpace.TYPE_CCLR;
            break;

        case icSigSpaceDCLR:
            theColorSpace = ColorSpace.TYPE_DCLR;
            break;

        case icSigSpaceECLR:
            theColorSpace = ColorSpace.TYPE_ECLR;
            break;

        case icSigSpaceFCLR:
            theColorSpace = ColorSpace.TYPE_FCLR;
            break;

        default:
            throw new IllegalArgumentException ("Unknown color space");
        }
        
        return theColorSpace;
    }
    
    
    static int
        intFromBigEndian (    byte[]    array,
                            int        index)
    {
        return (((array[index+0] & 0xff) << 24) |
                ((array[index+1] & 0xff) << 16) |
                ((array[index+2] & 0xff) <<  8) |
                 (array[index+3] & 0xff));
    }
    
    
    static void
        intToBigEndian (    int        value,
                            byte[]    array,
                            int        index)
    {
            array [index+0] = (byte) (value >> 24);
            array [index+1] = (byte) (value >> 16);
            array [index+2] = (byte) (value >>  8);
            array [index+3] = (byte) (value >>  0);
    }
    
    
    static short
        shortFromBigEndian (    byte[]    array,
                                int        index)
    {
        return (short) (((array[index+0] & 0xff) << 8) |
                         (array[index+1] & 0xff));
    }
    
    
    static void
        shortToBigEndian (    short    value,
                            byte[]    array,
                            int    index)
    {
            array [index+0] = (byte) (value >> 8);
            array [index+1] = (byte) (value >> 0);
    }


    /*
     * fileName may be an absolute or a relative file specification.
     * Relative file names are looked for in several places: first, relative
     * to any directories specified by the java.iccprofile.path property;
     * second, relative to any directories specified by the java.class.path
     * property; finally, in a directory used to store profiles always
     * available, such as a profile for sRGB.  Built-in profiles use .pf as
     * the file name extension for profiles, e.g. sRGB.pf.
     */
    private static FileInputStream openProfile(String fileName) {
        FileInputStream fis = null;
        String path, dir, fullPath;

        java.security.AccessController.beginPrivileged();

        try {
            fis = new FileInputStream(fileName);  /* absolute file name */
        }
        catch (FileNotFoundException e) {
        }

        if ((fis == null) &&
            ((path = System.getProperty("java.iccprofile.path")) != null)) {
                                    /* try relative to java.iccprofile.path */
            StringTokenizer st = new StringTokenizer(path, File.pathSeparator);
            while (st.hasMoreTokens() && (fis == null)) {
                dir = st.nextToken();
                try {
                    fullPath = dir + File.separatorChar + fileName;
                    fis = new FileInputStream(fullPath);
                }
                catch (FileNotFoundException e) {
                }
            }
        }

        if ((fis == null) &&
            ((path = System.getProperty("java.class.path")) != null)) {
                                    /* try relative to java.class.path */
            StringTokenizer st = new StringTokenizer(path, File.pathSeparator);
            while (st.hasMoreTokens() && (fis == null)) {
                dir = st.nextToken();
                try {
                    fullPath = dir + File.separatorChar + fileName;
                    fis = new FileInputStream(fullPath);
                }
                catch (FileNotFoundException e) {
                }
            }
        }

        if (fis == null) {        /* try the directory of built-in profiles */
            dir = System.getProperty("java.home") + File.separatorChar + "lib" +
                                     File.separatorChar + "cmm";
            fullPath = dir + File.separatorChar + fileName;
            try {
                fis = new FileInputStream(fullPath);
            }
            catch (FileNotFoundException e) {
            }
        }

        java.security.AccessController.endPrivileged();
        return fis;
    }

}
