/*
 * @(#)ICC_ProfileGray.java	1.10 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/**********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.color;

import java.awt.image.LookupTable;

/**
 *  
 * A subclass of the ICC_Profile class which represents profiles
 * which meet the following criteria: the color space type of the
 * profile is TYPE_GRAY, the PCS is the CIEXYZ space defined by the ICC
 * specification, and the profile includes the grayTRCTag and
 * mediaWhitePointTag tags.  Examples of this kind of profile are
 * monochrome input profiles, monochrome display profiles, and
 * monochrome output profiles.  The
 * <A href="java.awt.color.ICC_Profile.html#getInstance">
 * ICC_Profile.getInstance()
 * </A>
 * method will
 * return an ICC_ProfileGray object when the above conditions are
 * met.  The advantage of this class is that it provides a lookup
 * table that Java or native methods may be able to use directly to
 * optimize color conversion in some cases.
 * <p>
 * To transform from a GRAY device profile color space to the CIEXYZ Profile
 * Connection Space, the device gray component is transformed by
 * a lookup through the tone reproduction curve (TRC).  The result is
 * treated as the achromatic component of the PCS.
<pre>

&nbsp;               PCSY = grayTRC[deviceGray]

</pre>
 * The inverse transform is done by converting the PCS Y components to
 * device Gray via the inverse of the grayTRC.
 * <p>
 * @version 10 Feb 1997
 */



public class ICC_ProfileGray
extends ICC_Profile {

    /**
     * Constructs an empty ICC_ProfileGray.  Before using the profile
     * for any purpose all necessary profile data must be set to
     * appropriate values.
     */
    ICC_ProfileGray (long ID)
    {
        super (ID);
    }

    /**
     * Returns a new ICC_ProfileGray from a byte data array.
     */
    ICC_ProfileGray (byte[] data)
    {
        ICC_Profile theProfile = ICC_Profile.getInstance (data);
        
        if ( ! (theProfile instanceof ICC_ProfileGray)) {
            throw new ProfileDataException("Not a gray scale profile");
        }
    }


    /**
     * Returns a float array of length 3 containing the X, Y, and Z
     * components of the mediaWhitePointTag in the ICC profile.
     */
    public float[] getMediaWhitePoint()
    {
        return super.getMediaWhitePoint();
    }


    /**
     * Returns a gamma value representing the tone reproduction
     * curve (TRC).  If the profile represents the TRC as a table rather
     * than a single gamma value, then an exception is thrown.  In this
     * case the actual table can be obtained via getTRC().  When
     * using a gamma value, the PCS Y component is computed as follows:
<pre>

&nbsp;                         gamma
&nbsp;        PCSY = deviceGray

</pre>
     * @return the gamma value as a float.
     * @exception ProfileDataException if the profile does not specify
     *            the TRC as a single gamma value.
     */
    public float getGamma()
    {
    float    theGamma;
    
        theGamma = super.getGamma (ICC_Profile.icSigGrayTRCTag);
        
        return theGamma;
    }

    /**
     * Returns the TRC as an array of shorts.  If the profile has
     * specified the TRC as linear (gamma = 1.0) or as a simple gamma
     * value, this method throws an exception, and the getGamma() method
     * should be used to get the gamma value.  Otherwise the short array
     * returned here represents a lookup table where the input Gray value
     * is conceptually in the range [0.0, 1.0].  Value 0.0 maps
     * to array index 0 and value 1.0 maps to array index length-1.
     * Interpolation may be used to generate output values for
     * input values which do not map exactly to an index in the
     * array.  Output values also map linearly to the range [0.0, 1.0].
     * Value 0.0 is represented by an array value of 0x0000 and
     * value 1.0 by 0xFFFF, i.e. the values are really unsigned
     * short values, although they are returned in a short array.
     * @return a short array representing the TRC.
     * @exception ProfileDataException if the profile does not specify
     *            the TRC as a table.
     */
    public short[] getTRC()
    {
    short[]    theTRC;

        theTRC = super.getTRC (ICC_Profile.icSigGrayTRCTag);
        
        return theTRC;
    }

    /*
     * Returns a LookupTable object which represents the TRC.
     * If shortdata is true, a ShortLookupTable is returned.
     * Otherwise, a ByteLookupTable is returned.
     * The number of entries and offset for the
     * LookupTable are provided as arguments to this method.  The
     * first entry in the LookupTable will correspond to a TRC input
     * value of 0.0, and the last entry to 1.0.  The byte or short
     * entries in the LookupTable will be computed from the TRC of the
     * profile and are unsigned values ranging from 0x00 to 0xFF for
     * byte data and from 0x0000 to 0xFFFF for short data.
     * This method returns a valid LookupTable both for TRCs
     * specified as tables and those specified as gamma values in the
     * profile.
     * @return a LookupTable representing the TRC.
     */
    /* REMIND - do we need this?  */
    /*
    public LookupTable getTRCLUT(boolean shortdata,
                 int num_entries, int offset) {
    }
     */

}

