/*
 * @(#)DataFlavor.java	1.15 98/03/18
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.datatransfer;

import java.io.*;

import java.util.mime.*;

/**
 * Each instance represents the opaque concept of a data format as would
 * appear on a clipboard, during drag and drop, or in a file system.
 *
 * @version 	1.15, 03/18/98
 * @author	Blake Sullivan
 */

public class DataFlavor implements Serializable, Cloneable {

    static final long serialVersionUID = 8367026044764648243L;

    static final Class ioInputStreamClass = java.io.InputStream.class;

    /*
     * private initializer
     */

    static private DataFlavor createConstant(Class rc, String prn) {
	try {
	    return new DataFlavor(rc, prn);
	} catch (Exception e) {
	    return null;
	}
    }

    /*
     * private initializer
     */

    static private DataFlavor createConstant(String mt, String prn) {
	try {
	    return new DataFlavor(mt, prn);
	} catch (Exception e) {
	    return null;
	}
    }

   /**
    * The DataFlavor representing a Java Unicode String class, where:
    * <p>
    * representationClass = java.lang.String<br>
    * mimeType            = "application/x-java-serialized-object"        
    * <p> 
    */

    public static final DataFlavor stringFlavor = createConstant(java.lang.String.class, "Unicode String");

   /**
    * The DataFlavor representing plain text with unicode encoding, where:
    * <p>
    * representationClass = InputStream<br>
    * mimeType            = "text/plain; charset=unicode"        
    * <p> 
    */

    public static final DataFlavor plainTextFlavor = createConstant("text/plain; charset=unicode", "Plain Text"); 
		

    static final String serializedObjectMimeType = "application/x-java-serialized-object";


    private static MimeType javaSerializedObjectMimeType;

   /**
    * Mime type for this DataFlavor.  (See RFC 1521 for an explanation
    * of Mime types)
    */
    
    /**
     * Construct a fully specified DataFlavor
     */

   public DataFlavor(String primaryType, String subType, MimeTypeParameterList params, Class representationClass, String humanPresentableName) {
	super();

	if (params == null) params = new MimeTypeParameterList();

	params.set("class", representationClass.getName());
	
	if (humanPresentableName == null) 
	    humanPresentableName = primaryType + "/" + subType;

	params.set("humanPresentableName", humanPresentableName);

	try {
	    mimeType = new MimeType(primaryType, subType, params);
	} catch (MimeTypeParseException mtpe) {
	    throw new IllegalArgumentException("MimeType Parse Exception: " + mtpe.getMessage());
	}

	this.representationClass  = representationClass;
	this.humanPresentableName = humanPresentableName;
   }

   /**
    * Construct a DataFlavor that represents a Java class
    * <p>
    * The returned DataFlavor will have the following characteristics
    * <p>
    * representationClass = representationClass<br>
    * mimeType            = application/x-java-serialized-object        
    * <p>
    * @param representationClass the class used to transfer data in this flavor
    * @param humanPresentableName the human-readible string used to identify this flavor
    */
    public DataFlavor(Class representationClass, String humanPresentableName) {
        this("application", "x-java-serialized-object", null, representationClass, humanPresentableName);
    }

   /**
    * Construct a DataFlavor that represents a MimeType
    * <p>
    * The returned DataFlavor will have the following characteristics:
    * <p>
    * If the mimeType is
    * "application/x-javaserializedobject; class=&lt;representation class&gt;",
    * the result is the same as calling
    * new DataFlavor(Class:forName(&lt;representation class&gt;) as above
    * <p>
    * otherwise:
    * <p>
    * representationClass = InputStream<br>
    * mimeType            = mimeType         
    * <p>
    * @param mimeType the string used to identify the MIME type for this flavor
    * @param humanPresentableName the human-readible string used to identify this flavor
    */
    public DataFlavor(String mimeType, String humanPresentableName) {
	super();

	try {
            this.mimeType = new MimeType(mimeType);
	} catch (MimeTypeParseException mtpe) {
	    throw new IllegalArgumentException(mimeType);
	}

	String rcn = getParameter("class");
	Class  rc  = null;

	if (rcn != null) try {
	    rc = Class.forName(rcn);
	} catch (ClassNotFoundException cnfe) {
	    throw new IllegalArgumentException(rcn);
	}

	rc = this.representationClass = (rc == null ? ioInputStreamClass : rc);
	this.mimeType.setParameter("class", rc.getName());

      	this.humanPresentableName = (humanPresentableName == null ? this.mimeType.getPrimaryType() + "/" + this.mimeType.getSubType() : humanPresentableName);
	this.mimeType.setParameter("humanPresentableName", this.humanPresentableName);
    }       
  
   /**
    * construct a DataFlavor from a Mime Type string.
    * 
    * The string must specify a "class=<fully specified Java class name>"
    * parameter in order to succeed in constructing a DataFlavor.
    */

   public DataFlavor(String mimeType) throws MimeTypeParseException, ClassNotFoundException {
	super();

	this.mimeType = new MimeType(mimeType);

	String rc = (String)this.mimeType.getParameter("class");

	if (rc == null)
	    throw new IllegalArgumentException("no representation class specified");

	representationClass = Class.forName(rc);

	humanPresentableName = this.mimeType.getParameter("humanPresentableName");
	if (humanPresentableName == null)
	    setHumanPresentableName(this.mimeType.getPrimaryType() + "/" + this.mimeType.getSubType());
   }

   /**
    * used by clone implementation
    */

    private DataFlavor(MimeType mt, Class rc, String hrn, int a) {
	super();

	mimeType             = mt;
	representationClass  = rc;
	humanPresentableName = hrn;
	atom		     = a;
    }

   /**
    * Returns the MIME type string for this DataFlavor
    */ 
    public String getMimeType() {
    	return mimeType.toString();
    }

   /**
    *  Returns the Class which objects supporting this DataFlavor
    *  will return when this DataFlavor is requested.
    */
    public Class getRepresentationClass() {
    	return representationClass;
    }
  
   /**
    * Returns the human presentable name for the data foramt that this
    * DataFlavor represents.  This name would be localized for different
    * countries
    */
    public String getHumanPresentableName() {
    	return humanPresentableName;
    }
  
   /**
    * @return the primary MIME type of this DataFlavor
    */

    public String getPrimaryType() { return mimeType.getPrimaryType(); }

   /**
    * @return the Sub MIME type of this DataFlavor
    */

    public String getSubType() { return mimeType.getSubType(); }

   /**
    * @return the value of the name parameter
    */

    public String getParameter(String paramName) {
	return mimeType.getParameter(paramName);
    }

   /**
    * Sets the human presentable name for the data format that this
    * DataFlavor represents. This name would be localized for different
    * countries
    */
    public void setHumanPresentableName(String humanPresentableName) {
    	this.humanPresentableName = humanPresentableName;
	mimeType.setParameter("humanPresentableName", humanPresentableName);
    }

    /**
     * @return if the objects are equal
     */

    public boolean equals(Object o) {
	return ((o instanceof DataFlavor) && equals((DataFlavor)o)) ||
	       ((o instanceof MimeType)   && equals((MimeType)o))   ||
	       ((o instanceof String)     && equals((String)o));
	      
    }

    /**
     * @return if the objects are equal
     */

    public boolean equals(MimeType mt) {
	return mimeType.equals(mt);
    }

    /**
     * @return if the DataFlavors represent the same type.
     */

    public boolean equals(DataFlavor dataFlavor) {
	return mimeType.match(dataFlavor.mimeType) && representationClass.equals(dataFlavor.getRepresentationClass());
    }
  
    /**
     * @return if the String (MimeType) is equal
     */

    public boolean equals(String s) {
	return isMimeTypeEqual(s);
    }

    /**
     * Is the string representation of the MIME type passed in equivalent
     * to the MIME type of this DataFlavor.  This may involve adding default
     * attributes for some MIME types (like adding charset=US-ASCII to
     * text/plain MIME types that have no charset parameter specified)
     */

    public boolean isMimeTypeEqual(String mimeType) {
	try {
	    return this.mimeType.match(new MimeType(mimeType));
	} catch (MimeTypeParseException mtpe) {
	    return false; 
	}
    }
  
    /**
     * @return if the MimeTypes are equal
     */

    public final boolean isMimeTypeEqual(DataFlavor dataFlavor) {
	return isMimeTypeEqual(dataFlavor.mimeType);
    }
    
    /**
     * @return if the MimeTypes are equal
     */

    public boolean isMimeTypeEqual(MimeType mtype) {
	return mimeType.match(mtype);
    }

   /**
    * does the DataFlavor represent a serialized object?
    */

    public boolean isMimeTypeSerializedObject() {
	if (javaSerializedObjectMimeType == null) try {
	    javaSerializedObjectMimeType = new MimeType(serializedObjectMimeType);
	} catch (MimeTypeParseException mtpe) { return false; }

	return mimeType.equals(javaSerializedObjectMimeType);
    }

   /**
    * does the DataFlavor represent a serialized object?
    */

    public boolean isRepresentationClassInputStream() {
	return ioInputStreamClass.isAssignableFrom(representationClass);
    }

   /**
    * @return true if the representation class can be serialized
    */

    public boolean isRepresentationClassSerializable() {
	return java.io.Serializable.class.isAssignableFrom(representationClass);
    }

   /**
    * Serialize this DataFlavor
    */

   public void writeObject(ObjectOutputStream oos) throws IOException {
	oos.defaultWriteObject();
   }

   /**
    * restore this DataFlavor from an Serialized state
    */

   public void readObject(ObjectInputStream ois) throws IOException , ClassNotFoundException {
	ois.defaultReadObject();
   }

   public DataFlavor() { super(); } // this is here for serialization only.

   /**
    * @return a clone of this DataFlavor
    */

   public Object clone() throws CloneNotSupportedException {
	return new DataFlavor((MimeType)mimeType.clone(), representationClass, humanPresentableName, atom);
   }

   /**
    * Called on DataFlavor for every MIME Type parameter to allow DataFlavor
    * subclasses to handle special parameters like the text/plain charset
    * parameters, whose values are case insensitive.  (MIME type parameter
    * values are supposed to be case sensitive.
    * <p>
    * This method is called for each parameter name/value pair and should
    * return the normalized representation of the parameterValue
    *
    * This method is never invoked by this implementation from 1.1 onwards
    *
    * @depreciated
    */
    protected String normalizeMimeTypeParameter(String parameterName, String parameterValue) {
	return parameterName+"="+parameterValue;	
    }
  
   /**
    * Called for each MIME type string to give DataFlavor subtypes the
    * opportunity to change how the normalization of MIME types is accomplished.
    * One possible use would be to add default parameter/value pairs in cases
    * where none are present in the MIME type string passed in
    *
    * This method is never invoked by this implementation from 1.1 onwards
    *
    * @depreciated
    */
    protected String normalizeMimeType(String mimeType) {
	return mimeType;	
    }

    /*
     * fields
     */

    /* placeholder for caching any platform-specific data for flavor */

    int			atom;      
  
    /* Mime Type of DataFlavor */

    MimeType		mimeType;

    /** Human-presentable name for this DataFlavor. Localizable. **/

    private String	humanPresentableName;  
  
    /** Java class of objects this DataFlavor represents **/

    private Class	representationClass;
}
