/*
 * @(#)GlyphIterator.java	1.13 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package java.awt.font;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.Font;
import java.lang.Math;


/*
    GlyphIterator

    This class takes a set of glyphsets, and their visual ordering, and presents
    the glyphs to the client as a single sequence of glyphs having a logical and
    visual ordering.  In particular, this class treats a base glyph with component
    glyphs as a single 'glyph', which is convenient for clients which want to ignore
    component glyphs.

    Because the GlyphIterator groups component glyphs together, there may actually
    be more glyphs in the glyphsets managed by the iterator than the client sees.  The
    term 'index' is used to represent the client's point of view (that is, treating
    base glyphs and their components as one 'glyph').  The term 'number' is used inside
    the iterator to index into the glyphsets themselves (to which component glyphs are
    first-class glyphs, no different from any others).
*/

final class GlyphIterator {

    private TextLayout fLayout;
    private TextLayoutComponent[] fSets;
    private int[] fVisualSetsOrder;
    private boolean fIsVertical;
    private float[] fBaselines;

    /*
     * Visual position of current glyph set.  The current glyph set is:
     * fSets[fVisualSetsOrder[fVisualCurrentSet]]
     */
    private int fVisualCurrentSet;

    /*
     * Logical glyph index (not actual glyph number) of the first glyph in the
     * current set.
     */
    private int fCurrentSetLogicalStart;

    /*
     * Visual glyph index (not actual glyph number) of the first glyph in the
     * current set.
     */
    private int fCurrentSetVisualStart;

    /*
     * Advance to start of current set (from left).
     */
    private float fCurrentSetAdvance;

    /*
     * Glyph numbers (not indices) of the current base glyph and its last component
     * glyph within current set.  Numbers are in visual order;  so, for example, the
     * base glyph is given by:
     *
     * GlyphSet currentSet = fSets[fVisualSetsOrder[fVisualCurrentSet]];
     * int[] visualOrder = currentSet.visualOrder();
     * int baseGlyphCode = currentSet.getGlyphCodes()[visualOrder[fCurrentBaseGlyph]];
     */
    private int fCurrentBaseGlyph, fLastComponentGlyph;

    /*
     * Number of glyphs in iterator, from client's perspective.
     */
    private int fGlyphIndexLimit;

    /*
     * Advance of all sets in iterator.
     */
    private float fTotalAdvance;

    /*
     * When iterator is invalid, fVisualCurrentSet is set to this value.
     */
    private static final int INVALID_SET = -42;

    /*
     * Create a GlyphIterator set to visual position 0.
     */
    public GlyphIterator(TextLayout layout, TextLayoutComponent[] sets, int[] setsOrder) {

        fLayout = layout;
        fBaselines = layout.getBaselineOffsets();
        fSets = sets;
        fVisualSetsOrder = setsOrder;
        fIsVertical = false;

        fGlyphIndexLimit = 0;
        fTotalAdvance = 0;

        for (int i=0; i<fSets.length; i++) {
            fGlyphIndexLimit += fSets[i].getNumCharacters();
            fTotalAdvance += fSets[i].getAdvance();
        }

        setVisualGlyph(0);
    }

    public GlyphIterator(GlyphIterator rhs) {

        fLayout = rhs.fLayout;
        fSets = rhs.fSets;
        fVisualSetsOrder = rhs.fVisualSetsOrder;
        fIsVertical = false;
        fBaselines = rhs.fBaselines;

        fGlyphIndexLimit = rhs.fGlyphIndexLimit;
        fTotalAdvance = rhs.fTotalAdvance;

        fVisualCurrentSet = rhs.fVisualCurrentSet;
        fCurrentSetLogicalStart = rhs.fCurrentSetLogicalStart;
        fCurrentSetVisualStart = rhs.fCurrentSetVisualStart;
        fCurrentSetAdvance = rhs.fCurrentSetAdvance;
        fCurrentBaseGlyph = rhs.fCurrentBaseGlyph;
        fLastComponentGlyph = rhs.fLastComponentGlyph;
    }

    public boolean equals(Object obj) {

        try {
            return equals((GlyphIterator) obj);
        }
        catch (ClassCastException e) {
            return false;
        }
    }

    public boolean equals(GlyphIterator rhs) {

        return (fSets == rhs.fSets &&
            fVisualSetsOrder == rhs.fVisualSetsOrder &&
            fIsVertical == rhs.fIsVertical &&
            fBaselines == rhs.fBaselines &&

            fGlyphIndexLimit == rhs.fGlyphIndexLimit &&
            fTotalAdvance == rhs.fTotalAdvance &&

            fVisualCurrentSet == rhs.fVisualCurrentSet &&
            fCurrentSetLogicalStart == rhs.fCurrentSetLogicalStart &&
            fCurrentSetVisualStart == rhs.fCurrentSetVisualStart &&
            fCurrentSetAdvance == rhs.fCurrentSetAdvance &&
            fCurrentBaseGlyph == rhs.fCurrentBaseGlyph &&
            fLastComponentGlyph == rhs.fLastComponentGlyph);
    }

    /*
     * NOTE:  This method is not meant for general consumption!  It was
     * made accessible so the sublayout code can know when to start new
     * glyphsets, and the insert/delete API could easily convert a logical
     * position to a glyphset.  Instead of using the set returned from this
     * method directly, use the iterator methods.
     */
    protected TextLayoutComponent currentGlyphSet() {

        return (fVisualSetsOrder == null)?
                            fSets[fVisualCurrentSet] :
                            fSets[fVisualSetsOrder[fVisualCurrentSet]];
    }

    /*
     * NOTE:  This method is not meant for general consumption!  It was
     * added for the editing methods on TextLayout.
     */
     protected int logicalStartOfCurrentSet() {

        return fCurrentSetLogicalStart;
     }

    /*
     * Utility to find last component glyph once fCurrentBaseGlyph has been
     * found.
     */
    private void findLastComponentGlyph() {

        TextLayoutComponent currentSet = currentGlyphSet();
        int numGlyphs = currentSet.getNumGlyphs();

        int visTestGlyph = fCurrentBaseGlyph;
        int logTestGlyph = currentSet.visualToLogicalIndex(visTestGlyph);
        int dir = glyphIsLTR()? 1 : -1;

        do {
            visTestGlyph += dir;
            logTestGlyph++;
            if (logTestGlyph >= numGlyphs)
                break;

        } while (currentSet.isComponent(logTestGlyph));

        fLastComponentGlyph = visTestGlyph - dir;
    }

    public void setVisualGlyph(int visualIndex) {

        if (visualIndex >= fGlyphIndexLimit) {
            throw new IllegalArgumentException("Index is too large in setVisualGlyph.");
        }
        // first find correct glyphset
        {
            fVisualCurrentSet = 0;
            fCurrentSetVisualStart = 0;
            fCurrentSetAdvance = 0;

            int currentChars = currentGlyphSet().getNumCharacters();

            while (fCurrentSetVisualStart + currentChars <= visualIndex) {

                fCurrentSetVisualStart += currentChars;
                fCurrentSetAdvance += currentGlyphSet().getAdvance();
                fVisualCurrentSet++;
                currentChars = currentGlyphSet().getNumCharacters();
            }
        }

        // compute fCurrentSetLogicalStart
        {
            fCurrentSetLogicalStart = 0;

            int currentSetNumber = (fVisualSetsOrder == null) ?
                fVisualCurrentSet :
                fVisualSetsOrder[fVisualCurrentSet];
            int i;

            for (i=0; i<currentSetNumber; i++) {
                fCurrentSetLogicalStart += fSets[i].getNumCharacters();
            }
        }

        // now compute fCurrentBaseGlyph and fLastComponentGlyph

        int visIndexInSet = visualIndex - fCurrentSetVisualStart;
        TextLayoutComponent currentSet = currentGlyphSet();

        if (currentSet.getNumCharacters() == currentSet.getNumGlyphs()) {

            // easy case - no component glyphs
            fCurrentBaseGlyph = fLastComponentGlyph = visIndexInSet;
        } else {

            // hard case - walk through glyphset to find base glyph

            fCurrentBaseGlyph = 0;

            while (true) {

                if (!currentSet.isComponent(visualToLogical(fCurrentBaseGlyph))) {
                    if (visIndexInSet == 0) {
                        break;
                    } else {
                        visIndexInSet--;
                    }
                }

                fCurrentBaseGlyph++;
            }

            findLastComponentGlyph();
        }
    }

    public void setLogicalGlyph(int logicalIndex) {

        if (logicalIndex >= fGlyphIndexLimit) {
            throw new IllegalArgumentException("Index is too large in setLogicalGlyph.");
        }
        if (logicalIndex < 0) {
            throw new IllegalArgumentException("Index is negative in setLogicalGlyph.");
        }

        // first find correct glyphset

        int currentSetNumber = 0;
        {
            fCurrentSetLogicalStart = 0;

            int currentChars = fSets[currentSetNumber].getNumCharacters();

            while (fCurrentSetLogicalStart + currentChars <= logicalIndex) {

                fCurrentSetLogicalStart += currentChars;
                currentSetNumber++;
                currentChars = fSets[currentSetNumber].getNumCharacters();
            }
        }


        // compute fVisualCurrentSet, fCurrentSetVisualStart, fCurrentSetAdvance

        fCurrentSetAdvance = 0;
        fCurrentSetVisualStart = 0;
        for (fVisualCurrentSet=0;
             (fVisualSetsOrder == null) ?
                 fVisualCurrentSet != currentSetNumber :
                 fVisualSetsOrder[fVisualCurrentSet] != currentSetNumber;
             fVisualCurrentSet++)
        {
            TextLayoutComponent set = currentGlyphSet();
            fCurrentSetVisualStart += set.getNumCharacters();
            fCurrentSetAdvance += set.getAdvance();
        }


        // now compute fCurrentBaseGlyph and fLastComponentGlyph
        TextLayoutComponent currentSet = currentGlyphSet();
        int logIndexInSet = logicalIndex - fCurrentSetLogicalStart;
        GlyphMetrics currentMetrics;

        // first, find correct glyph number

        int glyphNumber = 0;

        while (true) {

            if (!currentSet.isComponent(glyphNumber)) {
                if (logIndexInSet == 0) {
                    break;
                } else {
                    logIndexInSet--;
                }
            }

            glyphNumber++;
        }


        // convert logical glyph number to fCurrentBaseGlyph

        if (currentSet.isCompletelyLTR()) {
            fCurrentBaseGlyph = glyphNumber;
        } else {
            for (fCurrentBaseGlyph=0;
                 currentSet.visualToLogicalIndex(fCurrentBaseGlyph) !=
                     glyphNumber;
                 fCurrentBaseGlyph++) {}
        }

        if (currentSet.getNumGlyphs() == currentSet.getNumCharacters()) {
            fLastComponentGlyph = fCurrentBaseGlyph;
        } else {
            findLastComponentGlyph();
        }
    }

    public void firstVisualGlyph() {
        setVisualGlyph(0);
    }

    public void firstLogicalGlyph() {
        setLogicalGlyph(0);
    }

    public void lastVisualGlyph() {
        setVisualGlyph(fGlyphIndexLimit-1);
    }

    public void lastLogicalGlyph() {
        setLogicalGlyph(fGlyphIndexLimit-1);
    }

    public int visualIndex() {

        TextLayoutComponent currentSet = currentGlyphSet();
        int visIndexInSet;

        if (currentSet.getNumCharacters() == currentSet.getNumGlyphs()) {
            visIndexInSet = fCurrentBaseGlyph;
        } else {

            // walk back through glyphset to compute visual index
            visIndexInSet = 0;

            int currentGlyph;
            //GlyphMetrics metrics;

            for (currentGlyph = fCurrentBaseGlyph-1;
                 currentGlyph >= 0; currentGlyph--) {
                int glyphNumber = currentSet.visualToLogicalIndex(currentGlyph);
                if (!currentSet.isComponent(glyphNumber)) {
                    visIndexInSet++;
                }
            }
        }

        return fCurrentSetVisualStart + visIndexInSet;
    }

    public int logicalIndex() {

        TextLayoutComponent currentSet = currentGlyphSet();
        int logIndexInSet;

        int logGlyphNumber = currentSet.visualToLogicalIndex(fCurrentBaseGlyph);

        if (currentSet.getNumCharacters() == currentSet.getNumGlyphs()) {
            logIndexInSet = logGlyphNumber;
        } else {

            logIndexInSet = 0;

            for (logGlyphNumber--; logGlyphNumber >= 0; logGlyphNumber--) {
                if (!currentSet.isComponent(logGlyphNumber)) {
                    logIndexInSet++;
                }
            }
        }

        return fCurrentSetLogicalStart + logIndexInSet;
    }

    private void moveByOneVisualGlyph(boolean next) {

        if (fVisualCurrentSet == INVALID_SET) {
            throw new Error("Iterator is invalid.");
        }

        TextLayoutComponent currentSet = currentGlyphSet();
        int numGlyphs = currentSet.getNumGlyphs();

        if (next) {
            if (fCurrentBaseGlyph == numGlyphs-1 ||
                fLastComponentGlyph == numGlyphs-1) {

                int visualIndex = fCurrentSetVisualStart +
                    currentSet.getNumCharacters();
                if (visualIndex < fGlyphIndexLimit) {
                    setVisualGlyph(visualIndex);
                } else {
                    fVisualCurrentSet = INVALID_SET;
                }

                return;
            }
        } else {
            if (fCurrentBaseGlyph == 0 || fLastComponentGlyph == 0) {

                if (fCurrentSetVisualStart > 0) {
                    setVisualGlyph(fCurrentSetVisualStart-1);
                } else {
                    fVisualCurrentSet = INVALID_SET;
                }
                return;
            }
        }

        if (numGlyphs == currentSet.getNumCharacters()) {
            fCurrentBaseGlyph += next? 1 : -1;
            fLastComponentGlyph = fCurrentBaseGlyph;
        } else {

            // move fCurrentBaseGlyph to noncomponent glyph

            int d = next? 1 : -1;

            int index;
            do {
                fCurrentBaseGlyph += d;
                index = currentSet.visualToLogicalIndex(fCurrentBaseGlyph);

            } while(currentSet.isComponent(index));

            findLastComponentGlyph();
        }
    }

    public void nextVisualGlyph() {
        moveByOneVisualGlyph(true);
    }

    public void previousVisualGlyph() {
        moveByOneVisualGlyph(false);
    }

    private void moveByOneLogicalGlyph(boolean next) {
        if (fVisualCurrentSet == INVALID_SET) {
            throw new Error("Iterator is invalid.");
        }

        TextLayoutComponent currentSet = currentGlyphSet();
        int numGlyphs = currentSet.getNumGlyphs();
        //int[] visualOrder = currentSet.getVisualOrder();
        int baseLogGlyph, lastCompLogGlyph;

        baseLogGlyph = currentSet.visualToLogicalIndex(fCurrentBaseGlyph);
        lastCompLogGlyph = currentSet.visualToLogicalIndex(fLastComponentGlyph);

        if (next) {
            if (baseLogGlyph == numGlyphs-1 ||
                lastCompLogGlyph == numGlyphs-1) {

                int logicalIndex = fCurrentSetLogicalStart +
                    currentSet.getNumCharacters();
                if (logicalIndex < fGlyphIndexLimit) {
                    setLogicalGlyph(logicalIndex);
                } else {
                    fVisualCurrentSet = INVALID_SET;
                }

                return;
            }
        } else {
            if (baseLogGlyph == 0 || lastCompLogGlyph == 0) {

                if (fCurrentSetLogicalStart > 0) {
                    setLogicalGlyph(fCurrentSetLogicalStart-1);
                } else {
                    fVisualCurrentSet = INVALID_SET;
                }
                return;
            }
        }

        int d = next? 1 : -1;

        if (numGlyphs == currentSet.getNumCharacters()) {
            baseLogGlyph += d;
        } else {

            do {
                baseLogGlyph += d;
            } while (currentSet.isComponent(baseLogGlyph));
        }

        // compute fCurrentBaseGlyph

        if (currentSet.isCompletelyLTR()) {
            fCurrentBaseGlyph = baseLogGlyph;
        } else {
            for (fCurrentBaseGlyph=0;
                        currentSet.visualToLogicalIndex(fCurrentBaseGlyph) != baseLogGlyph;
                        fCurrentBaseGlyph++) {}
        }

        if (numGlyphs == currentSet.getNumCharacters()) {
            fLastComponentGlyph = fCurrentBaseGlyph;
        } else {
            findLastComponentGlyph();
        }
    }

    public void nextLogicalGlyph() {
        moveByOneLogicalGlyph(true);
    }

    public void previousLogicalGlyph() {
        moveByOneLogicalGlyph(false);
    }

    public int limit() {
        return fGlyphIndexLimit;
    }

    public float totalAdvance() {
        return fTotalAdvance;
    }

    public void invalidate() {
        fVisualCurrentSet = INVALID_SET;
    }

    public boolean isValid() {
        return fVisualCurrentSet != INVALID_SET;
    }

    public boolean isVertical() {
        return fIsVertical;
    }


/*********************************************************
    Glyph measurement access
 *********************************************************/

    private int visualToLogical(int visual) {
        //int[] visualOrder = currentGlyphSet().getVisualOrder();

        //if (visualOrder == null)
        //    return visual;
        //else
        //    return visualOrder[visual];
        return currentGlyphSet().visualToLogicalIndex(visual);
    }

    public float glyphAngle() {
        return currentGlyphSet().getItalicAngle();
    }

    public float getXAdjust() {
        return currentGlyphSet().getXAdjust();
    }

    public float getYAdjust() {
        return currentGlyphSet().getYAdjust();
    }

    public byte glyphBaseline() {
        return currentGlyphSet().getBaseline();
    }

    public float glyphAscent() {
        return currentGlyphSet().getAscent();
    }

    public float glyphDescent() {
        return currentGlyphSet().getDescent(fLayout.getAscent());
    }

    public int glyphCodeCount() {
        return Math.abs(fLastComponentGlyph-fCurrentBaseGlyph) + 1;
    }

    /*
     * Returns glyph just infos of current glyph, in logical order
     */

    public GlyphJustificationInfo[] glyphJustificationInfos() {

        TextLayoutComponent currentSet = currentGlyphSet();

        int logBase, logLast;

        logBase = visualToLogical(fCurrentBaseGlyph);
        logLast = visualToLogical(fLastComponentGlyph);

        GlyphJustificationInfo[] rval =
                    new GlyphJustificationInfo[logLast - logBase + 1];

        for (int i=0; i < rval.length; i++) {
            rval[i] = currentSet.getGlyphJustificationInfo(logBase++);
        }

        return rval;
    }

    public float[] glyphAdvances() {
        TextLayoutComponent currentSet = currentGlyphSet();
        //float[] glyphAdvances = currentSet.getAdvances();

        int logBase, logLast;

        logBase = visualToLogical(fCurrentBaseGlyph);
        logLast = visualToLogical(fLastComponentGlyph);

        float[] rval = new float[logLast - logBase + 1];

        for (int i=0; i < rval.length; i++) {
            rval[i] = currentSet.getGlyphAdvance(logBase++);
        }

        return rval;
    }

    public byte glyphLevel() {
        TextLayoutComponent currentSet = currentGlyphSet();
        return currentSet.getLevel(visualToLogical(fCurrentBaseGlyph));
    }

    public boolean glyphIsLTR()  {
        return (glyphLevel() & 0x1) == 0;
    }

    /*
     * Utilities for computing X/Y positions
     */

    private static float computeXPosition(int upToGlyph,
                                          TextLayoutComponent currentSet,
                                          float pos) {

        for (int i=0; i < upToGlyph; i++) {
            pos += currentSet.
                getGlyphXAdvance(currentSet.visualToLogicalIndex(i));
        }

        return pos;
    }

    private static float computeYPosition(int upToGlyph,
                                          TextLayoutComponent currentSet,
                                          float pos)
    {

        for (int i=0; i < upToGlyph; i++) {
            pos += currentSet.
                getGlyphYAdvance(currentSet.visualToLogicalIndex(i));
        }

        return pos;
    }

    public float glyphXPosition() {
        TextLayoutComponent currentSet = currentGlyphSet();

        float x = currentSet.getXAdjust();
        x += fIsVertical ? -fBaselines[currentSet.getBaseline()] :
            fCurrentSetAdvance;

        return computeXPosition(fCurrentBaseGlyph, currentSet, x);
    }

    public float glyphYPosition() {

        TextLayoutComponent currentSet = currentGlyphSet();

        float y = currentSet.getYAdjust();
        y += fIsVertical ? fCurrentSetAdvance :
            fBaselines[currentSet.getBaseline()];

        return computeYPosition(fCurrentBaseGlyph, currentSet, y);
    }

    public Point2D glyphPosition() {
        return new Point2D.Float(glyphXPosition(), glyphYPosition());
    }

    // return glyphXPosition for horizontal lines, and glyphYPosition
    // for vertical lines
    public float glyphLinePosition() {
        return fIsVertical? glyphYPosition() : glyphXPosition();
    }

    public float distanceToNextGlyph() {

        // total up advances between fCurrentBaseGlyph and the next non-component glyph,
        // or the end of the glyphset

        float distance = 0;
        TextLayoutComponent currentSet = currentGlyphSet();

        int numGlyphs = currentSet.getNumGlyphs();

        int visNumber = fCurrentBaseGlyph;
        int logNumber = currentSet.visualToLogicalIndex(visNumber);

        do {
            distance += currentSet.getGlyphAdvance(logNumber);
            visNumber++;
            if (visNumber >= numGlyphs) {
                break;
            }

            logNumber = currentSet.visualToLogicalIndex(visNumber);
            if (!currentSet.isComponent(logNumber))
            {
                break;
            }

        } while(true);

        return distance;
    }

    public Rectangle2D glyphBounds() {

        Rectangle2D boundsOfGlyph = getBounds2D();

        float x, y;
        TextLayoutComponent currentSet = currentGlyphSet();
        int firstVisualGlyph = Math.min(fCurrentBaseGlyph, fLastComponentGlyph);

        {
            x = currentSet.getXAdjust();
            x += fIsVertical ? -fBaselines[currentSet.getBaseline()] :
                fCurrentSetAdvance;

            x = computeXPosition(firstVisualGlyph, currentSet, x);
        }

        {
            y = currentSet.getYAdjust();
            y += fIsVertical ? fCurrentSetAdvance :
                fBaselines[currentSet.getBaseline()];

            y = computeYPosition(firstVisualGlyph, currentSet, y);
        }

        boundsOfGlyph.setBounds(boundsOfGlyph.getX() + x,
                                boundsOfGlyph.getY() + y,
                                boundsOfGlyph.getWidth(),
                                boundsOfGlyph.getHeight());

        return boundsOfGlyph;
    }

/******************************************************
    GlyphMetrics substitutes
 ******************************************************/

    private GlyphMetrics baseGlyphMetrics() {

        TextLayoutComponent currentSet = currentGlyphSet();
        int log = currentSet.visualToLogicalIndex(fCurrentBaseGlyph);

        return currentSet.getGlyphMetrics(log, fLayout);
    }

    public boolean isCombining() {

        if (fCurrentBaseGlyph != fLastComponentGlyph)
            return false;

        return baseGlyphMetrics().isCombining();
    }

    public boolean isComponent() {

        return false;
    }

    public boolean isHangingPunctuation() {

        return baseGlyphMetrics().isHangingPunctuation();
    }

    public boolean isLigature() {

        if (fCurrentBaseGlyph != fLastComponentGlyph)
            return false;

        return baseGlyphMetrics().isLigature();
    }

    public boolean isStandard() {

        return baseGlyphMetrics().isStandard();
    }

    public boolean isFiller() {

        return baseGlyphMetrics().isFiller();
    }

    public boolean isWhitespace() {

        return baseGlyphMetrics().isWhitespace();
    }

    public float glyphAdvance() {

        return baseGlyphMetrics().getAdvance();
    }

    private Rectangle2D getBounds2D() {

        if (fCurrentBaseGlyph == fLastComponentGlyph)
            return baseGlyphMetrics().getBounds2D();

        TextLayoutComponent currentSet = currentGlyphSet();

        int visGlyph, stopGlyph;
        if (fCurrentBaseGlyph < fLastComponentGlyph) {
            visGlyph = fCurrentBaseGlyph;
            stopGlyph = fLastComponentGlyph;
        } else {
            stopGlyph = fCurrentBaseGlyph;
            visGlyph = fLastComponentGlyph;
        }

        float prevAdvance = 0;
        int logGlyph;

        // iterate from visGlyph to stopGlyph, accumulating bounds rects

        Rectangle2D bounds = null;

        for (; visGlyph <= stopGlyph; visGlyph++) {
            logGlyph = currentSet.visualToLogicalIndex(visGlyph);
            GlyphMetrics metrics = currentSet.getGlyphMetrics(logGlyph, fLayout);

            if (bounds == null) {
                // sigh - no copy ctor
                bounds = new Rectangle2D.Float();
                bounds.setRect(metrics.getBounds2D());
            } else {
                Rectangle2D r = new Rectangle2D.Float();
                r.setRect(metrics.getBounds2D());
                //r.translate(fIsVertical? 0 : prevAdvance, fIsVertical? prevAdvance : 0);
                float dx = fIsVertical? 0 : prevAdvance;
                float dy = fIsVertical? prevAdvance : 0;
                r.setBounds(r.getX()+dx, r.getY()+dy, r.getWidth(),
                            r.getHeight());

                //bounds.add(r);
                add(bounds, r);
            }
            prevAdvance += currentSet.getGlyphAdvance(logGlyph);
        }
        return bounds;
    }

    // lhs += rhs;
    private static void add(Rectangle2D lhs, Rectangle2D rhs) {
        double left = lhs.getLeft();
        double top = lhs.getTop();
        double right = lhs.getRight();
        double bottom = lhs.getBottom();

        double temp;

        if (left > (temp=rhs.getLeft())) {
            left = temp;
        }
        if (top > (temp=rhs.getTop())) {
            top = temp;
        }
        if (right < (temp=rhs.getRight())) {
            right = temp;
        }
        if (bottom < (temp=rhs.getBottom())) {
            bottom = temp;
        }

        lhs.setBounds(left, top, right-left, bottom-top);
    }

    public float getLSB() {

        if (fCurrentBaseGlyph <= fLastComponentGlyph) {
            return baseGlyphMetrics().getLSB();
        }

        TextLayoutComponent currentSet = currentGlyphSet();

        int index = currentSet.visualToLogicalIndex(fLastComponentGlyph);
        return currentSet.getGlyphMetrics(index, fLayout).getLSB();
    }

    public float getRSB() {

        if (fCurrentBaseGlyph >= fLastComponentGlyph) {
            return baseGlyphMetrics().getRSB();
        }

        TextLayoutComponent currentSet = currentGlyphSet();

        int index = currentSet.visualToLogicalIndex(fLastComponentGlyph);

        return currentSet.getGlyphMetrics(index, fLayout).getRSB();
    }
}
