/*
 * @(#)StyledStringIterator.java	1.13 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package java.awt.font;

import java.text.AttributedCharacterIterator;
import java.text.AttributeSet;
import java.text.CharacterIterator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * <code>StyledStringIterator</code> implements the
 * <code>AttributedCharacterIterater</code> protocol for a
 * <code>StyledString</code>.
 *
 * @see CharacterIterator
 */
public final class StyledStringIterator
    implements AttributedCharacterIterator {

    private StyledString text;
    private int begin;
    private int end;
    private int pos;
    private int runStart = -1;
    private int runLimit = -1;
    private AttributeSet attributes = null;

    /**
     * Allocates a new styled string iterator that iterates over the entire
     * range of the styled string argument, starting at index zero.
     *
     * @param text the styled string to iterate over
     */
    StyledStringIterator(StyledString s) {
        this(s, 0, s.length(), 0);
    }
    
    /**
     * Allocates a new styled string iterator that iterates over
     * the styled string argument in the specified range, starting at
     * the specified index.
     *
     * @param text the styled string to iterate over
     * @param beginIndex the start of the range to iterate over
     * @param endIndex the end of the range to iterate over
     * @param index the initial position of the iterator
     */
    public StyledStringIterator(StyledString text, int beginIndex,
                                int endIndex, int index) {
        if (beginIndex < 0 || beginIndex > endIndex ||
            endIndex > text.length()) {
            throw new IllegalArgumentException("Invalid substring range");
        }
        
        if (index < beginIndex || index > endIndex) {
            throw new IllegalArgumentException("Invalid position");
        }
        
        this.text = text;
        this.begin = beginIndex;
        this.end = endIndex;
        this.pos = index;
    }

    /**
     * Allocates a new styled string iterator that iterates over
     * the entire range of the string argument, with all characters
     * having the specified attributes.
     *
     * @param text the string to iterate over
     * @param attributes the attributes for the characters
     */
    public StyledStringIterator(String text, AttributeSet attributes) {
        this(new StyledString(text, attributes), 0, text.length(), 0);
    }

    /**
     * Sets the position to getBeginIndex() and returns the character at that
     * position.
     * @return The character at the current position, or DONE if the current
     * position is getEndIndex().
     */
    public char first() {
        pos = begin;
        if (begin < end) {
            return text.charAt(pos);
        } else {
            return DONE;
        }
    }

    /**
     * Sets the position to getEndIndex() - 1 and returns the character at that
     * position.  If the position would be before getBeginIndex(), sets the
     * position to getBeginIndex(). This only happens if the iterator's range
     * is zero length.
     * @return The character at the current position, or DONE if the current
     * position is getEndIndex().
     */
    public char last() {
        if (begin < end) {
            pos = end - 1;
            return text.charAt(pos);
        } else {
            pos = begin;
            return DONE;
        }
    }

    /**
     * Gets the character at the current position.
     * @return The character at the current position, or DONE if the current
     * position is getEndIndex().
     * @see getIndex
     */
    public char current() {
        if (pos < end) {
            return text.charAt(pos);
        } else {
            return DONE;
        }
    }

    /**
     * Increments the iterator's position by one and returns the character
     * there.  If the new position is greater or equal
     * to getEndIndex(), sets the position to getEndIndex() and returns DONE.
     * @return The character at the new position, or DONE if the new
     * position is getEndIndex().
     */
    public char next() {
        if (++pos < end) {
            return text.charAt(pos);
        } else {
            pos = end;
            return DONE;
        }
    }

    /**
     * Decrements the iterator's position by one and returns the character
     * there.  If the new position is
     * less than getBeginIndex(), sets the position to getBeginIndex()
     * and returns DONE.
     * @return The character at the new position, or DONE if the new
     * position is getEndIndex().
     */
    public char previous() {
        if (--pos >= begin) {
            return text.charAt(pos);
        } else {
            pos = begin;
            return DONE;
        }
    }

    /**
     * Sets the position in the text and returns the character there.
     * If the new current position is getEndIndex(), returns DONE.
     * @return The character at the current position, or DONE if the current
     * position is getEndIndex().
     */
    public char setIndex(int p) {
        if (p < begin || p > end) {
            throw new IllegalArgumentException("Invalid index");
        }
        
        pos = p;
        if (pos < end) {
            return text.charAt(pos);
        } else {
            return DONE;
        }
    }

    /**
     * Returns the index of the first character in the range of the iterator.
     * @return The index of the first character in the range of the iterator.
     */
    public final int getBeginIndex() {
        return begin;
    }

    /**
     * Returns the index after the last character in the range of the iterator.
     * @return The index after the last character in the range of the iterator.
     */
    public final int getEndIndex() {
        return end;
    }

    /**
     * Returns the current index.
     * @return The current index.
     */
    public final int getIndex() {
        return pos;
    }

    /**
     * Allocates a new copy of this object.
     * @return A copy of this object.
     */
    public Object clone() {
        try {
            return super.clone();
        }
        catch (CloneNotSupportedException e) {
            throw new InternalError();
        }
    }

    /**
     * Return an iterator whose range is the range given by
     * beginIndex and endIndex, intersected with this iterator's
     * begin and end indices.
     * @param beginIndex the beginning index of the new iterator
     * @param endIndex the ending index of the new iterator
     * @return an iterator with range beginIndex, endIndex
     */
    public CharacterIterator restrict(int beginIndex, int endIndex) {
        StyledStringIterator newIter = (StyledStringIterator)this.clone();

        if (beginIndex > begin) {
            newIter.begin = beginIndex;
            if (beginIndex > pos) {
                newIter.pos = beginIndex;
            }
        }

        if (endIndex < end) {
            newIter.end = endIndex;
            if (endIndex < pos) {
                newIter.pos = endIndex;
            }
        }

        if (newIter.begin > newIter.end) {
            throw new IllegalArgumentException("disjoint range");
        }
        
        // be conservative and invalidate cache
        newIter.runStart = -1;
        newIter.runLimit = -1;
        
        return newIter;
    }
    
    /**
     * Copy a range of characters from this object into the specified array.
     * <p>
     * For performance reasons, clients iterating over the characters
     * in the iterator may prefer to index into an array.  This method
     * extracts the iterator's characters into an array supplied by the
     * client.
     * @param srcBegin the index in the iterator of the first character
     *     to copy into the array
     * @param srcEnd the index in the iterator after the last character
     *     to copy into the array.  The number of characters copied will be
     *     <code>srcEnd - srcBegin</code>.
     * @param dst the array into which to copy
     * @param dstBegin the index in dst at which to begin copying
     */
    public void getChars(int srcBegin, int srcEnd, char[] dst, int dstBegin) {
        text.getChars(srcBegin, srcEnd, dst, dstBegin);
    }
           
    /**
     * Returns the index of the first character of the run
     * with respect to all attributes on the current character.
     */
    public int getRunStart() {
        validateCache();
        
        return runStart;
    }
    
    /**
     * Returns the index of the first character of the run
     * with respect to the named attribute on the current character.
     */
    public int getRunStart(String attributeName) {
        validateCache();
        
        if (runStart > begin) {
            Object value = attributes.get(attributeName);
            return getStartInternal(new SingleSearcher(attributeName, value));
        }
        
        return begin;
    }
    
    /**
     * Returns the index of the first character of the run
     * with respect to the named attributes on the current character.
     */
    public int getRunStart(Enumeration attributeNames) {
        validateCache();
        
        if (runStart > begin) {
            return getStartInternal(new MultiSearcher(attributes, attributeNames));
        }
        
        return begin;
    }
    
    /**
     * Returns the index of the first position following the run
     * with respect to all attributes on the current character.
     */
    public int getRunLimit() {
        validateCache();
        
        return runLimit;
    }
    
    /**
     * Returns the index of the first character following the run
     * with respect to the named attribute on the current character.
     */
    public int getRunLimit(String attributeName) {
        validateCache();
        
        if (runLimit < end) {
            Object value = attributes.get(attributeName);
            return getLimitInternal(new SingleSearcher(attributeName, value));
        }
        
        return end;
    }
    
    /**
     * Returns the index of the first character following the run
     * with respect to the named attributes on the current character.
     */
    public int getRunLimit(Enumeration attributeNames) {
        validateCache();
        
        if (runLimit < end) {
            Searcher searcher = new MultiSearcher(attributes, attributeNames);
            return getLimitInternal(searcher);
        }
        
        return end;
    }
    
    /**
     * Returns an attribute set with the attributes defined on the current
     * character.
     */
    public AttributeSet getAttributes() {
        validateCache();
        return attributes;
    }
    
    /**
     * Returns the value of the named attribute for the current character.
     * Returns null if the attribute is not defined.
     * @param attributeName the name of the attribute whose value is requested.
     */
    public Object getAttribute(String attributeName) {
        validateCache();
        return attributes.get(attributeName);
    }

    /**
     * Returns the names of all attributes defined on the iterator's text
     * range.
     * The enumeration is empty if no attributes are defined.
     */
    public Enumeration getAllAttributeNames() {
        
        // Build up a hashtable of names, and return its enumeration
        int p = text.findSegmentContaining(begin);
        int e = text.findSegmentContaining(end - 1);
        
        if (p == e) {
            return attributes.names();
        } else {
            Hashtable nameSet = new Hashtable();
            
            do {
                Enumeration names = text.attrs[p].names();
            while (names.hasMoreElements()) {
                    // !!! can't put null value into hashtable
                    nameSet.put(names.nextElement(), this);
                }
            } while (p++ < e);
            
            return nameSet.keys();
        }
    }

    /**
     * Compares the equality of two StyledStringIterator objects.
     * @param obj the Object to be compared with.
     * @return true if the given object is the same as this
     * StyledStringIterator object; false otherwise.
     */
    public boolean equals(Object obj) {
        try {
            return equals((StyledStringIterator)obj);
        } catch (ClassCastException e) {
            return false;
        }
    }

    /**
     * Compares the equality of two StyledStringIterator objects.
     * @param obj the StyledStringIterator to be compared with.
     * @return true if the given object is the same as this
     * StyledStringIterator object; false otherwise.
     */
    public boolean equals(StyledStringIterator iter) {
        if (this == iter) {
            return true;
        }

        return iter != null &&
            text == iter.text &&
            pos == iter.pos &&
            begin == iter.begin &&
            end == iter.end;
    }

    /**
     * Return a hash code.
     */
    public int hashCode() {
        return pos ^ begin ^ end;
    }

    private synchronized void validateCache() {
        if (pos < runStart || pos >= runLimit) {
            runStart = Math.max(text.getRunStart(pos), begin);
            runLimit = Math.min(text.getRunLimit(pos), end);
            attributes = text.getAttributesAt(pos);
        }
    }

    private int getStartInternal(Searcher searcher) {
        int s = text.findSegmentContaining(begin);
        int p = text.findSegmentContaining(pos);

        while (p > s) {
            if (!searcher.matches(text.attrs[--p])) {
                return text.indices[p+1];
            }
        }

        return begin;
    }

    private int getLimitInternal(Searcher searcher) {
        int e = text.findSegmentContaining(end - 1) + 1;
        int p = text.findSegmentContaining(pos);

        while (++p < e) {
            if (!searcher.matches(text.attrs[p])) {
                return text.indices[p];
            }
        }

        return end;
    }

    static interface Searcher {
        boolean matches(AttributeSet set);
    }

    // this attribute must be present with the same value (or missing if value is null)
    static final class SingleSearcher implements Searcher {
        private String name;
        private Object value;

        SingleSearcher(String name, Object value) {
            this.name = name;
            this.value = value;
        }

        public boolean matches(AttributeSet set) {
            return set.contains(name, value);
        }
    }

    // all attributes must be present with same values (or missing if corresponding value is null)
    static final class MultiSearcher implements Searcher {
        private Vector names = new Vector();
        private Vector values= new Vector();

        MultiSearcher(AttributeSet attributes, Enumeration attributeNames) {
            while (attributeNames.hasMoreElements()) {
                Object name = attributeNames.nextElement();
                Object value = attributes.get((String)name);
                names.addElement(name);
                values.addElement(value);
            }
        }

        public boolean matches(AttributeSet set) {
            for (int i = 0; i < names.size(); i++) {
                if (!set.contains((String)names.elementAt(i),
                                  values.elementAt(i))) {
                    return false;
                }
            }
            return true;
        }
    }
}
