/*
 * @(#)AffineTransform.java	1.51 98/03/18
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.geom;

import java.awt.Shape;
import java.awt.geom.Point2D;

/**
 * This class represents a 2D affine transform which performs a linear
 * mapping from 2D coordinates to other 2D coordinates in a manner which
 * preserves the "straightness" and "parallelness" of lines.  Affine
 * transformations can be constructed using sequences of translations,
 * scales, flips, rotations, and shears.
 * <p>
 * Such a coordinate transformation can be represented by a 3 row by
 * 3 column matrix with an implied last row of [ 0 0 1 ] which
 * transforms source coordinates <code>(x,&nbsp;y)</code> into
 * destination coordinates <code>(x',&nbsp;y')</code> by considering
 * them to be a column vector and multiplying the coordinate vector
 * by the matrix according to the following process.
 * <pre>
 *	[ x']   [  m00  m01  m02  ] [ x ]   [ m00x + m01y + m02 ]
 *	[ y'] = [  m10  m11  m12  ] [ y ] = [ m10x + m11y + m12 ]
 *	[ 1 ]   [   0    0    1   ] [ 1 ]   [         1         ]
 * </pre>
 *
 * @version 10 Feb 1997
 * @author Jim Graham
 */
public class AffineTransform implements Cloneable, java.io.Serializable {
    /**
     * This constant indicates that the transform defined by this object
     * is an identity transform.
     * An identity transform is one in which the output coordinates are
     * always the same as the input coordinates.
     * If this transform is anything other than the identity transform,
     * the type will be a combination of the appropriate flag bits for
     * the various coordinate conversions that this transform performs.
     * @see #TRANSLATION
     * @see #UNIFORM_SCALE
     * @see #GENERAL_SCALE
     * @see #QUADRANT_ROTATION
     * @see #GENERAL_ROTATION
     * @see #GENERAL_TRANSFORM
     * @see #getType
     */
    public static final int IDENTITY = 0;

    /**
     * This flag bit indicates that the transform defined by this object
     * performs a translation in addition to the conversions indicated
     * by other flag bits.
     * A translation moves the coordinates by a constant amount in x
     * and y without changing the length or angle of vectors.
     * @see #IDENTITY
     * @see #UNIFORM_SCALE
     * @see #GENERAL_SCALE
     * @see #QUADRANT_ROTATION
     * @see #GENERAL_ROTATION
     * @see #GENERAL_TRANSFORM
     * @see #getType
     */
    public static final int TRANSLATION = 1;

    /**
     * This flag bit indicates that the transform defined by this object
     * performs a uniform scale in addition to the conversions indicated
     * by other flag bits.
     * A uniform scale multiplies the length of vectors by the same amount
     * in both the x and y directions without changing the angle between
     * vectors.
     * This flag bit is mutually exclusive with the GENERAL_SCALE flag.
     * @see #IDENTITY
     * @see #TRANSLATION
     * @see #GENERAL_SCALE
     * @see #QUADRANT_ROTATION
     * @see #GENERAL_ROTATION
     * @see #GENERAL_TRANSFORM
     * @see #getType
     */
    public static final int UNIFORM_SCALE = 2;

    /**
     * This flag bit indicates that the transform defined by this object
     * performs a general scale in addition to the conversions indicated
     * by other flag bits.
     * A general scale multiplies the length of vectors by different
     * amounts in the x and y directions without changing the angle
     * between vectors.
     * This flag bit is mutually exclusive with the UNIFORM_SCALE flag.
     * @see #IDENTITY
     * @see #TRANSLATION
     * @see #UNIFORM_SCALE
     * @see #QUADRANT_ROTATION
     * @see #GENERAL_ROTATION
     * @see #GENERAL_TRANSFORM
     * @see #getType
     */
    public static final int GENERAL_SCALE = 4;

    /**
     * This flag is a bit mask for any of the scale flag bits.
     */
    public static final int ANY_SCALE_MASK = (UNIFORM_SCALE |
					      GENERAL_SCALE);

    /**
     * This flag bit indicates that the transform defined by this object
     * performs a quadrant rotation by some multiple of 90 degrees in
     * addition to the conversions indicated by other flag bits.
     * A rotation changes the angles of vectors by the same amount
     * regardless of the original direction of the vector without
     * changing the length of the vector.
     * This flag bit is mutually exclusive with the GENERAL_ROTATION flag.
     * @see #IDENTITY
     * @see #TRANSLATION
     * @see #UNIFORM_SCALE
     * @see #GENERAL_SCALE
     * @see #GENERAL_ROTATION
     * @see #GENERAL_TRANSFORM
     * @see #getType
     */
    public static final int QUADRANT_ROTATION = 8;

    /**
     * This flag bit indicates that the transform defined by this object
     * performs a rotation by an arbitrary angle in addition to the
     * conversions indicated by other flag bits.
     * A rotation changes the angles of vectors by the same amount
     * regardless of the original direction of the vector without
     * changing the length of the vector.
     * This flag bit is mutually exclusive with the QUADRANT_ROTATION flag.
     * @see #IDENTITY
     * @see #TRANSLATION
     * @see #UNIFORM_SCALE
     * @see #GENERAL_SCALE
     * @see #QUADRANT_ROTATION
     * @see #GENERAL_TRANSFORM
     * @see #getType
     */
    public static final int GENERAL_ROTATION = 16;

    /**
     * This flag is a bit mask for any of the rotation flag bits.
     */
    public static final int ANY_ROTATION_MASK = (QUADRANT_ROTATION |
						 GENERAL_ROTATION);

    /**
     * This flag bit indicates that the transform defined by this object
     * performs an arbitrary conversion of the input coordinates.
     * This flag bit is mutually exclusive with any of the other flags.
     * @see #IDENTITY
     * @see #TRANSLATION
     * @see #UNIFORM_SCALE
     * @see #GENERAL_SCALE
     * @see #QUADRANT_ROTATION
     * @see #GENERAL_ROTATION
     * @see #getType
     */
    public static final int GENERAL_TRANSFORM = 32;

    /**
     * This constant is used for the internal state variable to indicate
     * that the translation components of the matrix (m02 and m12) need
     * to be added to complete the transformation equation of this transform.
     * @see #IDENTITY
     * @see #APPLY_SCALE
     * @see #APPLY_SHEAR
     * @see #state
     */
    static final int APPLY_TRANSLATE = 1;

    /**
     * This constant is used for the internal state variable to indicate
     * that the scaling components of the matrix (m00 and m11) need
     * to be factored in to complete the transformation equation of
     * this transform.  If the APPLY_SHEAR bit is also set, then it
     * indicates that the scaling components are not both 0.0.  If the
     * APPLY_SHEAR bit is not also set, then it indicates that the
     * scaling components are not both 1.0.  If neither the APPLY_SHEAR
     * nor the APPLY_SCALE bits are set, then the scaling components
     * are both 1.0 which means that the x and y components contribute
     * to the transformed coordinate, but they are not multiplied by
     * any scaling factor.
     * @see #IDENTITY
     * @see #APPLY_TRANSLATE
     * @see #APPLY_SHEAR
     * @see #state
     */
    static final int APPLY_SCALE = 2;

    /**
     * This constant is used for the internal state variable to indicate
     * that the shearing components of the matrix (m01 and m10) need
     * to be factored in to complete the transformation equation of this
     * transform.  The presence of this bit in the state variable changes
     * the interpretation of the APPLY_SCALE bit as indicated in its
     * documentation.
     * @see #IDENTITY
     * @see #APPLY_TRANSLATE
     * @see #APPLY_SCALE
     * @see #state
     */
    static final int APPLY_SHEAR = 4;

    /*
     * For methods which combine together the state of two separate
     * transforms and dispatch based upon the combination, these constants
     * specify how far to shift one of the states so that the two states 
     * are mutually non-interfering and provide constants for testing the
     * bits of the shifted (HI) state.  The methods in this class use
     * the convention that the state of "this" transform is unshifted and
     * the state of the "other" or "argument" transform is shifted (HI).
     */
    private static final int HI_SHIFT = 3;
    private static final int HI_TRANSLATE = APPLY_TRANSLATE << HI_SHIFT;
    private static final int HI_SCALE = APPLY_SCALE << HI_SHIFT;
    private static final int HI_SHEAR = APPLY_SHEAR << HI_SHIFT;

    double m00, m10, m01, m11, m02, m12;

    /**
     * This field keeps track of which components of the matrix need to
     * be applied when performing a transformation.
     * @see #IDENTITY
     * @see #APPLY_TRANSLATE
     * @see #APPLY_SCALE
     * @see #APPLY_SHEAR
     */
    transient int state;

    private AffineTransform(double m00, double m10,
			    double m01, double m11,
			    double m02, double m12,
			    int state) {
	this.m00 = m00;
	this.m10 = m10;
	this.m01 = m01;
	this.m11 = m11;
	this.m02 = m02;
	this.m12 = m12;
	this.state = state;
    }

    /**
     * Constructs a new AffineTransform representing the Identity
     * transformation.
     */
    public AffineTransform() {
	m00 = m11 = 1.0;
	// m01 = m10 = m02 = m12 = 0.0;		/* Not needed. */
	// state = IDENTITY;			/* Not needed. */
    }

    /**
     * Constructs a new AffineTransform representing a copy of
     * another AffineTransform object.
     * @param Tx the AffineTransform object from which to copy the
     * transformation
     */
    public AffineTransform(AffineTransform Tx) {
	this.m00 = Tx.m00;
	this.m10 = Tx.m10;
	this.m01 = Tx.m01;
	this.m11 = Tx.m11;
	this.m02 = Tx.m02;
	this.m12 = Tx.m12;
	this.state = Tx.state;
    }

    /**
     * Constructs a new AffineTransform from 6 floating point values
     * representing the 6 specifiable entries of the 3x3 transformation matrix.
     */
    public AffineTransform(float m00, float m10,
			   float m01, float m11,
			   float m02, float m12) {
	this.m00 = m00;
	this.m10 = m10;
	this.m01 = m01;
	this.m11 = m11;
	this.m02 = m02;
	this.m12 = m12;
	updateState();
    }

    /**
     * Constructs a new AffineTransform from an array of floating point
     * values representing either the 4 non-translation or the 6 specifiable
     * entries of
     * the 3x3 transformation matrix.  The values are retrieved from the
     * array as { m00 m10 m01 m11 [m02 m12] }.
     * @param flatmatrix The float array containing the values to be set
     * in the new AffineTransform object. The length of the array is 
     * assumed to be at least 4. If the length of the array is 
     * less than 6, only the first 4 values are taken. If the length of
     * the array is greater than 6, the first 6 values are taken.
     */
    public AffineTransform(float[] flatmatrix) {
	m00 = flatmatrix[0];
	m10 = flatmatrix[1];
	m01 = flatmatrix[2];
	m11 = flatmatrix[3];
	if (flatmatrix.length > 5) {
	    m02 = flatmatrix[4];
	    m12 = flatmatrix[5];
	}
	updateState();
    }

    /**
     * Constructs a new AffineTransform from a 2-dimensional array of
     * floating point values representing either the 4 non-translation or the
     * 6 specifiable entries of the 3x3 transformation matrix.  The values are
     * retrieved from the 2-dimensional array using the row index as
     * the first subscript and the column index as the second.
     * @param matrix The matrix used to construct the new AffineTransform
     * object. Matrix is assumed to be at least 2x2. If the matrix
     * is smaller than 2x3, only the left upper 2x2 matrix is taken. If the
     * matrix is larger than 2x3, the left upper 2x3 matrix is taken.
     */
    public AffineTransform(float[][] matrix) {
	m00 = matrix[0][0];
	m10 = matrix[1][0];
	m01 = matrix[0][1];
	m11 = matrix[1][1];
	if (matrix[0].length > 2 && matrix[1].length > 2) {
	    m02 = matrix[0][2];
	    m12 = matrix[1][2];
	}
	updateState();
    }

    /**
     * Constructs a new AffineTransform from 6 double precision values
     * representing the 6 specifiable entries of the 3x3 transformation matrix.
     */
    public AffineTransform(double m00, double m10,
			   double m01, double m11,
			   double m02, double m12) {
	this.m00 = m00;
	this.m10 = m10;
	this.m01 = m01;
	this.m11 = m11;
	this.m02 = m02;
	this.m12 = m12;
	updateState();
    }

    /**
     * Constructs a new AffineTransform from an array of double precision
     * values representing either the 4 non-translation or the 6 specifiable
     * entries of
     * the 3x3 transformation matrix.  The values are retrieved from the
     * array as { m00 m10 m01 m11 [m02 m12] }.
     * @param flatmatrix The double array containing the values to be set
     * in the new AffineTransform object. The length of the array is 
     * assumed to be at least 4. If the length of the array is 
     * less than 6, only the first 4 values are taken. If the length of
     * the array is greater than 6, the first 6 values are taken.
     */
    public AffineTransform(double[] flatmatrix) {
	m00 = flatmatrix[0];
	m10 = flatmatrix[1];
	m01 = flatmatrix[2];
	m11 = flatmatrix[3];
	if (flatmatrix.length > 5) {
	    m02 = flatmatrix[4];
	    m12 = flatmatrix[5];
	}
	updateState();
    }

    /**
     * Constructs a new AffineTransform from a 2-dimensional array of
     * double precision values representing either the 4 non-translation or the
     * 6 specifiable entries of the 3x3 transformation matrix.  The values are
     * retrieved from the 2-dimensional array using the row index as
     * the first subscript and the column index as the second.
     * @param matrix The matrix used to construct the new AffineTransform
     * object. Matrix is assumed to be at least 2x2. If the matrix
     * is smaller than 2x3, only the left upper 2x2 matrix is taken. If the
     * matrix is larger than 2x3, the left upper 2x3 matrix is taken.
     */
    public AffineTransform(double[][] matrix) {
	m00 = matrix[0][0];
	m10 = matrix[1][0];
	m01 = matrix[0][1];
	m11 = matrix[1][1];
	if (matrix[0].length > 2 && matrix[1].length > 2) {
	    m02 = matrix[0][2];
	    m12 = matrix[1][2];
	}
	updateState();
    }

    /**
     * Returns a transform representing a translation transformation.
     * The matrix representing by the returned transform transform will be:
     * <pre>
     *		[   1    0    tx  ]
     *		[   0    1    ty  ]
     *		[   0    0    1   ]
     * </pre>
     * @param tx The distance by which coordinates are translated in the
     * X axis direction
     * @param ty The distance by which coordinates are translated in the
     * Y axis direction
     */
    public static AffineTransform getTranslateInstance(double tx, double ty) {
	AffineTransform Tx = new AffineTransform();
	Tx.setToTranslation(tx, ty);
	return Tx;
    }

    /**
     * Returns a transform representing a rotation transformation.
     * The matrix representing by the returned transform transform will be:
     * <pre>
     *		[   cos(theta)    -sin(theta)    0   ]
     *		[   sin(theta)     cos(theta)    0   ]
     *		[       0              0         1   ]
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     */
    public static AffineTransform getRotateInstance(double theta) {
	AffineTransform Tx = new AffineTransform();
	Tx.setToRotation(theta);
	return Tx;
    }

    /**
     * Returns a transform representing a translated rotation transformation.
     * This is equivalent to the following sequence of calls:
     * <pre>
     *		AffineTransform Tx = new AffineTransform();
     *		Tx.setToTranslation(x, y);
     *		Tx.rotate(theta);
     *		Tx.translate(-x, -y);
     * </pre>
     * The matrix representing by the returned transform transform will be:
     * <pre>
     *		[   cos(theta)    -sin(theta)    x-x*cos+y*sin  ]
     *		[   sin(theta)     cos(theta)    y-x*sin-y*cos  ]
     *		[       0              0               1        ]
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     * @param x The x coordinate of the origin of the rotation
     * @param y The x coordinate of the origin of the rotation
     */
    public static AffineTransform getRotateInstance(double theta,
						    double x, double y) {
	AffineTransform Tx = new AffineTransform();
	Tx.setToRotation(theta, x, y);
	return Tx;
    }

    /**
     * Returns a transform representing a scaling transformation.
     * The matrix representing by the returned transform transform will be:
     * <pre>
     *		[   sx   0    0   ]
     *		[   0    sy   0   ]
     *		[   0    0    1   ]
     * </pre>
     * @param sx The factor by which coordinates are scaled along the
     * X axis direction
     * @param sy The factor by which coordinates are scaled along the
     * Y axis direction
     */
    public static AffineTransform getScaleInstance(double sx, double sy) {
	AffineTransform Tx = new AffineTransform();
	Tx.setToScale(sx, sy);
	return Tx;
    }

    /**
     * Returns a transform representing a shearing transformation.
     * The matrix representing by the returned transform transform will be:
     * <pre>
     *		[   1   shx   0   ]
     *		[  shy   1    0   ]
     *		[   0    0    1   ]
     * </pre>
     * @param shx The factor by which coordinates are shifted towards the
     * positive X axis direction according to their Y coordinate
     * @param shy The factor by which coordinates are shifted towards the
     * positive Y axis direction according to their X coordinate
     */
    public static AffineTransform getShearInstance(double shx, double shy) {
	AffineTransform Tx = new AffineTransform();
	Tx.setToShear(shx, shy);
	return Tx;
    }

    /**
     * Retrieves the flag bits describing the conversion properties of
     * this transform.  The return value will be a combination of the
     * appriopriate conversion flag bits.
     * @return the OR combination of any of the indicated flags that
     * apply to this transform
     * @see #TRANSLATION
     * @see #UNIFORM_SCALE
     * @see #GENERAL_SCALE
     * @see #QUADRANT_ROTATION
     * @see #GENERAL_ROTATION
     * @see #GENERAL_TRANSFORM
     */
    public int getType() {
	int ret = IDENTITY;
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    ret = TRANSLATION;
	    /* NOBREAK */
	case (APPLY_SHEAR | APPLY_SCALE):
	    if (m00 * m01 + m10 * m11 != 0) {
		return GENERAL_TRANSFORM;
	    }
	    ret |= GENERAL_ROTATION;
	    if (m00 != m11 || m01 != -m10) {
		ret |= GENERAL_SCALE;
	    } else if (m00 * m11 - m01 * m10 != 1.0) {
		ret |= UNIFORM_SCALE;
	    }
	    break;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    ret = TRANSLATION;
	    /* NOBREAK */
	case (APPLY_SHEAR):
	    ret |= QUADRANT_ROTATION;
	    if (m01 != -m10) {
		ret |= GENERAL_SCALE;
	    } else if (m01 != 1.0 && m10 != 1.0) {
		ret |= UNIFORM_SCALE;
	    }
	    break;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    ret = TRANSLATION;
	    /* NOBREAK */
	case (APPLY_SCALE):
	    if (m00 < 0.0 && m11 < 0.0) {
		ret |= QUADRANT_ROTATION;
		// else it is some sort of flip or 0 rotation
		// let's not classify simple flips as general rotations
	    }
	    if (m00 == m11) {
		ret |= UNIFORM_SCALE;
	    } else {
		ret |= GENERAL_SCALE;
	    }
	    break;
	case (APPLY_TRANSLATE):
	    ret = TRANSLATION;
	    break;
	case (IDENTITY):
	    break;
	}
	return ret;
    }

    /**
     * Returns the determinant of the matrix representation of the transform.
     * The determinant is useful both to determine if the transform can
     * be inverted and also to get a single value representing the
     * combined X and Y scaling of the transform.
     * <p>
     * If the determinant is non-zero, then this transform will be
     * invertable and the various methods which depend on the inverse
     * transform will not need to throw a NoninvertibleTransformException.
     * If the determinant is zero, then this transform can not be
     * inverted since the transform maps all input coordinates onto
     * a line or a point.
     * If the determinant is near enough to zero, then inverse transform
     * operations may not carry enough precision to produce meaningful
     * results.
     * <p>
     * If this transform represents a uniform scale, as indicated by
     * the getType method, then the determinant will also represent
     * the square of the uniform scale factor by which all of the
     * points are expanded from or contracted towards the origin.
     * If this transform represents a non-uniform scale or more general
     * transform, then the determinant is not likely to represent a
     * value useful for any other purpose than determining if inverse
     * transforms are possible.
     * <p>
     * Mathematically, the determinant is calculated using the formula:
     * <pre>
     *		|  m00  m01  m02  |
     *		|  m10  m11  m12  |  =  m00 * m11 - m01 * m10
     *		|   0    0    1   |
     * </pre>
     *
     * @return the determinant of the matrix used to transform the
     * coordinates
     * @see #getType
     * @see #createInverse
     * @see #inverseTransform
     * @see #UNIFORM_SCALE
     */
    public double getDeterminant() {
	return m00 * m11 - m01 * m10;
    }

    /**
     * Manually recalculates the state of the transform when the matrix
     * changes too much to predict the effects on the state.
     */
    void updateState() {
	double M0, M1;
	if ((M0 = m01) == 0.0 && (M1 = m10) == 0.0) {
	    if ((M0 = m00) == 1.0 && (M1 = m11) == 1.0) {
		state = ((m02 != 0.0 || m12 != 0.0) ? APPLY_TRANSLATE : 0);
	    } else {
		state = (APPLY_SCALE |
			 ((m02 != 0.0 || m12 != 0.0) ? APPLY_TRANSLATE : 0));
	    }
	} else {
	    double M2, M3;
	    if ((M2 = m00) == 0.0 && (M3 = m11) == 0.0) {
		state = (APPLY_SHEAR |
			 ((m02 != 0.0 || m12 != 0.0) ? APPLY_TRANSLATE : 0));
	    } else {
		state = (APPLY_SHEAR |
			 APPLY_SCALE |
			 ((m02 != 0.0 || m12 != 0.0) ? APPLY_TRANSLATE : 0));
	    }
	}
    }

    /*
     * Convenience method used internally to throw exceptions when
     * a case was forgotten in a switch statement.
     */
    private void stateError() {
	throw new InternalError("missing case in transform state switch");
    }
    
    /**
     * Retrieves the 6 specifiable values in the 3x3 affine transformation
     * matrix into
     * an array of double precisions values.  The values are stored into
     * the array as { m00 m10 m01 m11 m02 m12 }.
     * @param flatmatrix The double array used to store the returned values.
     * The length of the array is assumed to be at least 6.
     * @see #getScaleX
     * @see #getScaleY
     * @see #getShearX
     * @see #getShearY
     * @see #getTranslateX
     * @see #getTranslateY
     */
    public void getMatrix(double[] flatmatrix) {
	flatmatrix[0] = m00;
	flatmatrix[1] = m10;
	flatmatrix[2] = m01;
	flatmatrix[3] = m11;
	if (flatmatrix.length > 5) {
	    flatmatrix[4] = m02;
	    flatmatrix[5] = m12;
	}
    }

    /**
     * Retrieves the 6 specifiable values in the 3x3 affine transformation
     * matrix into
     * a 2-dimensional array of double precision values.  The values are
     * stored into the 2-dimensional array using the row index as the
     * first subscript and the column index as the second.
     * @param matrix The 2-dimensional double array to store the returned
     * values.  The array is assumed to be at least 2x3.
     * @see #getScaleX
     * @see #getScaleY
     * @see #getShearX
     * @see #getShearY
     * @see #getTranslateX
     * @see #getTranslateY
     */
    public void getMatrix(double[][] matrix) {
	matrix[0][0] = m00;
	matrix[1][0] = m10;
	matrix[0][1] = m01;
	matrix[1][1] = m11;
	matrix[0][2] = m02;
	matrix[1][2] = m12;
    }

    /**
     * Returns the X coordinate scaling element (m00) of the 3x3
     * affine transformation matrix.
     * @see #getMatrix
     */
    public double getScaleX() {
	return m00;
    }

    /**
     * Returns the Y coordinate scaling element (m11) of the 3x3
     * affine transformation matrix.
     * @see #getMatrix
     */
    public double getScaleY() {
	return m11;
    }

    /**
     * Returns the X coordinate shearing element (m01) of the 3x3
     * affine transformation matrix.
     * @see #getMatrix
     */
    public double getShearX() {
	return m01;
    }

    /**
     * Returns the Y coordinate shearing element (m10) of the 3x3
     * affine transformation matrix.
     * @see #getMatrix
     */
    public double getShearY() {
	return m10;
    }

    /**
     * Returns the X coordinate translation element (m02) of the 3x3
     * affine transformation matrix.
     * @see #getMatrix
     */
    public double getTranslateX() {
	return m02;
    }

    /**
     * Returns the Y coordinate translation element (m12) of the 3x3
     * affine transformation matrix.
     * @see #getMatrix
     */
    public double getTranslateY() {
	return m12;
    }

    /**
     * Concatenates this transform with a translation transformation.
     * This is equivalent to calling concatenate(T), where T is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   1    0    tx  ]
     *		[   0    1    ty  ]
     *		[   0    0    1   ]
     * </pre>
     */
    public void translate(double tx, double ty) {
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    m02 = tx * m00 + ty * m01 + m02;
	    m12 = tx * m10 + ty * m11 + m12;
	    return;
	case (APPLY_SHEAR | APPLY_SCALE):
	    m02 = tx * m00 + ty * m01;
	    m12 = tx * m10 + ty * m11;
	    state = APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE;
	    return;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    m02 = ty * m01 + m02;
	    m12 = tx * m10 + m12;
	    return;
	case (APPLY_SHEAR):
	    m02 = ty * m01;
	    m12 = tx * m10;
	    state = APPLY_SHEAR | APPLY_TRANSLATE;
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    m02 = tx * m00 + m02;
	    m12 = ty * m11 + m12;
	    return;
	case (APPLY_SCALE):
	    m02 = tx * m00;
	    m12 = ty * m11;
	    state = APPLY_SCALE | APPLY_TRANSLATE;
	    return;
	case (APPLY_TRANSLATE):
	    m02 = tx + m02;
	    m12 = ty + m12;
	    return;
	case (IDENTITY):
	    m02 = tx;
	    m12 = ty;
	    state = APPLY_TRANSLATE;
	    return;
	}
    }

    /**
     * Concatenates this transform with a rotation transformation.
     * This is equivalent to calling concatenate(R), where R is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   cos(theta)    -sin(theta)    0   ]
     *		[   sin(theta)     cos(theta)    0   ]
     *		[       0              0         1   ]
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     */
    public void rotate(double theta) {
	double sin = Math.sin(theta);
	double cos = Math.cos(theta);
	if (Math.abs(sin) < 1E-15) {
	    if (cos < 0.0) {
		m00 = -m00;
		m11 = -m11;
		int state = this.state;
		if ((state & (APPLY_SHEAR)) != 0) {
		    // If there was a shear, then this rotation has no
		    // effect on the state.
		    m01 = -m01;
		    m10 = -m10;
		} else {
		    // No shear means the SCALE state may toggle when
		    // m00 and m11 are negated.
		    if (m00 == 1.0 && m11 == 1.0) {
			this.state = state & ~APPLY_SCALE;
		    } else {
			this.state = state | APPLY_SCALE;
		    }
		}
	    }
	    return;
	}
	if (Math.abs(cos) < 1E-15) {
	    if (sin < 0.0) {
		double M0 = m00;
		m00 = -m01;
		m01 = M0;
		M0 = m10;
		m10 = -m11;
		m11 = M0;
	    } else {
		double M0 = m00;
		m00 = m01;
		m01 = -M0;
		M0 = m10;
		m10 = m11;
		m11 = -M0;
	    }
	    state = rot90conversion[state];
	    return;
	}
	double M0, M1;
	M0 = m00;
	M1 = m01;
	m00 =  cos * M0 + sin * M1;
	m01 = -sin * M0 + cos * M1;
	M0 = m10;
	M1 = m11;
	m10 =  cos * M0 + sin * M1;
	m11 = -sin * M0 + cos * M1;
	updateState();
    }
    private static int rot90conversion[] = {
	APPLY_SHEAR, APPLY_SHEAR | APPLY_TRANSLATE,
	APPLY_SHEAR, APPLY_SHEAR | APPLY_TRANSLATE,
	APPLY_SCALE, APPLY_SCALE | APPLY_TRANSLATE,
	APPLY_SHEAR | APPLY_SCALE,
	APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE,
    };

    /**
     * Concatenates this transform with a translated rotation transformation.
     * This is equivalent to the following sequence of calls:
     * <pre>
     *		translate(x, y);
     *		rotate(theta);
     *		translate(-x, -y);
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     * @param x The x coordinate of the origin of the rotation
     * @param y The x coordinate of the origin of the rotation
     */
    public void rotate(double theta, double x, double y) {
	// REMIND: Simple for now - optimize later
	translate(x, y);
	rotate(theta);
	translate(-x, -y);
    }

    /**
     * Concatenates this transform with a scaling transformation.
     * This is equivalent to calling concatenate(S), where S is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   sx   0    0   ]
     *		[   0    sy   0   ]
     *		[   0    0    1   ]
     * </pre>
     */
    public void scale(double sx, double sy) {
	int state = this.state;
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SHEAR | APPLY_SCALE):
	    m00 *= sx;
	    m11 *= sy;
	    /* NOBREAK */
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	case (APPLY_SHEAR):
	    m01 *= sy;
	    m10 *= sx;
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SCALE):
	    m00 *= sx;
	    m11 *= sy;
	    return;
	case (APPLY_TRANSLATE):
	case (IDENTITY):
	    m00 = sx;
	    m11 = sy;
	    this.state = state | APPLY_SCALE;
	    return;
	}
    }

    /**
     * Concatenates this transform with a shearing transformation.
     * This is equivalent to calling concatenate(SH), where SH is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   1   shx   0   ]
     *		[  shy   1    0   ]
     *		[   0    0    1   ]
     * </pre>
     * @param shx The factor by which coordinates are shifted towards the
     * positive X axis direction according to their Y coordinate
     * @param shy The factor by which coordinates are shifted towards the
     * positive Y axis direction according to their X coordinate
     */
    public void shear(double shx, double shy) {
	int state = this.state;
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SHEAR | APPLY_SCALE):
	    double M0, M1;
	    M0 = m00;
	    M1 = m01;
	    m00 = M0 + M1 * shy;
	    m01 = M0 * shx + M1;

	    M0 = m10;
	    M1 = m11;
	    m10 = M0 + M1 * shy;
	    m11 = M0 * shx + M1;
	    return;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	case (APPLY_SHEAR):
	    m00 = m01 * shy;
	    m11 = m10 * shx;
	    this.state = state | APPLY_SCALE;
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SCALE):
	    m01 = m00 * shx;
	    m10 = m11 * shy;
	    this.state = state | APPLY_SHEAR;
	    return;
	case (APPLY_TRANSLATE):
	case (IDENTITY):
	    m01 = shx;
	    m10 = shy;
	    this.state = state | APPLY_SCALE | APPLY_SHEAR;
	    return;
	}
    }

    /**
     * Resets this transform to the Identity transform.
     */
    public void setToIdentity() {
	m00 = m11 = 1.0;
	m10 = m01 = m02 = m12 = 0.0;
	state = IDENTITY;
    }

    /**
     * Sets this transform to a translation transformation.
     * The matrix representing this transform becomes:
     * <pre>
     *		[   1    0    tx  ]
     *		[   0    1    ty  ]
     *		[   0    0    1   ]
     * </pre>
     * @param tx The distance by which coordinates are translated in the
     * X axis direction
     * @param ty The distance by which coordinates are translated in the
     * Y axis direction
     */
    public void setToTranslation(double tx, double ty) {
	m00 = 1.0;
	m10 = 0.0;
	m01 = 0.0;
	m11 = 1.0;
	m02 = tx;
	m12 = ty;
	state = APPLY_TRANSLATE;
    }

    /**
     * Sets this transform to a rotation transformation.
     * The matrix representing this transform becomes:
     * <pre>
     *		[   cos(theta)    -sin(theta)    0   ]
     *		[   sin(theta)     cos(theta)    0   ]
     *		[       0              0         1   ]
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     */
    public void setToRotation(double theta) {
	m02 = 0.0;
	m12 = 0.0;
	double sin = Math.sin(theta);
	double cos = Math.cos(theta);
	if (Math.abs(sin) < 1E-15) {
	    m01 = m10 = 0.0;
	    if (cos < 0) {
		m00 = m11 = -1.0;
		state = APPLY_SCALE;
	    } else {
		m00 = m11 = 1.0;
		state = IDENTITY;
	    }
	    return;
	}
	if (Math.abs(cos) < 1E-15) {
	    m00 = m11 = 0.0;
	    if (sin < 0.0) {
		m01 = 1.0;
		m10 = -1.0;
	    } else {
		m01 = -1.0;
		m10 = 1.0;
	    }
	    state = APPLY_SHEAR;
	    return;
	}
	m00 = cos;
	m01 = -sin;
	m10 = sin;
	m11 = cos;
	state = APPLY_SHEAR | APPLY_SCALE;
	return;
    }

    /**
     * Sets this transform to a translated rotation transformation.
     * This is equivalent to the following sequence of calls:
     * <pre>
     *		setToTranslation(x, y);
     *		rotate(theta);
     *		translate(-x, -y);
     * </pre>
     * The matrix representing this transform becomes:
     * <pre>
     *		[   cos(theta)    -sin(theta)    x-x*cos+y*sin  ]
     *		[   sin(theta)     cos(theta)    y-x*sin-y*cos  ]
     *		[       0              0               1        ]
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     * @param x The x coordinate of the origin of the rotation
     * @param y The x coordinate of the origin of the rotation
     */
    public void setToRotation(double theta, double x, double y) {
	setToRotation(theta);
	double sin = m10;
	double oneMinusCos = 1.0 - m00;
	m02 = x * oneMinusCos + y * sin;
	m12 = y * oneMinusCos - x * sin;
	state |= APPLY_TRANSLATE;
	return;
    }

    /**
     * Sets this transform to a scaling transformation.
     * The matrix representing this transform becomes:
     * <pre>
     *		[   sx   0    0   ]
     *		[   0    sy   0   ]
     *		[   0    0    1   ]
     * </pre>
     * @param sx The factor by which coordinates are scaled along the
     * X axis direction
     * @param sy The factor by which coordinates are scaled along the
     * Y axis direction
     */
    public void setToScale(double sx, double sy) {
	m00 = sx;
	m10 = 0.0;
	m01 = 0.0;
	m11 = sy;
	m02 = 0.0;
	m12 = 0.0;
	state = APPLY_SCALE;
    }

    /**
     * Sets this transform to a shearing transformation.
     * The matrix representing this transform becomes:
     * <pre>
     *		[   1   shx   0   ]
     *		[  shy   1    0   ]
     *		[   0    0    1   ]
     * </pre>
     * @param shx The factor by which coordinates are shifted towards the
     * positive X axis direction according to their Y coordinate
     * @param shy The factor by which coordinates are shifted towards the
     * positive Y axis direction according to their X coordinate
     */
    public void setToShear(double shx, double shy) {
	m00 = 1.0;
	m01 = shx;
	m10 = shy;
	m11 = 1.0;
	m02 = 0.0;
	m12 = 0.0;
	state = APPLY_SHEAR | APPLY_SCALE;
    }

    /**
     * Sets this transform to a copy of the transform in the indicated
     * AffineTransform object.
     * @param Tx the AffineTransform object from which to copy the transform
     */
    public void setTransform(AffineTransform Tx) {
	this.m00 = Tx.m00;
	this.m10 = Tx.m10;
	this.m01 = Tx.m01;
	this.m11 = Tx.m11;
	this.m02 = Tx.m02;
	this.m12 = Tx.m12;
	this.state = Tx.state;
    }

    /**
     * Sets this transform to the matrix specified by the 6
     * double precision values.
     */
    public void setTransform(double m00, double m10,
			     double m01, double m11,
			     double m02, double m12) {
	this.m00 = m00;
	this.m10 = m10;
	this.m01 = m01;
	this.m11 = m11;
	this.m02 = m02;
	this.m12 = m12;
	updateState();
    }

    /**
     * Concatenates an AffineTransform Tx to this AffineTransform Cx
     * in the most commonly useful way to provide a new user space
     * which is mapped to the former user space by Tx.
     * Cx is updated to perform the combined transformation.
     * Transforming a point p by the updated transform Cx' is
     * equivalent to first transforming p by Tx and then transforming
     * the result by the original transform Cx. In other words,
     * Cx'(p) = Cx(Tx(p)).  In matrix notation, if this transform Cx
     * is represented by the matrix [this] and Tx is represented by
     * the matrix [Tx], then this method does the following:
     * <pre>
     *		[this] = [this] x [Tx]
     * </pre>
     * @param Tx The AffineTransform object to be concatenated with
     * this AffineTransform object.
     * @see #preConcatenate
     */
    public void concatenate(AffineTransform Tx) {
        double M0, M1;
	double T00, T01, T10, T11;
	double T02, T12;
	int mystate = state;
	int txstate = Tx.state;
	switch ((txstate << HI_SHIFT) | mystate) {

	    /* ---------- Tx == IDENTITY cases ---------- */
	case (IDENTITY | IDENTITY):
	case (IDENTITY | APPLY_TRANSLATE):
	case (IDENTITY | APPLY_SCALE):
	case (IDENTITY | APPLY_SCALE | APPLY_TRANSLATE):
	case (IDENTITY | APPLY_SHEAR):
	case (IDENTITY | APPLY_SHEAR | APPLY_TRANSLATE):
	case (IDENTITY | APPLY_SHEAR | APPLY_SCALE):
	case (IDENTITY | APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    return;

	    /* ---------- this == IDENTITY cases ---------- */
	case (HI_SHEAR | HI_SCALE | HI_TRANSLATE | IDENTITY):
	    m01 = Tx.m01;
	    m10 = Tx.m10;
	    /* NOBREAK */
	case (HI_SCALE | HI_TRANSLATE | IDENTITY):
	    m00 = Tx.m00;
	    m11 = Tx.m11;
	    /* NOBREAK */
	case (HI_TRANSLATE | IDENTITY):
	    m02 = Tx.m02;
	    m12 = Tx.m12;
	    state = txstate;
	    return;
	case (HI_SHEAR | HI_SCALE | IDENTITY):
	    m01 = Tx.m01;
	    m10 = Tx.m10;
	    /* NOBREAK */
	case (HI_SCALE | IDENTITY):
	    m00 = Tx.m00;
	    m11 = Tx.m11;
	    state = txstate;
	    return;
	case (HI_SHEAR | HI_TRANSLATE | IDENTITY):
	    m02 = Tx.m02;
	    m12 = Tx.m12;
	    /* NOBREAK */
	case (HI_SHEAR | IDENTITY):
	    m01 = Tx.m01;
	    m10 = Tx.m10;
	    m00 = m11 = 0.0;
	    state = txstate;
	    return;

	    /* ---------- Tx == TRANSLATE cases ---------- */
	case (HI_TRANSLATE | APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    T02 = Tx.m02;
	    T12 = Tx.m12;
	    m02 = T02 * m00 + T12 * m01 + m02;
	    m12 = T02 * m10 + T12 * m11 + m12;
	    return;
	case (HI_TRANSLATE | APPLY_SHEAR | APPLY_SCALE):
	    T02 = Tx.m02;
	    T12 = Tx.m12;
	    m02 = T02 * m00 + T12 * m01;
	    m12 = T02 * m10 + T12 * m11;
	    state = APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE;
	    return;
	case (HI_TRANSLATE | APPLY_SHEAR | APPLY_TRANSLATE):
	    m02 = Tx.m12 * m01 + m02;
	    m12 = Tx.m02 * m10 + m12;
	    return;
	case (HI_TRANSLATE | APPLY_SHEAR):
	    m02 = Tx.m12 * m01;
	    m12 = Tx.m02 * m10;
	    state = APPLY_SHEAR | APPLY_TRANSLATE;
	    return;
	case (HI_TRANSLATE | APPLY_SCALE | APPLY_TRANSLATE):
	    m02 = Tx.m02 * m00 + m02;
	    m12 = Tx.m12 * m11 + m12;
	    return;
	case (HI_TRANSLATE | APPLY_SCALE):
	    m02 = Tx.m02 * m00;
	    m12 = Tx.m12 * m11;
	    state = APPLY_SCALE | APPLY_TRANSLATE;
	    return;
	case (HI_TRANSLATE | APPLY_TRANSLATE):
	    m02 = Tx.m02 + m02;
	    m12 = Tx.m12 + m12;
	    return;

	    /* ---------- Tx == SCALE cases ---------- */
	case (HI_SCALE | APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_SCALE | APPLY_SHEAR | APPLY_SCALE):
	    T00 = Tx.m00; T11 = Tx.m11;
	    m00 *= T00;
	    m11 *= T11;
	    m01 *= T11;
	    m10 *= T00;
	    return;
	case (HI_SCALE | APPLY_SHEAR | APPLY_TRANSLATE):
	case (HI_SCALE | APPLY_SHEAR):
	    m01 *= Tx.m11;
	    m10 *= Tx.m00;
	    return;
	case (HI_SCALE | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_SCALE | APPLY_SCALE):
	    m00 *= Tx.m00;
	    m11 *= Tx.m11;
	    return;
	case (HI_SCALE | APPLY_TRANSLATE):
	    m00 = Tx.m00;
	    m11 = Tx.m11;
	    state = APPLY_SCALE | APPLY_TRANSLATE;
	    return;

	    /* ---------- Tx == SHEAR cases ---------- */
	case (HI_SHEAR | APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_SHEAR | APPLY_SHEAR | APPLY_SCALE):
	    T01 = Tx.m01; T10 = Tx.m10;
	    M0 = m00;
	    m00 = m01 * T10;
	    m01 = M0 * T01;
	    M0 = m10;
	    m10 = m11 * T10;
	    m11 = M0 * T01;
	    return;
	case (HI_SHEAR | APPLY_SHEAR | APPLY_TRANSLATE):
	case (HI_SHEAR | APPLY_SHEAR):
	    m00 = m01 * Tx.m10;
	    m01 = 0.0;
	    m11 = m10 * Tx.m01;
	    m10 = 0.0;
	    state = mystate ^ (APPLY_SHEAR | APPLY_SCALE);
	    return;
	case (HI_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_SHEAR | APPLY_SCALE):
	    m01 = m00 * Tx.m01;
	    m00 = 0.0;
	    m10 = m11 * Tx.m10;
	    m11 = 0.0;
	    state = mystate ^ (APPLY_SHEAR | APPLY_SCALE);
	    return;
	case (HI_SHEAR | APPLY_TRANSLATE):
	    m00 = 0.0;
	    m01 = Tx.m01;
	    m10 = Tx.m10;
	    m11 = 0.0;
	    state = APPLY_TRANSLATE | APPLY_SHEAR;
	    return;
	}
	// If Tx has more than one attribute, it is not worth optimizing
	// all of those cases...
	T00 = Tx.m00; T01 = Tx.m01; T02 = Tx.m02;
	T10 = Tx.m10; T11 = Tx.m11; T12 = Tx.m12;
	switch (mystate) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE):
	    state = mystate | txstate;
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    M0 = m00;
	    M1 = m01;
	    m00  = T00 * M0 + T10 * M1;
	    m01  = T01 * M0 + T11 * M1;
	    m02 += T02 * M0 + T12 * M1;

	    M0 = m10;
	    M1 = m11;
	    m10  = T00 * M0 + T10 * M1;
	    m11  = T01 * M0 + T11 * M1;
	    m12 += T02 * M0 + T12 * M1;
	    return;

	case (APPLY_SHEAR | APPLY_TRANSLATE):
	case (APPLY_SHEAR):
	    M0 = m01;
	    m00  = T10 * M0;
	    m01  = T11 * M0;
	    m02 += T12 * M0;

	    M0 = m10;
	    m10  = T00 * M0;
	    m11  = T01 * M0;
	    m12 += T02 * M0;
	    break;

	case (APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SCALE):
	    M0 = m00;
	    m00  = T00 * M0;
	    m01  = T01 * M0;
	    m02 += T02 * M0;

	    M0 = m11;
	    m10  = T10 * M0;
	    m11  = T11 * M0;
	    m12 += T12 * M0;
	    break;

	case (APPLY_TRANSLATE):
	    m00  = T00;
	    m01  = T01;
	    m02 += T02;

	    m10  = T10;
	    m11  = T11;
	    m12 += T12;
	    state = txstate | APPLY_TRANSLATE;
	    return;
	}
	updateState();
    }

    /**
     * Concatenates an AffineTransform Tx to this AffineTransform Cx
     * in a less commonly used way such that Tx modifies the
     * coordinate transformation relative to the absolute pixel
     * space rather than relative to the existing user space.
     * Cx is updated to perform the combined transformation.
     * Transforming a point p by the updated transform Cx' is
     * equivalent to first transforming p by the original transform
     * Cx and then transforming the result by Tx.  In other words,
     * Cx'(p) = Tx(Cx(p)).  In matrix notation, if this transform Cx
     * is represented by the matrix [this] and Tx is represented by
     * the matrix [Tx], then this method does the following:
     * <pre>
     *		[this] = [Tx] x [this]
     * </pre>
     * @param Tx The AffineTransform object to be concatenated with
     * this AffineTransform object.
     * @see #concatenate
     */
    public void preConcatenate(AffineTransform Tx) {
	double M0, M1;
	double T00, T01, T10, T11;
	double T02, T12;
	int mystate = state;
	int txstate = Tx.state;
	switch ((txstate << HI_SHIFT) | mystate) {
	case (IDENTITY):
	case (APPLY_TRANSLATE):
	case (APPLY_SCALE):
	case (APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SHEAR):
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	case (APPLY_SHEAR | APPLY_SCALE):
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    // Tx is IDENTITY...
	    return;

	case (HI_TRANSLATE | IDENTITY):
	case (HI_TRANSLATE | APPLY_SCALE):
	case (HI_TRANSLATE | APPLY_SHEAR):
	case (HI_TRANSLATE | APPLY_SHEAR | APPLY_SCALE):
	    // Tx is TRANSLATE, this has no TRANSLATE
	    m02 = Tx.m02;
	    m12 = Tx.m12;
	    state = mystate | APPLY_TRANSLATE;
	    return;

	case (HI_TRANSLATE | APPLY_TRANSLATE):
	case (HI_TRANSLATE | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_TRANSLATE | APPLY_SHEAR | APPLY_TRANSLATE):
	case (HI_TRANSLATE | APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    // Tx is TRANSLATE, this has one too
	    m02 = m02 + Tx.m02;
	    m12 = m12 + Tx.m12;
	    return;

	case (HI_SCALE | APPLY_TRANSLATE):
	case (HI_SCALE | IDENTITY):
	    // Only these two existing states need a new state
	    state = mystate | APPLY_SCALE;
	    /* NOBREAK */
	case (HI_SCALE | APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_SCALE | APPLY_SHEAR | APPLY_SCALE):
	case (HI_SCALE | APPLY_SHEAR | APPLY_TRANSLATE):
	case (HI_SCALE | APPLY_SHEAR):
	case (HI_SCALE | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_SCALE | APPLY_SCALE):
	    // Tx is SCALE, this is anything
	    T00 = Tx.m00;
	    T11 = Tx.m11;
	    if ((mystate & APPLY_SHEAR) != 0) {
		m01 = m01 * T00;
		m10 = m10 * T11;
		if ((mystate & APPLY_SCALE) != 0) {
		    m00 = m00 * T00;
		    m11 = m11 * T11;
		}
	    } else {
		m00 = m00 * T00;
		m11 = m11 * T11;
	    }
	    if ((mystate & APPLY_TRANSLATE) != 0) {
		m02 = m02 * T00;
		m12 = m12 * T11;
	    }
	    return;
	case (HI_SHEAR | APPLY_SHEAR | APPLY_TRANSLATE):
	case (HI_SHEAR | APPLY_SHEAR):
	    mystate = mystate | APPLY_SCALE;
	    /* NOBREAK */
	case (HI_SHEAR | APPLY_TRANSLATE):
	case (HI_SHEAR | IDENTITY):
	case (HI_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_SHEAR | APPLY_SCALE):
	    state = mystate ^ APPLY_SHEAR;
	    /* NOBREAK */
	case (HI_SHEAR | APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (HI_SHEAR | APPLY_SHEAR | APPLY_SCALE):
	    // Tx is SHEAR, this is anything
	    T01 = Tx.m01;
	    T10 = Tx.m10;

	    M0 = m00;
	    m00 = m10 * T01;
	    m10 = M0 * T10;

	    M0 = m01;
	    m01 = m11 * T01;
	    m11 = M0 * T10;

	    M0 = m02;
	    m02 = m12 * T01;
	    m12 = M0 * T10;
	    return;
	}
	// If Tx has more than one attribute, it is not worth optimizing
	// all of those cases...
	T00 = Tx.m00; T01 = Tx.m01; T02 = Tx.m02;
	T10 = Tx.m10; T11 = Tx.m11; T12 = Tx.m12;
	switch (mystate) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    M0 = m02;
	    M1 = m12;
	    T02 += M0 * T00 + M1 * T01;
	    T12 += M0 * T10 + M1 * T11;

	    /* NOBREAK */
	case (APPLY_SHEAR | APPLY_SCALE):
	    m02 = T02;
	    m12 = T12;

	    M0 = m00;
	    M1 = m10;
	    m00 = M0 * T00 + M1 * T01;
	    m10 = M0 * T10 + M1 * T11;

	    M0 = m01;
	    M1 = m11;
	    m01 = M0 * T00 + M1 * T01;
	    m11 = M0 * T10 + M1 * T11;
	    break;

	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    M0 = m02;
	    M1 = m12;
	    T02 += M0 * T00 + M1 * T01;
	    T12 += M0 * T10 + M1 * T11;

	    /* NOBREAK */
	case (APPLY_SHEAR):
	    m02 = T02;
	    m12 = T12;

	    M0 = m10;
	    m00 = M0 * T01;
	    m10 = M0 * T11;

	    M0 = m01;
	    m01 = M0 * T00;
	    m11 = M0 * T10;
	    break;

	case (APPLY_SCALE | APPLY_TRANSLATE):
	    M0 = m02;
	    M1 = m12;
	    T02 += M0 * T00 + M1 * T01;
	    T12 += M0 * T10 + M1 * T11;

	    /* NOBREAK */
	case (APPLY_SCALE):
	    m02 = T02;
	    m12 = T12;

	    M0 = m00;
	    m00 = M0 * T00;
	    m10 = M0 * T10;

	    M0 = m11;
	    m01 = M0 * T01;
	    m11 = M0 * T11;
	    break;

	case (APPLY_TRANSLATE):
	    M0 = m02;
	    M1 = m12;
	    T02 += M0 * T00 + M1 * T01;
	    T12 += M0 * T10 + M1 * T11;

	    /* NOBREAK */
	case (IDENTITY):
	    m02 = T02;
	    m12 = T12;

	    m00 = T00;
	    m10 = T10;

	    m01 = T01;
	    m11 = T11;

	    state = mystate | txstate;
	    return;
	}
	updateState();
    }

    /**
     * Returns an AffineTransform object representing the inverse
     * transformation.
     * @exception NoninvertibleTransformException  if the matrix cannot be
     *                                         inverted.
     */
    public AffineTransform createInverse()
	throws NoninvertibleTransformException
    {
	double det;
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    det = m00 * m11 - m01 * m10;
	    if (Math.abs(det) <= Double.MIN_VALUE) {
		throw new NoninvertibleTransformException("Determinant is "+
							  det);
	    }
	    return new AffineTransform( m11 / det, -m10 / det,
				       -m01 / det,  m00 / det,
				       (m01 * m12 - m11 * m02) / det,
				       (m10 * m02 - m00 * m12) / det,
				       (APPLY_SHEAR |
					APPLY_SCALE |
					APPLY_TRANSLATE));
	case (APPLY_SHEAR | APPLY_SCALE):
	    det = m00 * m11 - m01 * m10;
	    if (Math.abs(det) <= Double.MIN_VALUE) {
		throw new NoninvertibleTransformException("Determinant is "+
							  det);
	    }
	    return new AffineTransform( m11 / det, -m10 / det,
				       -m01 / det,  m00 / det,
				        0.0,        0.0,
				       (APPLY_SHEAR | APPLY_SCALE));
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    if (m01 == 0.0 || m10 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    return new AffineTransform( 0.0,        1.0 / m01,
				        1.0 / m10,  0.0,
				       -m12 / m10, -m02 / m01,
				       (APPLY_SHEAR | APPLY_TRANSLATE));
	case (APPLY_SHEAR):
	    if (m01 == 0.0 || m10 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    return new AffineTransform(0.0,       1.0 / m01,
				       1.0 / m10, 0.0,
				       0.0,       0.0,
				       (APPLY_SHEAR));
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    if (m00 == 0.0 || m11 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    return new AffineTransform( 1.0 / m00,  0.0,
				        0.0,        1.0 / m11,
				       -m02 / m00, -m12 / m11,
				       (APPLY_SCALE | APPLY_TRANSLATE));
	case (APPLY_SCALE):
	    if (m00 == 0.0 || m11 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    return new AffineTransform(1.0 / m00, 0.0,
				       0.0,       1.0 / m11,
				       0.0,       0.0,
				       (APPLY_SCALE));
	case (APPLY_TRANSLATE):
	    return new AffineTransform( 1.0,  0.0,
				        0.0,  1.0,
				       -m02, -m12,
				       (APPLY_TRANSLATE));
	case (IDENTITY):
	    return new AffineTransform();
	}

	/* NOTREACHED */
    }

    /**
     * Transforms the specified ptSrc and stores the result in ptDst.
     * If ptDst is null, a new Point2D object will be allocated before
     * storing. In either case, ptDst containing the transformed point
     * is returned for convenience.
     * Note that ptSrc and ptDst can the same. In this case, the input
     * point will be overwritten with the transformed point.
     */
    public Point2D transform(Point2D ptSrc, Point2D ptDst) {
	if (ptDst == null) {
	    if (ptSrc instanceof Point2D.Double) {
		ptDst = new Point2D.Double();
	    } else {
		ptDst = new Point2D.Float();
	    }
	}
	// Copy source coords into local variables in case src == dst
	double x = ptSrc.getX();
	double y = ptSrc.getY();
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    ptDst.setLocation(x * m00 + y * m01 + m02,
			      x * m10 + y * m11 + m12);
	    return ptDst;
	case (APPLY_SHEAR | APPLY_SCALE):
	    ptDst.setLocation(x * m00 + y * m01, x * m10 + y * m11);
	    return ptDst;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    ptDst.setLocation(y * m01 + m02, x * m10 + m12);
	    return ptDst;
	case (APPLY_SHEAR):
	    ptDst.setLocation(y * m01, x * m10);
	    return ptDst;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    ptDst.setLocation(x * m00 + m02, y * m11 + m12);
	    return ptDst;
	case (APPLY_SCALE):
	    ptDst.setLocation(x * m00, y * m11);
	    return ptDst;
	case (APPLY_TRANSLATE):
	    ptDst.setLocation(x + m02, y + m12);
	    return ptDst;
	case (IDENTITY):
	    ptDst.setLocation(x, y);
	    return ptDst;
	}

	/* NOTREACHED */
    }

    /**
     * Transforms an array of point objects by this transform.
     * @param ptSrc The array containing the source point objects.
     * @param ptDst The array where the transform point objects are returned.
     * @param srcOff The offset to the first point object to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point object is stored in the destination array.
     * @param numPts The number of point objects to be transformed.
     */
    public void transform(Point2D[] ptSrc, int srcOff,
			  Point2D[] ptDst, int dstOff,
			  int numPts) {
	int state = this.state;
        while (--numPts >= 0) {
            // Copy source coords into local variables in case src == dst
	    Point2D src = ptSrc[srcOff++];
            double x = src.getX();
            double y = src.getY();
	    Point2D dst = ptDst[dstOff++];
	    if (dst == null) {
		if (src instanceof Point2D.Double) {
		    dst = new Point2D.Double();
		} else {
		    dst = new Point2D.Float();
		}
		ptDst[dstOff - 1] = dst;
	    }
	    switch (state) {
	    default:
		stateError();
		/* NOTREACHED */
	    case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
		dst.setLocation(x * m00 + y * m01 + m02,
				x * m10 + y * m11 + m12);
		break;
	    case (APPLY_SHEAR | APPLY_SCALE):
		dst.setLocation(x * m00 + y * m01, x * m10 + y * m11);
		break;
	    case (APPLY_SHEAR | APPLY_TRANSLATE):
		dst.setLocation(y * m01 + m02, x * m10 + m12);
		break;
	    case (APPLY_SHEAR):
		dst.setLocation(y * m01, x * m10);
		break;
	    case (APPLY_SCALE | APPLY_TRANSLATE):
		dst.setLocation(x * m00 + m02, y * m11 + m12);
		break;
	    case (APPLY_SCALE):
		dst.setLocation(x * m00, y * m11);
		break;
	    case (APPLY_TRANSLATE):
		dst.setLocation(x + m02, y + m12);
		break;
	    case (IDENTITY):
		dst.setLocation(x, y);
		break;
	    }
	}

	/* NOTREACHED */
    }

    /**
     * Transforms an array of floating point coordinates by this transform.
     * @param ptSrc The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param ptDst The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of points to be transformed.
     */
    public void transform(float[] srcPts, int srcOff,
			  float[] dstPts, int dstOff,
			  int numPts) {
	double M00, M01, M02, M10, M11, M12;	// For caching
	if (dstPts == srcPts &&
	    dstOff > srcOff && dstOff < srcOff + numPts * 2)
	{
	    // If the arrays overlap partially with the destination higher
	    // than the source and we transform the coordinates normally
	    // we would overwrite some of the later source coordinates
	    // with results of previous transformations.
	    // To get around this we use arraycopy to copy the points
	    // to their final destination with correct overwrite
	    // handling and then transform them in place in the new
	    // safer location.
	    System.arraycopy(srcPts, srcOff, dstPts, dstOff, numPts * 2);
	    // srcPts = dstPts;		// They are known to be equal.
	    srcOff = dstOff;
	}
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M01 = m01; M02 = m02;
	    M10 = m10; M11 = m11; M12 = m12;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = (float) (M00 * x + M01 * y + M02);
		dstPts[dstOff++] = (float) (M10 * x + M11 * y + M12);
	    }
	    return;
	case (APPLY_SHEAR | APPLY_SCALE):
	    M00 = m00; M01 = m01;
	    M10 = m10; M11 = m11;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = (float) (M00 * x + M01 * y);
		dstPts[dstOff++] = (float) (M10 * x + M11 * y);
	    }
	    return;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    M01 = m01; M02 = m02;
	    M10 = m10; M12 = m12;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = (float) (M01 * srcPts[srcOff++] + M02);
		dstPts[dstOff++] = (float) (M10 * x + M12);
	    }
	    return;
	case (APPLY_SHEAR):
	    M01 = m01; M10 = m10;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = (float) (M01 * srcPts[srcOff++]);
		dstPts[dstOff++] = (float) (M10 * x);
	    }
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M02 = m02;
	    M11 = m11; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = (float) (M00 * srcPts[srcOff++] + M02);
		dstPts[dstOff++] = (float) (M11 * srcPts[srcOff++] + M12);
	    }
	    return;
	case (APPLY_SCALE):
	    M00 = m00; M11 = m11;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = (float) (M00 * srcPts[srcOff++]);
		dstPts[dstOff++] = (float) (M11 * srcPts[srcOff++]);
	    }
	    return;
	case (APPLY_TRANSLATE):
	    M02 = m02; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = (float) (srcPts[srcOff++] + M02);
		dstPts[dstOff++] = (float) (srcPts[srcOff++] + M12);
	    }
	    return;
	case (IDENTITY):
	    if (srcPts != dstPts || srcOff != dstOff) {
		System.arraycopy(srcPts, srcOff, dstPts, dstOff,
				 numPts * 2);
	    }
	    return;
	}

	/* NOTREACHED */
    }

    /**
     * Transforms an array of double precision coordinates by this transform.
     * @param srcPts The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param dstPts The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of point objects to be transformed.
     */
    public void transform(double[] srcPts, int srcOff,
			  double[] dstPts, int dstOff,
			  int numPts) {
	double M00, M01, M02, M10, M11, M12;	// For caching
	if (dstPts == srcPts &&
	    dstOff > srcOff && dstOff < srcOff + numPts * 2)
	{
	    // If the arrays overlap partially with the destination higher
	    // than the source and we transform the coordinates normally
	    // we would overwrite some of the later source coordinates
	    // with results of previous transformations.
	    // To get around this we use arraycopy to copy the points
	    // to their final destination with correct overwrite
	    // handling and then transform them in place in the new
	    // safer location.
	    System.arraycopy(srcPts, srcOff, dstPts, dstOff, numPts * 2);
	    // srcPts = dstPts;		// They are known to be equal.
	    srcOff = dstOff;
	}
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M01 = m01; M02 = m02;
	    M10 = m10; M11 = m11; M12 = m12;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = M00 * x + M01 * y + M02;
		dstPts[dstOff++] = M10 * x + M11 * y + M12;
	    }
	    return;
	case (APPLY_SHEAR | APPLY_SCALE):
	    M00 = m00; M01 = m01;
	    M10 = m10; M11 = m11;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = M00 * x + M01 * y;
		dstPts[dstOff++] = M10 * x + M11 * y;
	    }
	    return;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    M01 = m01; M02 = m02;
	    M10 = m10; M12 = m12;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = M01 * srcPts[srcOff++] + M02;
		dstPts[dstOff++] = M10 * x + M12;
	    }
	    return;
	case (APPLY_SHEAR):
	    M01 = m01; M10 = m10;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = M01 * srcPts[srcOff++];
		dstPts[dstOff++] = M10 * x;
	    }
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M02 = m02;
	    M11 = m11; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = M00 * srcPts[srcOff++] + M02;
		dstPts[dstOff++] = M11 * srcPts[srcOff++] + M12;
	    }
	    return;
	case (APPLY_SCALE):
	    M00 = m00; M11 = m11;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = M00 * srcPts[srcOff++];
		dstPts[dstOff++] = M11 * srcPts[srcOff++];
	    }
	    return;
	case (APPLY_TRANSLATE):
	    M02 = m02; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = srcPts[srcOff++] + M02;
		dstPts[dstOff++] = srcPts[srcOff++] + M12;
	    }
	    return;
	case (IDENTITY):
	    if (srcPts != dstPts || srcOff != dstOff) {
		System.arraycopy(srcPts, srcOff, dstPts, dstOff,
				 numPts * 2);
	    }
	    return;
	}

	/* NOTREACHED */
    }

    /**
     * Transforms an array of floating point coordinates by this transform,
     * storing the results into an array of doubles.
     * @param ptSrc The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param ptDst The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of points to be transformed.
     */
    public void transform(float[] srcPts, int srcOff,
			  double[] dstPts, int dstOff,
			  int numPts) {
	double M00, M01, M02, M10, M11, M12;	// For caching
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M01 = m01; M02 = m02;
	    M10 = m10; M11 = m11; M12 = m12;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = M00 * x + M01 * y + M02;
		dstPts[dstOff++] = M10 * x + M11 * y + M12;
	    }
	    return;
	case (APPLY_SHEAR | APPLY_SCALE):
	    M00 = m00; M01 = m01;
	    M10 = m10; M11 = m11;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = M00 * x + M01 * y;
		dstPts[dstOff++] = M10 * x + M11 * y;
	    }
	    return;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    M01 = m01; M02 = m02;
	    M10 = m10; M12 = m12;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = M01 * srcPts[srcOff++] + M02;
		dstPts[dstOff++] = M10 * x + M12;
	    }
	    return;
	case (APPLY_SHEAR):
	    M01 = m01; M10 = m10;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = M01 * srcPts[srcOff++];
		dstPts[dstOff++] = M10 * x;
	    }
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M02 = m02;
	    M11 = m11; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = M00 * srcPts[srcOff++] + M02;
		dstPts[dstOff++] = M11 * srcPts[srcOff++] + M12;
	    }
	    return;
	case (APPLY_SCALE):
	    M00 = m00; M11 = m11;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = M00 * srcPts[srcOff++];
		dstPts[dstOff++] = M11 * srcPts[srcOff++];
	    }
	    return;
	case (APPLY_TRANSLATE):
	    M02 = m02; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = srcPts[srcOff++] + M02;
		dstPts[dstOff++] = srcPts[srcOff++] + M12;
	    }
	    return;
	case (IDENTITY):
	    while (--numPts >= 0) {
		dstPts[dstOff++] = srcPts[srcOff++];
		dstPts[dstOff++] = srcPts[srcOff++];
	    }
	    return;
	}

	/* NOTREACHED */
    }

    /**
     * Transforms an array of double precision coordinates by this transform,
     * storing the results into an array of floats.
     * @param srcPts The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param dstPts The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of point objects to be transformed.
     */
    public void transform(double[] srcPts, int srcOff,
			  float[] dstPts, int dstOff,
			  int numPts) {
	double M00, M01, M02, M10, M11, M12;	// For caching
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M01 = m01; M02 = m02;
	    M10 = m10; M11 = m11; M12 = m12;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = (float) (M00 * x + M01 * y + M02);
		dstPts[dstOff++] = (float) (M10 * x + M11 * y + M12);
	    }
	    return;
	case (APPLY_SHEAR | APPLY_SCALE):
	    M00 = m00; M01 = m01;
	    M10 = m10; M11 = m11;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = (float) (M00 * x + M01 * y);
		dstPts[dstOff++] = (float) (M10 * x + M11 * y);
	    }
	    return;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    M01 = m01; M02 = m02;
	    M10 = m10; M12 = m12;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = (float) (M01 * srcPts[srcOff++] + M02);
		dstPts[dstOff++] = (float) (M10 * x + M12);
	    }
	    return;
	case (APPLY_SHEAR):
	    M01 = m01; M10 = m10;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = (float) (M01 * srcPts[srcOff++]);
		dstPts[dstOff++] = (float) (M10 * x);
	    }
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M02 = m02;
	    M11 = m11; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = (float) (M00 * srcPts[srcOff++] + M02);
		dstPts[dstOff++] = (float) (M11 * srcPts[srcOff++] + M12);
	    }
	    return;
	case (APPLY_SCALE):
	    M00 = m00; M11 = m11;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = (float) (M00 * srcPts[srcOff++]);
		dstPts[dstOff++] = (float) (M11 * srcPts[srcOff++]);
	    }
	    return;
	case (APPLY_TRANSLATE):
	    M02 = m02; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = (float) (srcPts[srcOff++] + M02);
		dstPts[dstOff++] = (float) (srcPts[srcOff++] + M12);
	    }
	    return;
	case (IDENTITY):
	    while (--numPts >= 0) {
		dstPts[dstOff++] = (float) (srcPts[srcOff++]);
		dstPts[dstOff++] = (float) (srcPts[srcOff++]);
	    }
	    return;
	}

	/* NOTREACHED */
    }

    /**
     * Inverse transforms the specified ptSrc and stores the result in ptDst.
     * If ptDst is null, a new Point2D object will be allocated before
     * storing. In either case, ptDst containing the transformed point
     * is returned for convenience.
     * Note that ptSrc and ptDst can the same. In this case, the input
     * point will be overwritten with the transformed point.
     * @param ptSrc The point to be inverse transformed.
     * @param ptDst The resulting transformed point.
     * @exception NoninvertibleTransformException  if the matrix cannot be
     *                                         inverted.
     */
    public Point2D inverseTransform(Point2D ptSrc, Point2D ptDst)
	throws NoninvertibleTransformException
    {
	if (ptDst == null) {
	    if (ptSrc instanceof Point2D.Double) {
		ptDst = new Point2D.Double();
	    } else {
		ptDst = new Point2D.Float();
	    }
	}
	// Copy source coords into local variables in case src == dst
	double x = ptSrc.getX();
	double y = ptSrc.getY();
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    x -= m02;
	    y -= m12;
	    /* NOBREAK */
	case (APPLY_SHEAR | APPLY_SCALE):
	    double det = m00 * m11 - m01 * m10;
	    if (Math.abs(det) <= Double.MIN_VALUE) {
		throw new NoninvertibleTransformException("Determinant is "+
							  det);
	    }
	    ptDst.setLocation((x * m11 - y * m01) / det,
			      (y * m00 - x * m10) / det);
	    return ptDst;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    x -= m02;
	    y -= m12;
	    /* NOBREAK */
	case (APPLY_SHEAR):
	    if (m01 == 0.0 || m10 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    ptDst.setLocation(y / m10, x / m01);
	    return ptDst;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    x -= m02;
	    y -= m12;
	    /* NOBREAK */
	case (APPLY_SCALE):
	    if (m00 == 0.0 || m11 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    ptDst.setLocation(x / m00, y / m11);
	    return ptDst;
	case (APPLY_TRANSLATE):
	    ptDst.setLocation(x - m02, y - m12);
	    return ptDst;
	case (IDENTITY):
	    ptDst.setLocation(x, y);
	    return ptDst;
	}

	/* NOTREACHED */
    }

    /**
     * Inverse transforms an array of double precision coordinates by
     * this transform.  
     * @param srcPts The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param dstPts The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of point objects to be transformed.
     * @exception NoninvertibleTransformException  if the matrix cannot be
     *                                         inverted.
     */
    public void inverseTransform(double[] srcPts, int srcOff,
                                 double[] dstPts, int dstOff,
                                 int numPts)
	throws NoninvertibleTransformException
    {
	double M00, M01, M02, M10, M11, M12;	// For caching
	double det;
	if (dstPts == srcPts &&
	    dstOff > srcOff && dstOff < srcOff + numPts * 2)
	{
	    // If the arrays overlap partially with the destination higher
	    // than the source and we transform the coordinates normally
	    // we would overwrite some of the later source coordinates
	    // with results of previous transformations.
	    // To get around this we use arraycopy to copy the points
	    // to their final destination with correct overwrite
	    // handling and then transform them in place in the new
	    // safer location.
	    System.arraycopy(srcPts, srcOff, dstPts, dstOff, numPts * 2);
	    // srcPts = dstPts;		// They are known to be equal.
	    srcOff = dstOff;
	}
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M01 = m01; M02 = m02;
	    M10 = m10; M11 = m11; M12 = m12;
	    det = M00 * M11 - M01 * M10;
	    if (Math.abs(det) <= Double.MIN_VALUE) {
		throw new NoninvertibleTransformException("Determinant is "+
							  det);
	    }
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++] - M02;
		double y = srcPts[srcOff++] - M12;
		dstPts[dstOff++] = (x * M11 - y * M01) / det;
		dstPts[dstOff++] = (y * M00 - x * M10) / det;
	    }
	    return;
	case (APPLY_SHEAR | APPLY_SCALE):
	    M00 = m00; M01 = m01;
	    M10 = m10; M11 = m11;
	    det = M00 * M11 - M01 * M10;
	    if (Math.abs(det) <= Double.MIN_VALUE) {
		throw new NoninvertibleTransformException("Determinant is "+
							  det);
	    }
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = (x * M11 - y * M01) / det;
		dstPts[dstOff++] = (y * M00 - x * M10) / det;
	    }
	    return;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	    M01 = m01; M02 = m02;
	    M10 = m10; M12 = m12;
	    if (M01 == 0.0 || M10 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++] - M02;
		dstPts[dstOff++] = (srcPts[srcOff++] - M12) / M01;
		dstPts[dstOff++] = x / M10;
	    }
	    return;
	case (APPLY_SHEAR):
	    M01 = m01; M10 = m10;
	    if (M01 == 0.0 || M10 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = srcPts[srcOff++] / M01;
		dstPts[dstOff++] = x / M10;
	    }
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	    M00 = m00; M02 = m02;
	    M11 = m11; M12 = m12;
	    if (M00 == 0.0 || M11 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    while (--numPts >= 0) {
		dstPts[dstOff++] = (srcPts[srcOff++] - M02) / M00;
		dstPts[dstOff++] = (srcPts[srcOff++] - M12) / M11;
	    }
	    return;
	case (APPLY_SCALE):
	    M00 = m00; M11 = m11;
	    if (M00 == 0.0 || M11 == 0.0) {
		throw new NoninvertibleTransformException("Determinant is 0");
	    }
	    while (--numPts >= 0) {
		dstPts[dstOff++] = srcPts[srcOff++] / M00;
		dstPts[dstOff++] = srcPts[srcOff++] / M11;
	    }
	    return;
	case (APPLY_TRANSLATE):
	    M02 = m02; M12 = m12;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = srcPts[srcOff++] - M02;
		dstPts[dstOff++] = srcPts[srcOff++] - M12;
	    }
	    return;
	case (IDENTITY):
	    if (srcPts != dstPts || srcOff != dstOff) {
		System.arraycopy(srcPts, srcOff, dstPts, dstOff,
				 numPts * 2);
	    }
	    return;
	}

	/* NOTREACHED */
    }

    /**
     * Transforms the relative distance vector specified by ptSrc and
     * stores the result in ptDst.
     * A relative distance vector is transformed without applying the
     * translation components of the affine transformation matrix
     * using the following equations:
     * <pre>
     *	[ x']   [  m00  m01 (m02) ] [ x ]   [ m00x + m01y ]
     *	[ y'] = [  m10  m11 (m12) ] [ y ] = [ m10x + m11y ]
     *	[ 1 ]   [   0    0    1   ] [ 1 ]   [      1      ]
     * </pre>
     * If ptDst is null, a new Point2D object will be allocated before
     * storing. In either case, ptDst containing the transformed point
     * is returned for convenience.
     * Note that ptSrc and ptDst can the same. In this case, the input
     * point will be overwritten with the transformed point.
     * @param ptSrc The distance vector to be inverse transformed.
     * @param ptDst The resulting transformed distance vector.
     */
    public Point2D deltaTransform(Point2D ptSrc, Point2D ptDst) {
	if (ptDst == null) {
	    if (ptSrc instanceof Point2D.Double) {
		ptDst = new Point2D.Double();
	    } else {
		ptDst = new Point2D.Float();
	    }
	}
	// Copy source coords into local variables in case src == dst
	double x = ptSrc.getX();
	double y = ptSrc.getY();
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SHEAR | APPLY_SCALE):
	    ptDst.setLocation(x * m00 + y * m01, x * m10 + y * m11);
	    return ptDst;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	case (APPLY_SHEAR):
	    ptDst.setLocation(y * m01, x * m10);
	    return ptDst;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SCALE):
	    ptDst.setLocation(x * m00, y * m11);
	    return ptDst;
	case (APPLY_TRANSLATE):
	case (IDENTITY):
	    ptDst.setLocation(x, y);
	    return ptDst;
	}

	/* NOTREACHED */
    }

    /**
     * Transforms an array of relative distance vectors by this
     * transform.
     * A relative distance vector is transformed without applying the
     * translation components of the affine transformation matrix
     * using the following equations:
     * <pre>
     *	[ x']   [  m00  m01 (m02) ] [ x ]   [ m00x + m01y ]
     *	[ y'] = [  m10  m11 (m12) ] [ y ] = [ m10x + m11y ]
     *	[ 1 ]   [   0    0    1   ] [ 1 ]   [      1      ]
     * </pre>
     * @param srcPts The array containing the source distance vectors.
     * Each vector is stored as a pair of relative x,y coordinates.
     * @param dstPts The array where the transformed distance vectors are
     * returned.  Each vector is stored as a pair of relative x,y coordinates.
     * @param srcOff The offset to the first vector to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * vector is stored in the destination array.
     * @param numPts The number of vector coordinate pairs to be transformed.
     */
    public void deltaTransform(double[] srcPts, int srcOff,
			       double[] dstPts, int dstOff,
			       int numPts) {
	double M00, M01, M10, M11;	// For caching
	if (dstPts == srcPts &&
	    dstOff > srcOff && dstOff < srcOff + numPts * 2)
	{
	    // If the arrays overlap partially with the destination higher
	    // than the source and we transform the coordinates normally
	    // we would overwrite some of the later source coordinates
	    // with results of previous transformations.
	    // To get around this we use arraycopy to copy the points
	    // to their final destination with correct overwrite
	    // handling and then transform them in place in the new
	    // safer location.
	    System.arraycopy(srcPts, srcOff, dstPts, dstOff, numPts * 2);
	    // srcPts = dstPts;		// They are known to be equal.
	    srcOff = dstOff;
	}
	switch (state) {
	default:
	    stateError();
	    /* NOTREACHED */
	case (APPLY_SHEAR | APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SHEAR | APPLY_SCALE):
	    M00 = m00; M01 = m01;
	    M10 = m10; M11 = m11;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		dstPts[dstOff++] = x * M00 + y * M01;
		dstPts[dstOff++] = x * M10 + y * M11;
	    }
	    return;
	case (APPLY_SHEAR | APPLY_TRANSLATE):
	case (APPLY_SHEAR):
	    M01 = m01; M10 = m10;
	    while (--numPts >= 0) {
		double x = srcPts[srcOff++];
		dstPts[dstOff++] = srcPts[srcOff++] * M01;
		dstPts[dstOff++] = x * M10;
	    }
	    return;
	case (APPLY_SCALE | APPLY_TRANSLATE):
	case (APPLY_SCALE):
	    M00 = m00; M11 = m11;
	    while (--numPts >= 0) {
		dstPts[dstOff++] = srcPts[srcOff++] * M00;
		dstPts[dstOff++] = srcPts[srcOff++] * M11;
	    }
	    return;
	case (APPLY_TRANSLATE):
	case (IDENTITY):
	    if (srcPts != dstPts || srcOff != dstOff) {
		System.arraycopy(srcPts, srcOff, dstPts, dstOff,
				 numPts * 2);
	    }
	    return;
	}

	/* NOTREACHED */
    }

    /**
     * Transforms a shape object by this transform.
     */
    public Shape createTransformedShape(Shape pSrc) {
        if (pSrc == null) {
            return null;
        }
        
        if (pSrc instanceof GeneralPath) {
            return ((GeneralPath)pSrc).createTransformedShape(this);
        } else {
            PathIterator pi = pSrc.getPathIterator(this);
            GeneralPath gp = new GeneralPath(pi.getWindingRule());
	    gp.append(pi, false);
            return gp;
        }

	/* NOTREACHED */
    }

    // Round values to sane precision for printing
    // Note that Math.sin(Math.PI) has an error of about 10^-16
    private static double _matround(double matval) {
	return Math.rint(matval * 1E15) / 1E15;
    }

    /**
     * Returns a String that represents the value of this Object.
     */
    public String toString() {
	return ("AffineTransform[["
		+ _matround(m00) + ", "
		+ _matround(m01) + ", "
		+ _matround(m02) + "], ["
		+ _matround(m10) + ", "
		+ _matround(m11) + ", "
		+ _matround(m12) + "]]");
    }

    /**
     * Returns the boolean true value if this AffineTransform is an
     * identity transform. Returns false otherwise.
     */
    public boolean isIdentity() {
        return (state == IDENTITY);
    }

    /**
     * Returns a copy of this AffineTransform object.
     */
    public Object clone() {
	try {
	    return super.clone();
	} catch (CloneNotSupportedException e) { 
	    // this shouldn't happen, since we are Cloneable
	    throw new InternalError();
	}
    }

    public boolean equals(Object obj) {
        if (!(obj instanceof AffineTransform)) {
            return false;
        }

        AffineTransform a = (AffineTransform)obj;

	return ((m00 == a.m00) && (m01 == a.m01) && (m02 == a.m02) &&
		(m10 == a.m10) && (m11 == a.m11) && (m12 == a.m12));
    }

    /* Serialization support.  A readObject method is neccessary because
     * the state field is part of the implementation of this particular
     * AffineTransform and not part of the public specification.  The
     * state variable's value needs to be recalculated on the fly by the
     * readObject method as it is in the 6-argument matrix constructor.
     */

    private void writeObject(java.io.ObjectOutputStream s)
	throws java.lang.ClassNotFoundException, java.io.IOException
    {
	s.defaultWriteObject();
    }

    private void readObject(java.io.ObjectInputStream s)
	throws java.lang.ClassNotFoundException, java.io.IOException
    {
	s.defaultReadObject();
	updateState();
    }
}
