/*
 * @(#)BilinearAffineTransformOp.java	1.16 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.awt.GraphicsEnvironment;
import java.awt.geom.AffineTransform;


/**
 * This class uses an affine transformation with bilinear interpolation
 * to transform an image or Raster.
 * <p>
 * Note that the following constraints have to be met:
 * <li>Source and Destination must be different.
 * <li>For Rasters, the number of Channels in the Source must be equal
 * to the number of Channels in the Destination.
 * @version 10 Feb 1997
 */
public class BilinearAffineTransformOp extends AffineTransformOp {

    /**
     * Constructs a BilinearAffineTransformOp given an affine
     * transform.  
     */
    public BilinearAffineTransformOp(AffineTransform xform) {
        super (xform);
    }

    /**
     * Transforms the BufferedImage using the stored affine transform.
     * If the color models for the two images do not match, a color
     * conversion into the destination color model will be performed.
     * If the destination image is null,
     * a BufferedImage will be created with the source ColorModel.
     * The IllegalArgumentException may be thrown if the source is
     * the same as the destination.
     */
    public BufferedImage filter (BufferedImage src, BufferedImage dst) {
        if (src == null) {
            throw new NullPointerException("src image is null");
        }
            
        if (dst == null) {
            dst = createCompatibleDestImage(src, null);
        }
        GraphicsEnvironment ge =
            GraphicsEnvironment.getLocalGraphicsEnvironment();
        ImagingLib imlib = ge.getImagingLib();

        if (imlib.filter(this, src, dst) == null) {
            throw new ImagingOpException ("Unable to transform src image");
        }
        return dst;
    }

    /**
     * Transforms the Raster using the stored affine transform.
     * This operation will perform the transform channel by channel.
     * If the destination Raster is null, a new Raster will be created.
     * The IllegalArgumentException may be thrown if the source is
     * the same as the destination or if the number of channels in
     * the source is not equal to the number of channels in the
     * destination.
     */

    public WritableRaster filter (Raster src, WritableRaster dst) {
        if (src == null) {
            throw new NullPointerException("src image is null");
        }
        if (dst == null) {
            dst = createCompatibleDestRaster(src);
        }
        GraphicsEnvironment ge =
            GraphicsEnvironment.getLocalGraphicsEnvironment();
        ImagingLib imlib = ge.getImagingLib();

        if (imlib.filter(this, src, dst) == null) {
            throw new ImagingOpException ("Unable to transform src image");
        }
        return dst;
    }

}
