/*
 * @(#)ColorConvertOp.java	1.15 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/**********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.image;

import java.awt.Point;
import java.awt.color.*;
import sun.awt.color.ICC_Transform;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Point2D;

/**
 * This class performs a pixel-by-pixel color conversion of the data in
 * the source image.  The resulting color values are scaled to the precision
 * of the destination image data type.  Color conversion can be specified
 * via an array of ColorSpace objects or an array of ICC_Profile objects.
 * <p>
 * If the source is a BufferedImage with premultiplied alpha, the
 * color components are divided by the alpha component before color conversion.
 * If the destination is a BufferedImage with premultiplied alpha, the
 * color components are multiplied by the alpha component after conversion.
 * Rasters are treated as having no alpha channel, i.e. all channels are
 * color channels.
 * <p>
 * Note Source and Destination may be the same object.
 * <p>
 */
public class ColorConvertOp implements BufferedImageOp, RasterOp {
    ICC_Profile[]    profileList;
    ICC_Transform    thisTransform = null, thisRasterTransform = null;
    ICC_Profile      thisSrcProfile = null, thisDestProfile = null;

    /**
     * Constructs a new ColorConvertOp which will convert
     * from a source color space to a destination color space.
     * This Op can be used only with BufferedImages, and will convert
     * directly from the ColorSpace of the source image to that of the
     * destination.  The destination cannot be specified as null.
     */
    public ColorConvertOp ()
    {
        profileList = new ICC_Profile [0];    /* 0 length list */
    }

    /**
     * Constructs a new ColorConvertOp from a color space identified by
     * an integer constant representing one of the specific predefined
     * color spaces from the ColorSpace class.  This Op can be used only
     * with BufferedImages, and is primarily useful when the filter method
     * is invoked with a destination argument of null.  In that case,
     * the color space defines the destination color space for the
     * destination created by the filter method.  Otherwise, the color space
     * defines an intermediate space to which the source is converted before
     * being converted to the destination space.
     */
    public ColorConvertOp (int colorspace)
    {
        profileList = new ICC_Profile [1];    /* 1 profile in the list */

        profileList [0] = ICC_Profile.getInstance (colorspace);
    }


    /**
     * Constructs a new ColorConvertOp from a ColorSpace object.  This
     * Op can be used only with BufferedImages, and is primarily useful
     * when the filter method is invoked with a destination argument of null.
     * In that case, the ColorSpace defines the destination color space
     * for the destination created by the filter method.  Otherwise, the
     * ColorSpace defines an intermediate space to which the source is
     * converted before being converted to the destination space.
     */
    public ColorConvertOp (ColorSpace cspace)
    {
        if (cspace instanceof ICC_ColorSpace) {
            profileList = new ICC_Profile [1];    /* 1 profile in the list */

            profileList [0] = ((ICC_ColorSpace) cspace).getProfile();
        }
        else {
            throw new IllegalArgumentException(
                "ColorSpace is not an ICC_ColorSpace");
        }
    }


    /**
     * Constructs a new ColorConvertOp from two color spaces identified by
     * integer constants representing specific predefined
     * color spaces from the ColorSpace class.  This Op is primarily
     * useful for calling the filter method on Rasters, in which case
     * the two color spaces define the operation to be performed on the
     * Rasters.  In that case, the number of bands in the source Raster
     * must match the number of components in srcColorspace, and
     * the number of bands in the destination Raster must match the number
     * of components in dstColorspace.  For BufferedImages, the two
     * color spaces define intermediate spaces through which the source
     * is converted before being converted to the destination space.
     */
    public ColorConvertOp (int srcColorspace, int dstColorspace)
    {
        profileList = new ICC_Profile [2];    /* 2 profiles in the list */

        profileList [0] = ICC_Profile.getInstance (srcColorspace);
        profileList [1] = ICC_Profile.getInstance (dstColorspace);
    }


    /**
     * Constructs a new ColorConvertOp from two ColorSpace objects.
     * This Op is primarily useful for calling the filter method on
     * Rasters, in which case the two ColorSpaces define the operation
     * to be performed on the Rasters.  In that case, the number of bands
     * in the source Raster must match the number of components in
     * srcCspace, and the number of bands in the destination Raster
     * must match the number of components in dstCspace.  For BufferedImages,
     * the two ColorSpaces define intermediate spaces through which the
     * source is converted before being converted to the destination space.
     */
    public ColorConvertOp (ColorSpace srcCspace, ColorSpace dstCspace)
    {
        if ((srcCspace instanceof ICC_ColorSpace) &&
            (dstCspace instanceof ICC_ColorSpace)) {
            profileList = new ICC_Profile [2];    /* 2 profiles in the list */

            profileList [0] = ((ICC_ColorSpace) srcCspace).getProfile();
            profileList [1] = ((ICC_ColorSpace) dstCspace).getProfile();
        }
        else {
            throw new IllegalArgumentException(
                "ColorSpace is not an ICC_ColorSpace");
        }
    }


     /**
     * Constructs a new ColorConvertOp from an array of ICC_Profiles.
     * The sequence of profiles may include profiles that represent color
     * spaces, profiles that represent effects, etc.  If the whole sequence
     * does not represent a well-defined color conversion, an exception is
     * thrown.
     * <p>For BufferedImages, if the ColorSpace
     * of the source BufferedImage does not match the requirements of the
     * first profile in the array,
     * the first conversion is to an appropriate ColorSpace.
     * If the requirements of the last profile in the array are not met
     * by the ColorSpace of the destination BufferedImage,
     * the last conversion is to the destination's ColorSpace.
     * <p>For Rasters, the number of bands in the source Raster must match
     * the requirements of the first profile in the array, and the
     * number of bands in the destination Raster must match the requirements
     * of the last profile in the array.  The array must have at least two
     * elements or calling the filter method for Rasters will throw an
     * IllegalArgumentException.
     * @exception IllegalArgumentException when the profile sequence does not
     *             specify a well-defined color conversion
     */
    public ColorConvertOp (ICC_Profile[] profiles)
    {
        profileList = profiles;    /* save the profiles */
    }


    /**
     * Returns the array of ICC_Profiles used to construct this ColorConvertOp.
     * Returns null if the ColorConvertOp was constructed from an array of
     * ColorSpaces.
     */
    public ICC_Profile[] getICC_Profiles() {
        return    profileList;
    }
     
    /**
     * ColorConvert the BufferedImage.  
     * If the destination image is null,
     * a BufferedImage will be created with an appropriate ColorModel.
     */
    public BufferedImage filter (BufferedImage src, BufferedImage dest) {
    ICC_Profile[]    theProfiles;
    int              nProfiles = profileList.length;
    ColorSpace       srcColorSpace, destColorSpace;
    int              i1, nTransforms, whichTrans, renderState;
    ICC_Transform[]  theTransforms;
    ICC_Profile      srcProfile = null, destProfile = null;

        srcColorSpace = src.getColorModel().getColorSpace();
        if ( !(srcColorSpace instanceof ICC_ColorSpace)) {
            throw new IllegalArgumentException(
                "Source ColorSpace is not an ICC_ColorSpace");
        }
        else {
            srcProfile = ((ICC_ColorSpace) srcColorSpace).getProfile();
        }

        if (dest == null) {        /* last profile in the list defines
                                      the output color space */
            if (nProfiles == 0) {
                throw new IllegalArgumentException(
                    "Destination ColorSpace is undefined");
            }
            nTransforms = nProfiles + 1;
            destProfile = profileList [nProfiles - 1];
            dest = createCompatibleDestImage(src, null);
        }
        else {
            if (src.getHeight() != dest.getHeight() ||
                src.getWidth() != dest.getWidth()) {
                throw new IllegalArgumentException(
                    "Width or height of BufferedImages do not match");
            }
            nTransforms = nProfiles +2;
            destColorSpace = dest.getColorModel().getColorSpace();
            if ( !(destColorSpace instanceof ICC_ColorSpace)) {
                throw new IllegalArgumentException(
                    "Destination ColorSpace is not an ICC_ColorSpace");
            }
            else {
                destProfile = ((ICC_ColorSpace) destColorSpace).getProfile();
            }
        }

        /* make a new transform if needed */
        if ((thisTransform == null) || (thisSrcProfile != srcProfile) ||
            (thisDestProfile != destProfile) ) {

            /* make the profile list */
            theProfiles = new ICC_Profile[nTransforms]; /* the list of profiles
                                                           for this Op */

            theProfiles [0] = srcProfile;  /* insert source as first profile */

            for (i1 = 1; i1 < nTransforms - 1; i1++) {
                                       /* insert profiles defined in this Op */
                theProfiles [i1] = profileList [i1 -1];
            }
            
            theProfiles [nTransforms - 1] = destProfile; /* insert dest as last
                                                            profile */

            /* make the transform list */
            theTransforms = new ICC_Transform [nTransforms];

            /* initialize transform get loop */
            if (theProfiles[0].getProfileClass() == ICC_Profile.CLASS_OUTPUT) {
                                            /* if first profile is a printer
                                               render as colorimetric */
                renderState = ICC_Profile.icRelativeColorimetric;
            }
            else {
                renderState = ICC_Profile.icPerceptual; /* render any other
                                                           class perceptually */
            }

            whichTrans = ICC_Transform.In;

            /* get the transforms from each profile */
            for (i1 = 0; i1 < nTransforms; i1++) {
                if (i1 == nTransforms -1) {         /* last profile? */
                    whichTrans = ICC_Transform.Out; /* get output transform */
                }

                theTransforms[i1] = new ICC_Transform (theProfiles[i1],
                                                       renderState, whichTrans);

                /* get this profile's rendering intent to select transform
                   from next profile */
                renderState = getRenderingIntent(theProfiles[i1]);

                /* "middle" profiles use simulation (or abstract) transform */
                whichTrans = ICC_Transform.Simulation;
            }

            /* make the net transform */
            thisTransform = new ICC_Transform (theTransforms);
            
            /* update corresponding source and dest profiles */
            thisSrcProfile = srcProfile;
            thisDestProfile = destProfile;
        }

        /* color convert the image */
        thisTransform.colorConvert(src, dest);

        return dest;
    }

    /**
     * ColorConvert the image data in the source Raster.
     * If the destination Raster is null, a new Raster will be created.
     * The number of bands in the source and destination Rasters must
     * meet the requirements explained above.  The constructor used to
     * create this ColorConvertOp must have provided enough information
     * to define both source and destination color spaces.  See above.
     * Otherwise, an exception is thrown.
     * @exception IllegalArgumentException if the number of source or
     *             destination bands is incorrect or the source or
     *             destination color spaces are undefined
     */
    public WritableRaster filter (Raster src, WritableRaster dest)  {
        int              nProfiles = profileList.length;


        if (nProfiles < 2) {
            throw new IllegalArgumentException(
                "Source or Destination ColorSpace is undefined");
        }
        if (src.getNumBands() != profileList[0].getNumComponents()) {
            throw new IllegalArgumentException(
                "Numbers of source Raster bands and source color space " +
                "components do not match");
        }
        if (dest == null) {
            dest = createCompatibleDestRaster(src);
        }
        else {
            if (src.getHeight() != dest.getHeight() ||
                src.getWidth() != dest.getWidth()) {
                throw new IllegalArgumentException(
                    "Width or height of Rasters do not match");
            }
            if (dest.getNumBands() !=
                profileList[nProfiles-1].getNumComponents()) {
                throw new IllegalArgumentException(
                    "Numbers of destination Raster bands and destination " +
                    "color space components do not match");
            }
        }

        /* make a new transform if needed */
        if (thisRasterTransform == null) {
            int              i1, whichTrans, renderState;
            ICC_Transform[]  theTransforms;

            /* make the transform list */
            theTransforms = new ICC_Transform [nProfiles];

            /* initialize transform get loop */
            if (profileList[0].getProfileClass() == ICC_Profile.CLASS_OUTPUT) {
                                            /* if first profile is a printer
                                               render as colorimetric */
                renderState = ICC_Profile.icRelativeColorimetric;
            }
            else {
                renderState = ICC_Profile.icPerceptual; /* render any other
                                                           class perceptually */
            }

            whichTrans = ICC_Transform.In;

            /* get the transforms from each profile */
            for (i1 = 0; i1 < nProfiles; i1++) {
                if (i1 == nProfiles -1) {         /* last profile? */
                    whichTrans = ICC_Transform.Out; /* get output transform */
                }

                theTransforms[i1] = new ICC_Transform (profileList[i1],
                                                       renderState, whichTrans);

                /* get this profile's rendering intent to select transform
                   from next profile */
                renderState = getRenderingIntent(profileList[i1]);

                /* "middle" profiles use simulation (or abstract) transform */
                whichTrans = ICC_Transform.Simulation;
            }

            /* make the net transform */
            thisRasterTransform = new ICC_Transform (theTransforms);
        }

        /* color convert the raster */
        thisRasterTransform.colorConvert(src, dest);

        return dest;
    }

    /**
     * Returns the bounding box of the destination, given this source.
     * Note that this will be the same as the the bounding box of the
     * source.
     */
    public Rectangle2D getDestBounds (BufferedImage src) {
        return getDestBounds(src.getRaster());
    }

    /**
     * Returns the bounding box of the destination, given this source.  
     * Note that this will be the same as the the bounding box of the
     * source.
     */
    public Rectangle2D getDestBounds (Raster src) {
        /*        return new Rectangle (src.getXOffset(),
                              src.getYOffset(),
                              src.getWidth(), src.getHeight()); */
        return src.getBounds();
    }

    /**
     * Creates an empty destination image with the correct size and number of
     * components, given this source.
     * @param src       Source image for the filter operation
     * @param destCM    ColorModel of the destination.  If null, an
     *                  appropriate ColorModel will be used.
     */
    public BufferedImage createCompatibleDestImage (BufferedImage src,
                                                    ColorModel destCM) {
        BufferedImage image;
        if (destCM == null) {
            int nProfiles = profileList.length;
            if (nProfiles == 0) {
                throw new IllegalArgumentException(
                    "Destination ColorSpace is undefined");
            }
            ICC_Profile destProfile = profileList[nProfiles - 1];
            ColorModel srcCM = src.getColorModel();
            ICC_ColorSpace cs = new ICC_ColorSpace(destProfile);
            int nbands = cs.getNumComponents();
            boolean hasAlpha = srcCM.hasAlpha();
            if (hasAlpha) {
               nbands += 1;
            }
            int[] nbits = new int[nbands];
            for (int i = 0; i < nbands; i++) {
                nbits[i] = 8;
            }
            destCM = new ComponentColorModel(cs, nbits, hasAlpha,
                                             srcCM.isAlphaPremultiplied(),
                                             srcCM.getTransparency());
        }
        int w = src.getWidth();
        int h = src.getHeight();
        image = new BufferedImage(destCM,
                                  destCM.createCompatibleWritableRaster(w, h),
                                  destCM.isAlphaPremultiplied());
        return image;
    }

    
    /**
     * Creates an empty destination Raster with the correct size and number of
     * bands, given this source.
     */
    public WritableRaster createCompatibleDestRaster (Raster src) {
        int nProfiles = profileList.length;

        if (nProfiles < 2) {
            throw new IllegalArgumentException(
                "Destination ColorSpace is undefined");
        }
        WritableRaster dest = Raster.createComponentRaster(DataBuffer.BYTE_DATA,
                                  src.getWidth(),
                                  src.getHeight(),
                                  profileList[nProfiles-1].getNumComponents(),
                                  new Point(src.getMinX(), src.getMinY()));
        return dest;
    }
    
    /**
     * Returns the location of the destination point given a
     * point in the source image.  If dstPt is non-null, it will
     * be used to hold the return value.
     */
    public Point2D getDestPoint (Point2D srcPt, Point2D dstPt) {
        if (dstPt == null) {
            dstPt = new Point2D.Float();
        }
        dstPt.setLocation(srcPt.getX(), srcPt.getY());

        return dstPt;
    }


    /**
     * Returns the RenderingIntent from the specified ICC Profile.
     */
    private int getRenderingIntent (ICC_Profile profile) {
        byte[] header = profile.getData(ICC_Profile.icSigHead);
        int index = ICC_Profile.icHdrRenderingIntent;
        return (((int)header[index+0] << 24) |
                ((int)header[index+1] << 16) |
                ((int)header[index+2] <<  8) |
                ((int)header[index+3] <<  0) );
    }
    
}
