/*
 * @(#)ColorModel.java	1.40 98/03/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.awt.Transparency;
import java.awt.color.ColorSpace;
import java.awt.Toolkit;

/**
 * An abstract class that encapsulates the methods for translating from pixel 
 * values to color components (e.g. red, green, blue) for an image.  
 * Pixels may be represented as 32-bit ints or as arrays of 32-bit ints.
 *
 * @see IndexColorModel
 * @see ComponentColorModel
 * @see PackedColorModel
 * @see DirectColorModel
 * @version 10 Feb 1997
 */
public abstract class ColorModel implements Transparency{
    private long pData;		// Placeholder for data for native functions

    protected int pixel_bits;
    int nBits[] = null;
    int transparency = Transparency.TRANSLUCENT;
    boolean supportsAlpha = true;
    boolean isAlphaPremultiplied = false;
    int numComponents = -1;
    int numColorComponents = -1;
    ColorSpace colorSpace = ColorSpace.getInstance(ColorSpace.CS_sRGB);
    int colorSpaceType = ColorSpace.TYPE_RGB;
    int maxBits = 0;
    boolean is_sRGB = true;
    protected int transferType;

    /**
     * This is copied from java.awt.Toolkit since we need the library
     * loaded in java.awt.image also:
     *
     * WARNING: This is a temporary workaround for a problem in the
     * way the AWT loads native libraries. A number of classes in the
     * AWT package have a native method, initIDs(), which initializes
     * the JNI field and method ids used in the native portion of
     * their implementation.
     *
     * Since the use and storage of these ids is done by the
     * implementation libraries, the implementation of these method is
     * provided by the particular AWT implementations
     * (i.e. "Toolkit"s/Peer), such as Motif, Win32 or Tiny. The
     * problem is that this means that the native libraries must be
     * loaded by the java.* classes, which do not necessarily know the
     * names of the libraries to load. A better way of doing this
     * would be to provide a separate library which defines java.awt.*
     * initIDs, and exports the relevant symbols out to the
     * implementation libraries.
     * 
     * For now, we know it's done by the implementation, and we assume
     * that the name of the library is "awt".  -br.
     */
    static void loadLibraries() {
        try {
	    java.security.AccessController.beginPrivileged();
	    System.loadLibrary("fontmanager");
	    System.loadLibrary("awt");
	} finally {
	    java.security.AccessController.endPrivileged();
	}
    }
    private static native void initIDs();
    static {
	/* ensure that the proper libraries are loaded */
        loadLibraries();
	initIDs();
    }
    private static ColorModel RGBdefault;

    /**
     * Return a DirectColorModel that describes the default format for
     * integer RGB values used in many of the methods in the
     * AWT image interfaces for the convenience of the programmer.
     * The color space is the default ColorSpace, sRGB.
     * The format for the RGB values is an integer with 8 bits
     * each of alpha, red, green, and blue color components ordered
     * correspondingly from the most significant byte to the least
     * significant byte, as in:  0xAARRGGBB.  Color components are
     * not premultiplied by the alpha component.  This format does not
     * necessarily represent the native or the most efficient
     * ColorModel for a particular device or for all images.  It is
     * merely used as a common color model format.
     */
    public static ColorModel getRGBdefault() {
	if (RGBdefault == null) {
	    RGBdefault = new DirectColorModel(32,
					      0x00ff0000,	// Red
					      0x0000ff00,	// Green
					      0x000000ff,	// Blue
					      0xff000000	// Alpha
					      );
	}
	return RGBdefault;
    }

    /**
     * Constructs a ColorModel which describes a pixel of the specified
     * number of bits.  The color space is the default RGB ColorSpace,
     * which is sRGB.  
     */
    public ColorModel(int bits) {
	pixel_bits = bits;
        if (bits < 1) {
            throw new IllegalArgumentException("Number of bits must be > 0");
        }
        // REMIND: make sure transferType is set correctly
        if (pixel_bits > 16) {
           transferType = DataBuffer.INT_DATA;
        } else if (pixel_bits > 8) {
           transferType = DataBuffer.SHORT_DATA;
        } else {
           transferType = DataBuffer.BYTE_DATA;
        }
    }

    /**
     * Constructs a ColorModel which describes a pixel of the specified
     * number of bits in the color space of the specified ColorSpace.
     * The boolean, isAlphaPremultiplied, specifies how to interpret color
     * data that are passed to this ColorModel.  If the boolean is true, the
     * color data have been multiplied with alpha.  The transparency 
     * specifies what alpha values can be represented by this color model.
     *
     * Any ColorModel object that uses this constructor should override
     * the getComponentSize() method to return the actual bits per
     * component array.  This constructor is here to support legacy
     * code.
     * @see java.awt.Transparency
     */
    protected ColorModel(int bits, ColorSpace cspace,
                         boolean hasAlpha,
                         boolean isAlphaPremultiplied,
                         int transparency) {
        if (bits <= 0) {
            throw new IllegalArgumentException("Number of bits must be > 0");
        }
        pixel_bits = bits;
        colorSpace                = cspace;
        colorSpaceType            = cspace.getType();
        numColorComponents        = cspace.getNumComponents();
        numComponents             = numColorComponents + (hasAlpha ? 1 : 0);
        supportsAlpha             = hasAlpha;
        if (supportsAlpha == false) {
            this.isAlphaPremultiplied = true;
            this.transparency = Transparency.OPAQUE;
        }
        else {
            this.isAlphaPremultiplied = isAlphaPremultiplied;
            this.transparency         = transparency;
        }

        // Save this since we always need to check if it is the default CS
        if (cspace != ColorSpace.getInstance(ColorSpace.CS_sRGB)) {
            is_sRGB = false;
        }
            
        // Get the bits array
        nBits = getComponentSize();

        // REMIND: make sure transferType is set correctly
        if (pixel_bits > 16) {
           transferType = DataBuffer.INT_DATA;
        } else if (pixel_bits > 8) {
           transferType = DataBuffer.SHORT_DATA;
        } else {
           transferType = DataBuffer.BYTE_DATA;
        }
    }

    /**
     * Constructs a ColorModel which describes a pixel 
     * in the color space of the specified ColorSpace.  The bits array
     * specifies the number of significant bits per component in the
     * color model (color components and alpha).  
     * The boolean, isAlphaPremultiplied, specifies how to
     * interpret color data that
     * are passed to this ColorModel.  If the boolean is true, the
     * color data have been multiplied with alpha.  The transparency 
     * specifies what alpha values can be represented by this color model.
     * @see java.awt.Transparency
     */
    protected ColorModel(int[] bits, ColorSpace cspace,
                         boolean hasAlpha,
                         boolean isAlphaPremultiplied,
                         int transparency) {
        colorSpace                = cspace;
        colorSpaceType            = cspace.getType();
        numColorComponents        = cspace.getNumComponents();
        numComponents             = numColorComponents + (hasAlpha ? 1 : 0);
        supportsAlpha             = hasAlpha;
        if (bits.length < numComponents) {
            throw new IllegalArgumentException("Number of color/alpha "+
                                               "components should be "+
                                               numComponents+
                                               " but length of bits array is "+
                                               bits.length);
        }
        if (supportsAlpha == false) {
            this.isAlphaPremultiplied = true;
            this.transparency = Transparency.OPAQUE;
        }
        else {
            this.isAlphaPremultiplied = isAlphaPremultiplied;
            this.transparency         = transparency;
        }

        nBits = new int[bits.length];
        
        // Sum all of the bits in the array
        pixel_bits = 0;
        maxBits = 0;
        for (int i=0; i < bits.length; i++) {
            if (bits[i] <= 0) {
                throw new
                    IllegalArgumentException("Number of bits must be > 0");
            }
            nBits[i] = bits[i];
            pixel_bits += bits[i];
            if (maxBits < nBits[i]) {
                maxBits = nBits[i];
            }
        }

        // REMIND: make sure transferType is set correctly
        if (maxBits <= 8) {
           transferType = DataBuffer.BYTE_DATA;
        } else if (maxBits <= 16) {
           transferType = DataBuffer.SHORT_DATA;
        } else {
           transferType = DataBuffer.INT_DATA;
        }

        // Save this since we always need to check if it is the default CS
        if (cspace != ColorSpace.getInstance(ColorSpace.CS_sRGB)) {
            is_sRGB = false;
        }
            
    }

    /**
     * Returns whether or not alpha is supported in this ColorModel.
     */
    final public boolean hasAlpha() {
        return supportsAlpha;
    }
    
    /**
     * Returns whether or not the alpha has been premultiplied in the
     * Raster associated with this ColorModel.
     */
    final public boolean isAlphaPremultiplied() {
        return isAlphaPremultiplied;
    }
    
    /**
     * Returns the number of bits per pixel described by this ColorModel.
     */
    public int getPixelSize() {
	return pixel_bits;
    }

    /**
     * Returns the number of bits for the given component.  
     */
    public int getComponentSize(int componentIdx) {
        // REMIND:
        if (nBits == null) {
            throw new NullPointerException("Number of bits array is null.");
        }
        
        return nBits[componentIdx];
    }
    
    /**
     * Returns an array of the number of bits per component.  
     */
    public int[] getComponentSize() {
        if (nBits != null) {
            return (int[]) nBits.clone();
        }

        return null;
    }

    /**
     * Returns the maximum number of bits for all components.
     */
    final public int getMaxComponentSize() {
        if (nBits == null) {
            throw new NullPointerException("Number of bits array is null.");
        }
        int maxSize = nBits[0];
        for (int i=1; i < nBits.length; i++) {
            if (maxSize < nBits[i]) {
                maxSize = nBits[i];
            }
        }
        return maxSize;
    }
    
    /**
     * Returns the transparency.  Returns either OPAQUE, BITMASK,
     * or TRANSLUCENT
     * @see Transparency#OPAQUE
     * @see Transparency#BITMASK
     * @see Transparency#TRANSLUCENT
     */
    public int getTransparency() {
        return transparency;
    }

    /**
     * Returns the number of components (including alpha) in this color model.
     * This is equal to the number of color components, optionally plus
     * one, if there is an alpha component.
     */
    public int getNumComponents() {
        return numComponents;
    }
    
    /**
     * Returns the number of color components in this color model.
     * This is the number of components returned by
     * ColorSpace.getNumComponents().
     * @see ColorSpace#getNumComponents
     */
    public int getNumColorComponents() {
        return numColorComponents;
    }

    /**
     * Returns the red
     * color component for the specified pixel, scaled from 0 to 255
     * in the default RGB ColorSpace, sRGB.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.  The returned value will 
     * be a non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public abstract int getRed(int pixel);

    /**
     * Returns the green
     * color component for the specified pixel, scaled from 0 to 255
     * in the default RGB ColorSpace, sRGB.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.  The returned value will be
     * a non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public abstract int getGreen(int pixel);

    /**
     * Returns the blue
     * color component for the specified pixel, scaled from 0 to 255
     * in the default RGB ColorSpace, sRGB.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.  The returned value will be
     * a non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public abstract int getBlue(int pixel);

    /**
     * Returns the alpha
     * component for the specified pixel, scaled from 0 to 255
     * in the default RGB ColorSpace, sRGB.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.  
     */
    public abstract int getAlpha(int pixel);

    /**
     * Returns the color of the pixel in the default RGB color model.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.  The returned value will be
     * a non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     * @see ColorModel#getRGBdefault
     */
    public int getRGB(int pixel) {
	return (getAlpha(pixel) << 24)
	    | (getRed(pixel) << 16)
	    | (getGreen(pixel) << 8)
	    | (getBlue(pixel) << 0);
    }

    /**
     * Returns the red
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getRed(Object inData) {
        int pixel=0,length=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
               length = bdata.length;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
               length = sdata.length;
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
               length = idata.length;
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        if (length == 1) {
            return getRed(pixel);
        }
        else {
            throw new NoSuchMethodError
                ("This method is not supported by this color model");
        }
    }

    /**
     * Returns the green
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of
     * channel elements, one per channel.  The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getGreen(Object inData) {
        int pixel=0,length=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
               length = bdata.length;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
               length = sdata.length;
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
               length = idata.length;
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        if (length == 1) {
            return getGreen(pixel);
        }
        else {
            throw new NoSuchMethodError
                ("This method is not supported by this color model");
        }
    }
    
    /**
     * Returns the blue
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of
     * channel elements, one per channel.  The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getBlue(Object inData) {
        int pixel=0,length=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
               length = bdata.length;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
               length = sdata.length;
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
               length = idata.length;
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        if (length == 1) {
            return getBlue(pixel);
        }
        else {
            throw new NoSuchMethodError
                ("This method is not supported by this color model");
        }
    }

    /**
     * Returns the alpha
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of
     * channel elements, one per channel.
     */
    public int getAlpha(Object inData) {
        int pixel=0,length=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
               length = bdata.length;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
               length = sdata.length;
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
               length = idata.length;
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        if (length == 1) {
            return getAlpha(pixel);
        }
        else {
            throw new NoSuchMethodError
                ("This method is not supported by this color model");
        }
    }

    /**
     * Returns a pixel in the default RGB color model given a pixel array
     * in the current color model.
     * The pixel is described by a discrete array of
     * channel elements, one per channel.  The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     * @see ColorModel#getRGBdefault
     */
    public int getRGB(Object inData) {
        return (getAlpha(inData) << 24)
            | (getRed(inData) << 16)
            | (getGreen(inData) << 8)
            | (getBlue(inData) << 0);
    }

    /**
     * Returns a pixel array given a pixel in the default RGB color model.
     * This pixel array can then be passed to the putPixelData method of
     * a Raster object.  If the pixel variable is null, a new array will
     * be allocated.  The pixel array will be returned.
     */
    public Object getPixelData(int rgb, Object pixel) {
        throw new NoSuchMethodError
            ("This method is not supported by this color model.");
    }
    
    /**
     * Returns an array of all of the components given a pixel 
     * in the current color model.  If the color model is not
     * an IndexColorModel or DirectColorModel, an IllegalArgumentException
     * will be thrown.  This can be used to retrieve the
     * components for a specific pixel in a PackedColorModel.  If the
     * components array is null, a new array will be allocated.  The
     * components array will be returned.
     */
    public int[] getComponents(int pixel, int[] components, int offset) {
        throw new NoSuchMethodError
            ("This method is not supported by this color model.");
    }
    
    /**
     * Returns an array of all of the components given a pixel array
     * in the current color model.  This can be used to retrieve the
     * components for a specific pixel in a PackedColorModel.  If the
     * components array is null, a new array will be allocated.  The
     * components array will be returned.
     */
    public int[] getComponents(Object obj, int[] components, int offset) {
        throw new NoSuchMethodError
            ("This method is not supported by this color model.");
    }

    /**
     * Returns an array of all of the components given a normalized
     * component array.
     */
    public int[] getComponents(float[] normComponents, int normOffset,
                               int[] components, int offset) {
        // Make sure that someone isn't using a custom color model
        // that called the super(bits) constructor.
        if (colorSpace == null) {
            throw new NoSuchMethodError("This method is not supported "+
                                        "by this color model.");
        }

        if (nBits == null) {
            throw new NoSuchMethodError ("This method is not supported.  "+
                                         "Unable to determine #bits per "+
                                         "component.");
        }
        if ((normComponents.length - normOffset) < numColorComponents) {
            throw new
                IllegalArgumentException(
                        "Incorrect number of color components.  Expecting "+
                        numColorComponents);
        }
        
        if (components == null) {
            components = new int[offset+numColorComponents];
        }

        for (int i=0; i < numColorComponents; i++) {
            components[offset+i] = (int) (normComponents[normOffset+i]
                                   * (1<<nBits[i] - 1));
        }

        return components;
    }

    /**
     * Returns an array of normalized components given an array of all
     * components.
     */
    public float[] getNormalizedComponents(int[] components, int offset,
                                           float[] normComponents,
                                           int normOffset) {
        // Make sure that someone isn't using a custom color model
        // that called the super(bits) constructor.
        if (colorSpace == null) {
            throw new NoSuchMethodError("This method is not supported by "+
                                        "this color model.");
        }
        if (nBits == null) {
            throw new NoSuchMethodError ("This method is not supported.  "+
                                         "Unable to determine #bits per "+
                                         "component.");
        }
        
        if ((components.length - offset) < numColorComponents) {
            throw new
                IllegalArgumentException(
                        "Incorrect number of color components.  Expecting "+
                        numColorComponents);
        }
        
        if (normComponents == null) {
            normComponents = new float[numComponents];
        }

        for (int i=0; i < numComponents; i++) {
            normComponents[normOffset+i] =
                (float) components[offset+i] / ((1<<nBits[i]) - 1.f);
        }

        return normComponents;
    }

    /**
     * This method will pack the components into a pixel given an
     * array of discrete components in this color model.  This method
     * will throw an IllegalArgumentException if the components cannot
     * fit in a single integer pixel.
     */
    public int setComponents(int[] components, int offset) {
        throw new NoSuchMethodError("This method is not supported "+
                                    "by this color model.");
    }
    
    /**
     * Returns a pixel, given a discrete components (unpacked color/alpha
     * values) in this color model.
     */
    public Object setComponents(int[] components, int offset, Object obj) {
        throw new NoSuchMethodError("This method has not been implemented "+
                                    "for this color model.");
    }

    public boolean equals(Object obj) {
        if (!(obj instanceof ColorModel)) {
            return false;
        }
        ColorModel cm = (ColorModel) obj;
        
        if (this == cm) {
            return true;
        }
        if (supportsAlpha != cm.hasAlpha() ||
            isAlphaPremultiplied != cm.isAlphaPremultiplied() ||
            pixel_bits != cm.getPixelSize() ||
            transparency != cm.getTransparency() ||
            numComponents != cm.getNumComponents())
        {
            return false;
        }
        int i;
        int[] nb = cm.getComponentSize();
        for (i=0; i < numComponents; i++) {
            if (nBits[i] != nb[i]) {
                return false;
            }
        }

        return true;
    }
    
    /**
     * Returns the ColorSpace associated with this ColorModel
     */
    final public ColorSpace getColorSpace() {
        return colorSpace;
    }

    /**
     * Forces the Raster data to match the state specified in the
     * isAlphaPremultiplied variable, assuming the data is currently 
     * correctly described by this ColorModel.  It may multiply or divide
     * the color Raster data by alpha, or do nothing if the data is
     * in the correct state.  If the data needs to be coerced, this
     * method will also return an instance of this ColorModel with
     * the isAlphaPremultiplied flag set appropriately.
     */
    public ColorModel coerceData (WritableRaster raster,
                                  boolean isAlphaPremultiplied) {
        throw new NoSuchMethodError
            ("This method is not supported by this color model");
    }

    /**
      * Returns true if raster is compatible with this ColorModel and
      * false if it is not.
      */
    public boolean isCompatibleRaster(Raster raster) {
	throw new NoSuchMethodError(
            "This method has not been implemented for this ColorModel.");
    }

    /**
     * Creates a WritableRaster with the specified width and height, that 
     * has a channel layout compatible with this ColorModel.  
     * @see WritableRaster
     */
    public WritableRaster createCompatibleWritableRaster(int w, int h) {
	throw new NoSuchMethodError
            ("This method is not supported by this color model");
    }
    
    public void finalize() {
    }

    /**
     * Prints the contents of this object
     */
    public String toString() {
       return new String("ColorModel: #pixelBits = "+pixel_bits
                         + " numComponents = "+numComponents
                         + " color space = "+colorSpace
                         + " transparency = "+transparency
                         + " has alpha = "+supportsAlpha
                         + " isAlphaPre = "+isAlphaPremultiplied
                         );
    }
}
