/*
 * @(#)PackedColorModel.java	1.19 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.awt.Transparency;
import java.awt.color.ColorSpace;

/**
 * An abstract ColorModel class that represents pixel values
 * that have the color components embedded directly in the bits of an
 * integer pixel.  A subclass is the DirectColorModel, which
 * is similar to an X11 TrueColor visual.
 *
 * @see DirectColorModel
 * @version 10 Feb 1997
 */

public abstract class PackedColorModel extends ColorModel {
    int[] maskArray;
    int[] maskOffsets;
    double[] scaleFactors;

    private static native void initIDs();
    static {
        loadLibraries();
        initIDs();
    }
    
    /**
     * Constructs a PackedColorModel from a mask array, which specifies
     * which bits in the integer pixel contain each of the color
     * components in the ColorSpace.  All of the bits in each mask
     * must be contiguous and fit in the specified number of least
     * significant bits of the integer.  If the alphaMask is 0,
     * there is no alpha.  If there is alpha, the boolean,
     * isAlphaPremultiplied, specifies how to interpret color data that
     * are passed to this ColorModel.  If it is true, the color data will
     * have been multiplied with alpha.
     */
    public PackedColorModel (ColorSpace space, int bits,
                             int[] colorMaskArray, int alphaMask,
                             boolean isAlphaPremultiplied,
                             int trans) {
        super(bits, space, (alphaMask == 0 ? false : true),
              isAlphaPremultiplied, trans);
        if (bits < 1 || bits > 32) {
            throw new IllegalArgumentException("Number of bits must be between"
                                               +" 1 and 32.");
        }
        maskArray   = new int[numComponents];
        maskOffsets = new int[numComponents];
        nBits       = new int[numComponents];
        scaleFactors = new double[numComponents];

        for (int i=0; i < numColorComponents; i++) {
            // Get the mask offset and #bits
            DecomposeMask(colorMaskArray[i], i, space.getName(i));
        }
        if (alphaMask != 0) {
            DecomposeMask(alphaMask , numColorComponents,
                          space.getName(numColorComponents));
            if (nBits[numComponents-1] == 1) {
                transparency = Transparency.BITMASK;
            }
        }
    }
    /**
     * Constructs a PackedColorModel from the given masks specifying
     * which bits in the pixel contain the alpha, red, green and blue
     * color components.  All of the bits in each mask must be contiguous
     * and fit in the specified number of least significant bits of the
     * integer.  The boolean,
     * isAlphaPremultiplied, specifies how to interpret color data that
     * are passed to this ColorModel.  If it is true, the color data will
     * have been multiplied with alpha.
     * The ColorSpace should be of type ColorSpace.TYPE_RGB.
     * @see ColorSpace
     */
    public PackedColorModel(ColorSpace space, int bits, int rmask, int gmask,
                            int bmask, int amask,
                            boolean isAlphaPremultiplied,
                            int trans) {
        super (bits, space, (amask == 0 ? false : true),
               isAlphaPremultiplied, trans);

        if (space.getType() != ColorSpace.TYPE_RGB) {
            throw new IllegalArgumentException("ColorSpace must be TYPE_RGB.");
        }
        maskArray = new int[numComponents];
        maskOffsets = new int[numComponents];
        nBits       = new int[numComponents];
        scaleFactors = new double[numComponents];
        
        DecomposeMask(rmask, 0, "red");
        
        DecomposeMask(gmask, 1, "green");
        
        DecomposeMask(bmask, 2, "blue");
        
        if (amask != 0) {
            DecomposeMask(amask, 3, "alpha");
            if (nBits[3] == 1) {
                transparency = Transparency.BITMASK;
            }
        }
    }

    /**
     * Returns the mask indicating which bits in a pixel contain the 
     * color component.  The index corresponds to the placement of
     * color component names in the color space.  Thus, index 0 for
     * a CMYK ColorSpace would correspond to Cyan and index 1 would
     * correspond to Magenta.  If there is alpha, the alpha index
     * would be:
     * <pre>
     *      alphaIndex = numComponents() - 1;
     * </pre>
     */
    final public int getMask(int index) {
	return maskArray[index];
    }

    /**
     * Returns the mask array indicating which bits in a pixel contain the 
     * color and alpha components.  
     */
    final public int[] getMask() {
	return (int[]) maskArray.clone();
    }
    
    /*
     * A utility function to decompose a single mask and verify that it
     * fits in the specified pixel size, and that it does not overlap any
     * other color component.  
     */
    private void DecomposeMask(int mask,  int idx, String componentName) {
	int off = 0;
	int count = 0;

        // Store the mask
        maskArray[idx]   = mask;

        // Now find the #bits and shifts
	if (mask != 0) {
	    while ((mask & 1) == 0) {
		mask >>>= 1;
		off++;
	    }
	    while ((mask & 1) == 1) {
		mask >>>= 1;
		count++;
	    }
	}
	if (mask != 0) {
	    throw new IllegalArgumentException(componentName +
                                               " mask bits not contiguous");
	}
	if (off + count > pixel_bits) {
	    throw new IllegalArgumentException(componentName +
                                               " mask overflows pixel");
	}

	maskOffsets[idx] = off;
	nBits[idx]       = count;
	if (count == 0) {
	    // High enough to scale any 0-ff value down to 0.0, but not
	    // high enough to get Infinity when scaling back to pixel bits
	    scaleFactors[idx] = 256.0;
	} else {
	    scaleFactors[idx] = 255. /((1 << count) - 1);
	}

        // Store the maximum #bits per component
        if (maxBits < count) {
            maxBits = count;
        }
    }

    public boolean equals(Object obj) {
        if (!(obj instanceof PackedColorModel)) {
            return false;
        }

        if (!super.equals(obj)) {
            return false;
        }

        PackedColorModel cm = (PackedColorModel) obj;
        int numC = cm.getNumComponents();
        if (numC != numComponents) {
            return false;
        }
        for(int i=0; i < numC; i++) {
            if (maskArray[i] != cm.getMask(i)) {
                return false;
            }
        }
        return true;
    }
}

