/*
 * @(#)ContextualRenderedImageFactory.java	1.4 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* ********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.image.renderable;
import java.awt.geom.Rectangle2D;
import java.awt.image.RenderedImage;

/**
 * ContextualRenderedImageFactory provides an interface for the
 * functionality that may differ between instances of
 * RenderableImageOp.  In this way there does not need to be a
 * subclass of RenderableImageOp for each rendering-independent
 * operator.
 *
 * <p> All operations that are to be used in a rendering-independent
 * chain must implement ContextualRenderedImageFactory.
 *
 * <p> This interface is intended to be used only by the
 * RenderableImageOp class.  It is not intended to be used by
 * developers, unless they are writing a RenderableImageOp.
 *
 * <p> Classes that implement this interface must provide a
 * constructor with no arguments so they can be instantiated by the
 * class loader.
 */
public interface ContextualRenderedImageFactory extends RenderedImageFactory {
 
    /**
     * Maps the operation's output RenderContext into a RenderContext
     * for each of the operation's sources.  This is useful for
     * operations that can be expressed in whole or in part simply as
     * alterations in the RenderContext, such as an affine mapping, or
     * operations that wish to obtain lower quality renderings of
     * their sources in order to save processing effort or
     * transmission bandwith.  Some operations, such as blur, can also
     * use this mechanism to avoid obtaining sources of higher quality
     * than necessary.
     *
     * @param i the index of the source image.
     * @param renderContext the RenderContext being applied to the operation.
     * @param paramBlock a ParameterBlock containing the operation's
     *        sources and parameters.
     * @param parentOp the RenderableImageOp from which this method is
     *        being called.
     */
    RenderContext mapRenderContext(int i,
                                   RenderContext renderContext,
                                   ParameterBlock paramBlock,
                                   RenderableImageOp parentOp);
       
    /**
     * Creates a rendering, given a RenderContext and a ParameterBlock
     * containing the operation's sources and parameters.  The output
     * is a RenderedImage that takes the RenderContext into account to
     * determine its dimensions and placement on the image plane.
     * This method houses the "intelligence" that allows a
     * rendering-independent operation to adapt to a specific render
     * context.
     *
     * @param renderContext The RenderContext that specifies the rendering.
     * @param paramBlock a ParameterBlock containing the operation's
     *        sources and parameters.
     */
    RenderedImage createOp(RenderContext renderContext,
                           ParameterBlock paramBlock);
    
    /**
     * Returns the bounding box for the output of the operation,
     * performed on a given set of sources.  The bounds are returned
     * as a Rectangle2D, that is, and axis-aligned rectangle with
     * floating-point corner coordinates.
     *
     * @param paramBlock a ParameterBlock containing the operation's
     *        sources and parameters.
     * @return a Rectangle2D specifying the bounding box of the output.
     */
    Rectangle2D getBoundingBox(ParameterBlock paramBlock);

    /**
     * Gets the appropriate instance of the property specified by the name 
     * parameter.  This method must determine which instance of a property to
     * return when there are multiple sources that each specify the property.
     *
     * @param paramBlock a ParameterBlock containing the operation's
     *        sources and parameters.
     * @param name a String naming the desired property.
     * @return an object reference to the value of the property requested.
     */
    Object getProperty(ParameterBlock pb, String name);
}
