/*
 * @(#)RenderableImage.java	1.7 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* ********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.image.renderable;
import java.util.Vector;
import java.util.Hashtable;
import java.awt.image.*;

/**
 * A RenderableImage is a common interface for rendering-independent
 * images (a notion which subsumes resolution independence).  That is,
 * images which are described and have operations applied to them
 * independent of any specific rendering of the image.  For example, a
 * RenderableImage can be rotated and cropped in
 * resolution-independent terms.  Then, it can be rendered for various
 * specific contexts, such as a draft preview, a high-quality screen
 * display, or a printer, each in an optimal fashion.
 * 
 * <p> A RenderedImage is returned from a RenderableImage via the getImage()
 * method, which takes a RenderContext.  The RenderContext specifies
 * how the RenderedImage should be constructed.  Note that it is not
 * possible to extract pixels directly from a RenderableImage.
 * 
 * <p> The getDefaultRendering() and getScaledRendering() methods are
 * convenience methods that construct an appropriate RenderContext
 * internally.
 */
public interface RenderableImage {
  
    /** 
     * Returns a vector of RenderableImages that are the sources of
     * Image data for this RenderableImage. Note that this method will
     * often return an empty vector.
     *
     * @return a Vector of RenderableImages.
     */
    Vector getSources();
    
    /** 
     * Gets a property from the property set of this image.
     *
     * @param name the name of the property to get, as a String.
     * @return a reference to the property Object.
     */
    Object getProperty(String name);

    /** 
     * Return a list of names recognized by getProperty(String). 
     */
    String[] getPropertyNames();
    
    /** 
     * Gets the width in user coordinate space.  The default width
     * and height for image sources should be 1 unit high, by aspect
     * ratio wide.
     *
     * @return the width of the image in user coordinates.
     */
    float getWidth();
  
    /** 
     * Gets the height in user coordinate space.  The default width
     * and height for image sources should be 1 unit high, by aspect
     * ratio wide.
     *
     * @return the height of the image in user coordinates.
     */
    float getHeight();
    
    /** 
     * Gets the minimum X coordinate for image data.  Be aware that a
     * RenderableImage may have "infinite" extent, in which case this
     * function will return -Float.MAX_VALUE.
     *
     * @return The minimum X coordinate of the rendering-independent image.
     */
    float getMinXCoord();
  
    /** 
     * Gets the maxumum X coordinate for image data.  Be aware that a
     * RenderableImage may have "infinite" extent, in which case this
     * function will return Float.MAX_VALUE.
     *
     * @return The maximum X coordinate of the rendering-independent image.
     */
    float getMaxXCoord();

    /** 
     * Gets the minimum Y coordinate for image data.  Be aware that a
     * RenderableImage may have "infinite" extent, in which case this
     * function will return -Float.MAX_VALUE.
     *
     * @return The minimum Y coordinate of the rendering-independent image.
     */
    float getMinYCoord();

    /** 
     * Gets the maximum Y coordinate for image data.  Be aware that a
     * RenderableImage may have "infinite" extent, in which case this
     * function will return Float.MAX_VALUE.
     *
     * @return The maximum Y coordinate of the rendering-independent image.
     */
    float getMaxYCoord();
  
    /** 
     * Gets a RenderedImage instance of this image with width w, and
     * height h in pixels.  The RenderContext is built automatically
     * with an appropriate usr2dev transform and an area of interest
     * of the full image.  All the rendering hints come from hints
     * passed in.
     *
     * @param w the width of rendered image in pixels.
     * @param h the height of rendered image in pixels.
     * @param hints a Hashtable of Hints to use during rendering.
     * @return a RenderedImage containing the rendered data.
     */
    RenderedImage getScaledRendering(int w, int h, Hashtable hints);
  
    /** 
     * Gets a RenderedImage instance of this image with a default
     * width and height in pixels.  The RenderContext is built
     * automatically with an appropriate usr2dev transform and an area
     * of interest of the full image.  All the rendering hints come
     * from hints passed in.  Implementors of this interface must be
     * sure that there is a defined default width and height.
     *
     * @return a RenderedImage containing the rendered data.
     */
    RenderedImage getDefaultRendering();
  
    /** 
     * Gets a RenderedImage instance of this image from a
     * RenderContext.  This is the most general way to obtain a
     * rendering of a RenderableImage.
     *
     * @param renderContext the RenderContext to use to produce the rendering.
     * @return a RenderedImage containing the rendered data.
     */
    RenderedImage getImage(RenderContext renderContext);
}
