/*
 * @(#)RenderableImageOp.java	1.5 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* ********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.image.renderable;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.RenderedImage;
import java.util.Hashtable;
import java.util.Vector;

/**
 * This class handles the renderable aspects of an operation with help
 * from its associated instance of a ContextualRenderedImageFactory.
 */
public class RenderableImageOp implements RenderableImage {

    /** A ParameterBlock containing source and parameters. */
    ParameterBlock pb;

    /** The sources from the ParameterBlock. */
    Vector sources;

    /** The name of the associated ContextualRenderedImageFactory class. */
    String CRIFName;

    /** The associated ContextualRenderedImageFactory. */
    ContextualRenderedImageFactory myCRIF;

    /** The bounding box of the results of this RenderableImageOp. */
    Rectangle2D boundingBox;

    /** 
     * Constructs a RenderedImageOp given the name of a class
     * implementing the ContextualRenderedImageFactory interface, and
     * a ParameterBlock containing RenderableImage sources and other
     * parameters.  Any RenderedImage sources referenced by the
     * ParameterBlock will be ignored.
     *
     * @param name a String containing the name of a class
     *        implementing ContextualRenderedImageFactory that
     *        implements the desired operation.  This will be
     *        dynamically loaded by Class.forName().
     * @param pb a ParameterBlock conating this operation's source
     *        images and other parameters necessary for the operation
     *        to run.
     * @exception java.lang.ClassNotFoundException if the class cannot
     *            be found.
     * @exception java.lang.IllegalAccessException if the class or
     *            initializer is not accessible.
     * @exception java.lang.InstantiationException if the CRIF class
     *            is an abstract class, an interface, not an instance
     *            of ContextualRenderedImageFactory, or if its
     *            instantiation fails for any other reason.
     * @see java.lang.Class */
    public RenderableImageOp(String CRIFName, ParameterBlock pb) 
        throws ClassNotFoundException, IllegalAccessException, 
               InstantiationException {
        this.CRIFName = CRIFName;
        this.pb = (ParameterBlock) pb.clone();
        boundingBox = null;
        
        if (pb.getNumSources() > 0) {
            sources = new Vector();
            int i = 0;
            while (i < pb.getNumSources()) {
                Object o = pb.getSource(i);
                if (o instanceof RenderableImage) {
                    sources.add((RenderableImage)o);
                    i++;
                } else {
                    break;
                }
            }
            
        }
        
        // Dynamically load the CRIF by name.
        Object obj = (Class.forName(CRIFName)).newInstance();
        if (obj instanceof ContextualRenderedImageFactory) {
            myCRIF = (ContextualRenderedImageFactory)obj;
        } else {
            throw new InstantiationException(obj.getClass().getName() + 
                         " does not implement ContextualRenderedImageFactory");
        }
    }
    
    /** 
     * Produces a RenderedImage which represents this
     * RenderableImageOp (including its Renderable sources) rendered
     * according to the given RenderContext.
     *
     * <p> This method supports chaining of either Renderable or
     * RenderedImage operations.  If the first source of sources in
     * the ParameterBlock used to construct the RenderableImageOp are
     * RenderableImages, then a three step process is followed:
     *
     * <ol>
     * <li> It calls mapRenderContext() on its associated CRIF for
     * each RenderableImage source in the ParameterBlock;
     * <li> It calls getImage() on each of its RenderableImage sources
     * using the backwards-mapped RenderContexts obtaing in step 1,
     * resulting in a rendering of each source;
     * <li> It calls the ContextualRenderedImageFactory.createOp()
     * method with a new ParameterBlock containing the parameters of
     * the RenderableImageOp and the RenderedImages that were created in the
     * getImage() calls.
     * </ol>
     *
     * <p> If the first element or elements in the source Vector of
     * the ParameterBlock used to construct the RenderableImageOp are
     * instances of RenderedImage, then the CRIF.createOp() method is
     * called immediately using the original ParameterBlock.
     * This provides a basis case for the recursion.
     *
     * @param renderContext The RenderContext to use to perform the rendering.
     * @return a RenderedImage containing the desired output image.
     */
    public RenderedImage getImage(RenderContext renderContext) {
        RenderedImage image = null;
        RenderContext rcOut = null;

        // Clone the original ParameterBlock; if the ParameterBlock
        // contains RenderableImage sources, they will be replaced by
        // RenderedImages.
        ParameterBlock renderedPB = (ParameterBlock)pb.clone();
        int i;
        try {
            // This assumes that if there is no renderable source, that there
            // is a rendered source in pb
            
            if (sources != null) {
                Vector renderedSources = new Vector();
                for (i = 0; i < sources.size(); i++) {
                    rcOut =
                        myCRIF.mapRenderContext(i, renderContext, pb, this);
                    RenderedImage rdrdImage =
                       ((RenderableImage)sources.elementAt(i)).getImage(rcOut);
                    if (rdrdImage == null) {
                        return null;
                    }
                    
                    // Add this rendered image to the ParameterBlock's
                    // list of RenderedImages.
                    renderedSources.addElement(rdrdImage);
                }
                
                if (renderedSources.size() > 0) {
                    renderedPB.setSources(renderedSources);
                }
            }

            return myCRIF.createOp(renderContext,renderedPB);
        } catch (ArrayIndexOutOfBoundsException e) {
            // This should never happen
            return null;
        }
    }
    
    /** 
     * Returns a vector of RenderableImages that are the sources of
     * image data for this RenderableImage. Note that this method will
     * often return an empty vector.
     *
     * @return a Vector of RenderableImages.
     */
    public Vector getSources() {
        return (Vector)sources.clone();
    }
  
    /** 
     * Gets a property from the property set of this image.  
     *
     * @param name a String containging the name of the desired property.
     * @return a reference to the property Object.
     */
    public Object getProperty(String name) {
        return myCRIF.getProperty(pb, name);
    }

    /**
     * Return a list of names recognized by getProperty(String).
     */
    public String[] getPropertyNames() {
        return null;
    }
  
    /** 
     * Gets the width in user coordinate space.  The default width
     * and height for image sources should be 1 unit high, by aspect
     * ratio wide.
     *
     * @return the width of the image in user coordinates.
     */
    public float getWidth() {
        if (boundingBox == null) {
            boundingBox = myCRIF.getBoundingBox(pb);
        }
        return (float)boundingBox.getWidth();
    }

    /** 
     * Gets the height in user coordinate space.  The default width
     * and height for image sources should be 1 unit high, by aspect
     * ratio wide.
     *
     * @return the height of the image in user coordinates.
     */
    public float getHeight() {
        if (boundingBox == null) {
            boundingBox = myCRIF.getBoundingBox(pb);
        }
        return (float)boundingBox.getHeight();
    }
  
    /** 
     * Gets the minimum X coordinate for image data.  Be aware that a
     * RenderableImage may have "infinite" extent, in which case this
     * function will return -Float.MAX_VALUE.
     *
     * @return The minimum X coordinate of the rendering-independent image.
     */
    public float getMinXCoord() {
        if (boundingBox == null) {
            boundingBox = myCRIF.getBoundingBox(pb);
        }
        return (float)boundingBox.getLeft();
    }

    /** 
     * Gets the maxumum X coordinate for image data.  Be aware that a
     * RenderableImage may have "infinite" extent, in which case this
     * function will return Float.MAX_VALUE.
     *
     * @return The maximum X coordinate of the rendering-independent image.
     */
    public float getMaxXCoord() {
        if (boundingBox == null) {
            boundingBox = myCRIF.getBoundingBox(pb);
        }
        return (float)boundingBox.getRight();
    }

    /** 
     * Gets the minimum Y coordinate for image data.  Be aware that a
     * RenderableImage may have "infinite" extent, in which case this
     * function will return -Float.MAX_VALUE.
     *
     * @return The minimum Y coordinate of the rendering-independent image.
     */
    public float getMinYCoord() {
        if (boundingBox == null) {
            boundingBox = myCRIF.getBoundingBox(pb);
        }
        return (float)boundingBox.getTop();
    }

    /** 
     * Gets the maximum Y coordinate for image data.  Be aware that a
     * RenderableImage may have "infinite" extent, in which case this
     * function will return Float.MAX_VALUE.
     *
     * @return The maximum Y coordinate of the rendering-independent image.
     */
    public float getMaxYCoord() {
        if (boundingBox == null) {
            boundingBox = myCRIF.getBoundingBox(pb);
        }
        return (float)boundingBox.getBottom();
    }
  
    /** 
     * Gets a RenderedImage instance of this image with width w, and
     * height h in pixels.  The RenderContext is built automatically
     * with an appropriate usr2dev transform and an area of interest
     * of the full image.  All the rendering hints come from hints
     * passed in.
     *
     * @param w the width of rendered image in pixels.
     * @param h the height of rendered image in pixels.
     * @param hints a Hashtable of Hints to use during rendering.
     * @return a RenderedImage containing the rendered data.
     */
    public RenderedImage getScaledRendering(int w, int h, Hashtable hints) {
        AffineTransform usr2dev = new AffineTransform();
        usr2dev.setToScale((double)w,(double)h);
        
        RenderContext newRC = new RenderContext(usr2dev,hints);
        return getImage(newRC);
    }

    /** 
     * Gets a RenderedImage instance of this image with a default
     * width and height in pixels.  The RenderContext is built
     * automatically with an appropriate usr2dev transform and an area
     * of interest of the full image.  All the rendering hints come
     * from hints passed in.  Implementors of this interface must be
     * sure that there is a defined default width and height.
     *
     * @return a RenderedImage containing the rendered data.
     */
    public RenderedImage getDefaultRendering() {
        AffineTransform usr2dev = new AffineTransform(); // Identity
        RenderContext newRC = new RenderContext(usr2dev);
        return getImage(newRC);
    }
    
    /**
     * Change the current ParameterBlock of the operation, allowing
     * editing of image rendering chains.  The effects of such a
     * change will be visible when a new rendering is created from
     * this RenderableImageOp or any dependent RenderableImageOp.
     *
     * @param pb the new ParameterBlock.
     * @return the old ParameterBlock.
     */
    public ParameterBlock setParameterBlock(ParameterBlock pb) {
        ParameterBlock oldPb = this.pb;

        this.pb = (ParameterBlock)pb.clone();
        sources = null;

        if (pb.getNumSources() > 0) {
            sources = new Vector();
            int i = 0;
            while (i < pb.getNumSources()) {
                Object o = pb.getSource(i);
                if (o instanceof RenderableImage) {
                    sources.add((RenderableImage)o);
                    i++;
                } else {
                    break;
                }
            }

        }

        return oldPb;
    }
}
