/*
 * @(#)BasicToggleButtonUI.java	1.35 98/02/02
 * 
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 */
 
package java.awt.swing.plaf.basic;

import java.awt.*;
import java.awt.event.*;

import java.awt.swing.*;
import java.awt.swing.border.*;
import java.awt.swing.plaf.*;

import java.io.Serializable;

/**
 * BasicToggleButton implementation
 * <p>
 * Warning: serialized objects of this class will not be compatible with
 * future swing releases.  The current serialization support is appropriate
 * for short term storage or RMI between Swing1.0 applications.  It will
 * not be possible to load serialized Swing1.0 objects with future releases
 * of Swing.  The JDK1.2 release of Swing will be the compatibility
 * baseline for the serialized form of Swing objects.
 *
 * @version 1.35 02/02/98
 * @author Jeff Dinkins
 */
public class BasicToggleButtonUI extends ToggleButtonUI implements Serializable {
    private static ToggleButtonUI toggleButtonUI;
    private static BasicButtonListener listener;

    // visual constants
    private static final int defaultTextIconGap = 4;
    private final static Insets defaultMargin = new Insets(2,2,2,2);

    protected static Color toggleButtonFocus;

    public static ComponentUI createUI(JComponent b) {
        if(toggleButtonUI == null) {
            toggleButtonUI = new BasicToggleButtonUI();
        }
        return toggleButtonUI;
    }
    
    /**
     * insets and margin
     */

    public Insets getDefaultMargin(AbstractButton b) {
	return defaultMargin;
    }

    public Border getDefaultBorder(AbstractButton b) {
	return UIManager.getBorder("ToggleButton.border");
    }

    public int getDefaultTextIconGap(AbstractButton b) {
	return defaultTextIconGap;
    }


    /**
     * ToggleButtonUI Implementation
     *
     */

    public void installUI(JComponent c) {
      installDefaults(c);
      installListeners(c);
      installKeyboardActions(c);
    }

    protected void installDefaults(JComponent c){
      LookAndFeel.installColorsAndFont(c, 
					    "ToggleButton.background", 
					    "ToggleButton.foreground", 
					    "ToggleButton.font");
      LookAndFeel.installBorder(c,"ToggleButton.border");
      toggleButtonFocus = UIManager.getColor("ToggleButton.focus");

    }

    protected void installListeners(JComponent c){
	if ((listener = createListener(c)) != null){
	    c.addMouseListener(listener);
	    c.addMouseMotionListener(listener);
	    c.addFocusListener(listener);
	    ((AbstractButton)c).addChangeListener(listener);
	}
    }

    protected void installKeyboardActions(JComponent c){
	listener.setupKeyboard((AbstractButton) c);
    }

    public void uninstallUI(JComponent c) {
	uninstallKeyboardActions(c);
	uninstallListeners(c);
	uninstallDefaults(c);
    }
    
    protected void uninstallKeyboardActions(JComponent c) {
	c.resetKeyboardActions();
    }
    
    protected void uninstallListeners(JComponent c) {
        c.removeMouseListener(listener);
        c.removeMouseMotionListener(listener);
	c.removeFocusListener(listener);
        ((AbstractButton)c).removeChangeListener(listener);
    }

    protected void uninstallDefaults(JComponent c){
	LookAndFeel.uninstallBorder(c);
    }


    protected BasicButtonListener createListener(JComponent c) {
	return new BasicButtonListener((AbstractButton) c);
    }


    public void paint(Graphics g, JComponent c) {
        AbstractButton b = (AbstractButton) c;
        ButtonModel model = b.getModel();

        Dimension size = b.getSize();
        FontMetrics fm = g.getFontMetrics();

	Insets i = c.getInsets();

        Rectangle viewRect = new Rectangle(size);

	viewRect.x += i.left;
	viewRect.y += i.top;
	viewRect.width -= (i.right + viewRect.x);
	viewRect.height -= (i.bottom + viewRect.y);

        Rectangle iconRect = new Rectangle();
        Rectangle textRect = new Rectangle();

	Font f = c.getFont();
	g.setFont(f);

	// layout the text and icon
        String text = SwingUtilities.layoutCompoundLabel(
	    fm, b.getText(), b.getIcon(),
	    b.getVerticalAlignment(), b.getHorizontalAlignment(),
	    b.getVerticalTextPosition(), b.getHorizontalTextPosition(),
	    viewRect, iconRect, textRect, b.getText() == null ? 0 : defaultTextIconGap
	);


        g.setColor(b.getBackground());

	if (model.isArmed() && model.isPressed() || model.isSelected())
	  {
	    paintButtonPressed(g,b);
	  }

	// Paint the Icon
        if(b.getIcon() != null) { 
	    paintIcon(g,c,iconRect);
        }

	// Draw the Text
        if(text != null && !text.equals("")) {
	    paintText(g, c, textRect, text);
        }
	  
	// draw the dashed focus line.
        if (b.isFocusPainted() && b.hasFocus()) {
	  paintFocus(g,size);
	}
    }

    protected void paintIcon(Graphics g, JComponent c, Rectangle iconRect) {
	AbstractButton b = (AbstractButton) c;
	ButtonModel model = b.getModel();
	Icon icon = null;
	
	if(!model.isEnabled()) {
	    icon = (Icon) b.getDisabledIcon();
	} else if(model.isPressed() && model.isArmed()) {
	    icon = (Icon) b.getPressedIcon();
	    if(icon == null) {
		// Use selected icon
		       icon = (Icon) b.getSelectedIcon();
	    } 
	} else if(model.isSelected()) {
	    icon = (Icon) b.getSelectedIcon();
	} else if(b.isRolloverEnabled() && model.isRollover()) {
	    icon = (Icon) b.getRolloverIcon();
	} 
	
	if(icon == null) {
	    icon = (Icon) b.getIcon();
	}
	
	icon.paintIcon(c, g, iconRect.x, iconRect.y);
    }

    protected void paintText(Graphics g, JComponent c, Rectangle textRect, String text){
	AbstractButton b = (AbstractButton)c;
	ButtonModel model = b.getModel();

        FontMetrics fm = g.getFontMetrics();

	if(model.isEnabled()) {
	    // *** paint the text normally
		       g.setColor(b.getForeground());
	    BasicGraphicsUtils.drawString(g,text, model.getMnemonic(),
					  textRect.x,
					  textRect.y + fm.getAscent());
	} else {
	    // *** paint the text disabled
		       g.setColor(b.getBackground().brighter());
	    BasicGraphicsUtils.drawString(g,text,model.getMnemonic(),
					  textRect.x, textRect.y + fm.getAscent());
	    g.setColor(b.getBackground().darker());
	    BasicGraphicsUtils.drawString(g,text,model.getMnemonic(),
					  textRect.x - 1, textRect.y + fm.getAscent() - 1);
	}
    }

    protected void paintFocus(Graphics g, Dimension size){
    }

    // Basic does nothing - L&F like Motif shade the button
    protected void paintButtonPressed(Graphics g, AbstractButton b){  }


    public Dimension getPreferredSize(JComponent c) {
	AbstractButton b = (AbstractButton)c;
	return BasicGraphicsUtils.getPreferredButtonSize(b, defaultTextIconGap);
    }


}
