/*
 * @(#)SortedVector.java	1.9 97/12/09
 * 
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 */
package java.awt.swing.text;

import java.util.Vector;

/**
 * Like a vector except that it does sorted inserts using a binary
 * search for the sort index.  This seems like it should be somewhere
 * like java.util.
 *
 * @author  Timothy Prinzing
 * @version 1.9 12/09/97
 */
abstract class SortedVector
{
    /**
     * Returns the number of elements in the vector.
     *
     * @return the number of elements
     */
    public int size() {
	return v.size();
    }

    /**
     * Gets the element at a specified index.
     *
     * @param index the index
     * @return the object at the given index
     */
    public Object elementAt(int index) {
	return v.elementAt(index);
    }

    /**
     * Inserts an object into the vector, preserving sorted order.
     *
     * @param o the object
     * @exception Error for duplicate insertions
     */
    public void insert(Object o) {
	int index = findSortIndex(o);
	if (index < v.size()) {
	    Object atIndex = v.elementAt(index);
	    if (compare(o, atIndex) == 0) {
		throw new Error("Attempt to insert a duplicate");
	    }
	}
	v.insertElementAt(o, index);
    }

    /**
     * Compares two objects.
     *
     * @param o1 the first object
     * @param o2 the second object
     * @return < 0 if o1 < o2, 0 if the same, > 0 if o1 > o2
     */
    protected abstract int compare(Object o1, Object o2);

    /**
     * Finds the index of where to insert a new object.
     *
     * @param o the object to insert
     * @return the index
     */
    protected int findSortIndex(Object o) {
	int lower = 0; 
	int upper = v.size() - 1;
	int mid = 0;
	
	if (upper == -1) {
	    return 0;
	}

	int cmp = 0;
	Object last = v.elementAt(upper);
	cmp = compare(o, last);
	if (cmp > 0)
	    return upper + 1;
	
	while (lower <= upper) {
	    mid = lower + ((upper - lower) / 2);
	    Object entry = v.elementAt(mid);
	    cmp = compare(o, entry);

	    if (cmp == 0) {
		// found a matching description
		return mid;
	    } else if (cmp < 0) {        
		upper = mid - 1;
	    } else {
		lower = mid + 1;
	    }
	}

	// didn't find it, but we indicate the index of where it would belong.
	return (cmp < 0) ? mid : mid + 1;
    }

    /**
     * Removes an element.
     *
     * @param elem the element
     */
    public void removeElement(Object elem) {
	v.removeElement(elem);
    }

    /**
     * Copies the element list into an external array.
     *
     * @param anArray the array
     */
    public void copyInto(Object anArray[]) {
	v.copyInto(anArray);
    }

    /**
     * Converts the list to a string.
     *
     * @return the string
     */
    public String toString() {
	return v.toString();
    }

    /**
     * The underlying vector.
     */
    protected Vector v = new Vector();
}
