/*
 * @(#)HTMLEditorKit.java	1.23 98/01/28
 * 
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 */
package java.awt.swing.text.html;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.awt.swing.Action;
import java.awt.swing.text.*;
import java.awt.swing.*;
import java.awt.swing.event.*;
import java.util.*;

/**
 * This is the default implementation of html editing
 * functionality.  The primary goal with this is to
 * be small, but flexible.  It is not intended to be
 * an all singing and all dancing html implementation.
 * This is provided to meet more modest needs, with
 * the idea that more substantial needs can be met 
 * with alternative implementations.
 *
 * @author  Timothy Prinzing
 * @author  Makarand Gokhale
 * @version 1.23 01/28/98
 */
public class HTMLEditorKit extends StyledEditorKit {
   
    /**
     * Constructs an HTMLEditorKit, creates a StyleContext,
     * and loads the style sheet.
     */
    public HTMLEditorKit() {
	styleContext = new StyleContext();
	loadStyleSheet(styleContext);
    }

    /**
     * Create a copy of the editor kit.  This
     * allows an implementation to serve as a prototype
     * for others, so that they can be quickly created.
     */
    public Object clone() {
	return new HTMLEditorKit();
    }

    /**
     * Get the MIME type of the data that this
     * kit represents support for.  This kit supports
     * the type <code>text/html</code>.
     */
    public String getContentType() {
	return "text/html";
    }

    /**
     * Fetch a factory that is suitable for producing 
     * views of any models that are produced by this
     * kit.  
     */
    public ViewFactory getViewFactory() {
	return new HTMLFactory();
    }

    /**
     * Create an uninitialized text storage model
     * that is appropriate for this type of editor.
     */
    public Document createDefaultDocument() {
	StyledDocument doc = new HTMLDocument(styleContext);

	return doc;
    }

    /**
     * Create and initialize a model from the given
     * stream which is expected to be in a format appropriate
     * for this kind of editor.  This is implemented to read
     * html 3.2 text.
     * 
     * @param in  The stream to read from
     * @param doc The destination for the insertion.
     * @param pos The location in the document to place the
     *   content.
     * @exception IOException on any I/O error
     * @exception BadLocationException if pos represents an invalid
     *   location within the document.
     */
    public void read(Reader in, Document doc, int pos) throws IOException, BadLocationException {

	if (doc instanceof HTMLDocument) {
	    HTMLDocument hdoc = (HTMLDocument) doc;
	    hdoc.read(in, pos);
	} else {
	    super.read(in, doc, pos);
	}
    }

    /**
     * Write content from a document to the given stream
     * in a format appropriate for this kind of content handler.
     * 
     * @param out  The stream to write to
     * @param doc The source for the write.
     * @param pos The location in the document to fetch the
     *   content.
     * @param len The amount to write out.
     * @exception IOException on any I/O error
     * @exception BadLocationException if pos represents an invalid
     *   location within the document.
     */
    public void write(Writer out, Document doc, int pos, int len) 
	throws IOException, BadLocationException {

	if (doc instanceof StyledDocument) {
	    try {
		HTMLWriter w = new HTMLWriter();
		w.write(out, (StyledDocument)doc);
	    } catch (HTMLException h) {
		throw new IOException("HTMLException: " + h.getMessage());
	    }
	} else {
	    super.write(out, doc, pos, len);
	}
    }

    /**
     * Called when the kit is being installed into the
     * a JEditorPane. 
     *
     * @param c the JEditorPane
     */
    public void install(JEditorPane c) {
	c.addMouseListener(linkHandler);
	super.install(c);
    }

    /**
     * Called when the kit is being removed from the
     * JEditorPane.  This is used to unregister any 
     * listeners that were attached.
     *
     * @param c the JEditorPane
     */
    public void deinstall(JEditorPane c) {
	c.removeMouseListener(linkHandler);
	super.deinstall(c);
    }

    /**
     * Default Cascading Style Sheet file that sets
     * up the tag views.
     */
    public static final String DEFAULT_CSS = "default.css";

    /**
     * Load in the default.css file.
     */
    private void loadStyleSheet(StyleContext sc) {
	InputStream is = this.getClass().getResourceAsStream(DEFAULT_CSS);
	if (is == null) 
  	  System.out.println("HTMLEditorKit.loadStyleSheet: " + DEFAULT_CSS + " file not found");
	else {
	  //
	  // Create the StyleReader and call it.
	  //
	  StyleReader stylereader = new StyleReader(sc);
	  stylereader.read(sc, 0, is);
	}

    }

    /**
     * Fetches the command list for the editor.  This is
     * the list of commands supported by the superclass
     * augmented by the collection of commands defined
     * locally for style operations.
     *
     * @return the command list
     */
    public Action[] getActions() {
	return TextAction.augmentList(super.getActions(), this.defaultActions);
    }

    public static class LinkController extends MouseAdapter {

	/**
         * Called for a mouse click event.
	 * If the component is read-only (ie a browser) then 
	 * the clicked event is used to drive an attempt to
	 * follow the reference specified by a link.
	 *
	 * @param e the mouse event
	 * @see MouseListener#mouseClicked
	 */
        public void mouseClicked(MouseEvent e) {
	    JEditorPane editor = (JEditorPane) e.getSource();
	    if (! editor.isEditable()) {
		Point pt = new Point(e.getX(), e.getY());
		int pos = editor.viewToModel(pt);
		if (pos >= 0) {
		    activateLink(pos, editor);
		}
	    }
	}

	/**
	 * Calls linkActivated on the associated JEditorPane
	 * if the given position represents a link.
         *
         * @param pos the position
	 */
        protected final void activateLink(int pos, JEditorPane html) {
	    Document doc = html.getDocument();
	    if (doc instanceof StyledDocument) {
		StyledDocument sdoc = (StyledDocument) doc;
		Element e = sdoc.getCharacterElement(pos);
		AttributeSet a = e.getAttributes();
		String href = (String) a.getAttribute("href");
		if (href != null) {
		    URL u;
			try {
			    u = new URL(html.getPage(), href);
			} catch (MalformedURLException m) {
			    u = null;
			}
			HyperlinkEvent linkEvent = 
			    new HyperlinkEvent(html, HyperlinkEvent.EventType.ACTIVATED, u);
			html.fireHyperlinkUpdate(linkEvent);
		}
	    }
	}
    }

    private StyleContext styleContext = null;
    private MouseListener linkHandler = new LinkController();

/** The bold action identifier
*/
    public static final String	BOLD_ACTION = "html-bold-action";
/** The italic action identifier
*/
    public static final String	ITALIC_ACTION = "html-italic-action";
/** The paragraph left indent action identifier
*/
    public static final String	PARA_INDENT_LEFT = "html-para-indent-left";
/** The paragraph right indent action identifier
*/
    public static final String	PARA_INDENT_RIGHT = "html-para-indent-right";
/** The  font size increase to next value action identifier
*/
    public static final String	FONT_CHANGE_BIGGER = "html-font-bigger";
/** The font size decrease to next value action identifier
*/
    public static final String	FONT_CHANGE_SMALLER = "html-font-smaller";
/** The Color choice action identifier
     The color is passed as an argument
*/
    public static final String	COLOR_ACTION = "html-color-action";
/** The logical style choice action identifier
     The logical style is passed in as an argument
*/
    public static final String	LOGICAL_STYLE_ACTION = "html-logical-style-action";
    public static final String	IMG_ALIGN_TOP = "html-image-align-top";
    public static final String	IMG_ALIGN_MIDDLE = "html-image-align-middle";
    public static final String	IMG_ALIGN_BOTTOM = "html-image-align-bottom";
    public static final String	IMG_BORDER = "html-image-border";

    // --- Action implementations ---------------------------------

    private static final Action[] defaultActions = {
	new HTMLBoldAction(),
	new HTMLItalicAction(),
	new FontSizeChangeAction(FONT_CHANGE_BIGGER, true),
	new FontSizeChangeAction(FONT_CHANGE_SMALLER, false),
	new ParagraphIndentAction(PARA_INDENT_LEFT, true),
	new ParagraphIndentAction(PARA_INDENT_RIGHT, false),
	new ImgAlignAction(IMG_ALIGN_TOP, ImageView.TOP),
	new ImgAlignAction(IMG_ALIGN_MIDDLE, ImageView.MIDDLE),
	new ImgAlignAction(IMG_ALIGN_BOTTOM, ImageView.BOTTOM),
	new ImgBorderAction(IMG_BORDER),
	new ForegroundAction(COLOR_ACTION, Color.black),
	new LogicalStyleAction("Normal")
    };

    static abstract class HtmlAction extends StyledTextAction {

	HtmlAction(String nm) {
	    super(nm);
	}

	/**
	 * Applies the given attributes to character 
	 * content.   The attributes are applied to a range. The range
	 * must be within current selection
	 * If the the start and end point are the same then 
	 * the attributes are applied to
	 * the input attribute set which defines the attributes
	 * for any new text that gets inserted.
	 *
	 * @param attr the attributes
	 * @param start The starting position for a range
	 * @param end   the ending position for a range
	 * @param replace if true, then replace the existing attributes first
	 */
        protected final void setCharacterAttributes(JEditorPane editor, 
					      AttributeSet attr, int start,
					      int end, boolean replace) {
	    if(start > end) {
		int tmp = end;
		end = start;
		start = tmp;
	    }
	    int p0 = editor.getSelectionStart();
	    int p1 = editor.getSelectionEnd();
	    if(start < p0) start = p0;
	    if(end > p1 ) end = p1;
	    if (start != end) {
		StyledDocument doc = getStyledDocument(editor);
		doc.setCharacterAttributes(start, end - start, attr, replace);
	    } else {
		StyledEditorKit k = getStyledEditorKit(editor);
		MutableAttributeSet inputAttributes = k.getInputAttributes();
		if (replace) {
		    inputAttributes.removeAttributes(inputAttributes);
		}
		inputAttributes.addAttributes(attr);
	    }
        }
    }


    /**
     * An action to toggle the bold attribute
     */
    static class HTMLBoldAction extends HtmlAction {

	public HTMLBoldAction() {
	    super(BOLD_ACTION);
	}

 	 public void actionPerformed(ActionEvent ae) {
	    JTextComponent target = getFocusedComponent();
  	    if (target != null) {
		JEditorPane pane = (JEditorPane) target;
		MutableAttributeSet attr = new SimpleAttributeSet();
// Check if the event source is a toggle button
// If so set or unset bold based on toggle state
		Object o = ae.getSource();
		boolean bSet=true;
		if (o != null && o instanceof JToggleButton)
		    bSet = ((JToggleButton)o).isSelected();
		StyleConstants.setBold(attr, bSet);
	    	setCharacterAttributes(pane, attr, false);
 	    }
	}
    }

    /**
     * An action to toggle the italic attribute
     */
    static class HTMLItalicAction extends HtmlAction {

	public HTMLItalicAction() {
	    super(ITALIC_ACTION);
	}

 	 public void actionPerformed(ActionEvent ae) {
	    JTextComponent target = getFocusedComponent();
  	    if (target != null) {
		JEditorPane pane = (JEditorPane) target;
		MutableAttributeSet attr = new SimpleAttributeSet();
// Check if the event source is a toggle button
// If so set or unset italic based on toggle state
		Object o = ae.getSource();
		boolean bSet=true;
		if (o != null && o instanceof JToggleButton)
		    bSet = ((JToggleButton)o).isSelected();
		StyleConstants.setItalic(attr, bSet);
	    	setCharacterAttributes(pane,attr, false);
 	    }
	}
    }


/**
 *
 * This class supports the font size change action. When the button
 * is clicked on a toolbar it sends an action event to this class
 * This class defines a local variable to specify, whether to increas
 * or decrease the font size. It obtains the next size using HTMLStyleSheet
 *
 * The font size is changed only if there is only one style in effect
 * for a selection. Otherwise the change request is ignored.
 *
 * The following is the order of processing
 * 1.	Check if there is a range of selection.
 * 2. If not set the fontsize for the input
 & 3. If ma range is selected then get all the effective styles
 * 4. If more than one size exists on a Font Ignore change request
 * 5. Otherwise change font
 */


    static class FontSizeChangeAction extends HtmlAction {
	private boolean bIncrease = true;

        FontSizeChangeAction( String key, boolean b) {
	    super(key);
	    bIncrease = b;
        }
      
	public void actionPerformed(ActionEvent event) {
	    JTextComponent target = getFocusedComponent();
	    StyleSheet ss = StyleReader.getStyleSheet();
  	    if (target != null) {
		JEditorPane pane = (JEditorPane) target;

		Caret   ip = pane.getCaret();
		int pos = ip.getDot();
		if (ip.getMark()  ==  pos) {
/* ASK TIM HOW TO GET INPUTATTRIBUTES
		    AttributeSet as = pane.getInputAttributes();
		    */
		    AttributeSet as = null;
		    if (as != null) {
		      	int size = StyleConstants.getFontSize(as);
		      	size= (bIncrease?ss.getBigger(size):ss.getSmaller(size));
			MutableAttributeSet attr = new SimpleAttributeSet();
			StyleConstants.setFontSize(attr, size);
			setCharacterAttributes(pane,attr, false);
		    }
		}
		else {
		    int start = ip.getMark();
		    int end = ip.getDot();
		    Vector v = getCharacterElements(pane, start,end);
		    if (v != null ) {
			int size=-1;
			int nextSize=-1;
			int vSize=v.size();
			if (vSize > 0 ) {
			    boolean bAbort = false;
			    int SizeArray[] = new int[vSize];
			    for (int i=0; i < vSize && !bAbort; i++) {
				try {
				    AttributeSet as = ((Element)v.elementAt(i)).getAttributes();
				    size= StyleConstants.getFontSize(as);
				    nextSize = (bIncrease?ss.getBigger(size):ss.getSmaller(size));
				    if (size == nextSize) {
					System.out.println("Bigger/Smaller ABORTING:Size limit reached for element:"+((Element)v.elementAt(i)).getName());
					bAbort = true;
					continue;
				    }
				    SizeArray[i] = nextSize;
				} catch(NullPointerException ee1) {
				    System.out.println("Exception: Cannot find font size for element:"+i);
				    SizeArray[i] = -1;
				    bAbort = true;
				    continue;
				}
			    }
			    if (!bAbort) {
				int eStart, eEnd;
				for (int i=0; i <vSize;i++) {
				    Element elem = (Element)v.elementAt(i);
				    eStart = elem.getStartOffset();
				    eEnd = elem.getEndOffset();
				    if (eStart < start)  eStart = start;
				    if (eEnd > end)  eEnd = end;
				    if (SizeArray[i] > 0) {
					MutableAttributeSet attr = new SimpleAttributeSet();
					StyleConstants.setFontSize(attr, SizeArray[i]);
					setCharacterAttributes(pane,attr,eStart, eEnd, false);
				    }
				}
			    }
			    else{
				System.out.println("Atleast one size at the limit, Aborting bigger/smaller");
			    }
	    		}
	  	    }
		}
	    }
	}
	
    }
	 
  /**
     * ParagraphIndentAction sets a new alignment.  If it is instantiated
     * with bInc=true then the next alignment will be chosen in order
     * from left to right.  If bInc=false then the next alignment will be
     * in the order from right to left.
     */
    static class ParagraphIndentAction extends  HtmlAction {
       	boolean	bLeft = false;

      	ParagraphIndentAction(String key, boolean b) {
	    super(key);
	    bLeft = b;
        }
      
        public void actionPerformed(ActionEvent e) {
	    JEditorPane pane = (JEditorPane)getFocusedComponent();
	    if (pane != null) {

		StyledEditorKit k = getStyledEditorKit(pane);
		MutableAttributeSet attr =  k.getInputAttributes();
		if (attr != null) {
    		    int align = StyleConstants.getAlignment(attr);
	    	    if (!bLeft) {
	      		if (align == StyleConstants.ALIGN_LEFT) {
	        	    align = StyleConstants.ALIGN_CENTER;
			}
	      	   	else if (align == StyleConstants.ALIGN_CENTER) {
	        	    align = StyleConstants.ALIGN_RIGHT;
          		}
		    }
	 	    else {
	     	 	if (align == StyleConstants.ALIGN_CENTER) {
	     	   	    align = StyleConstants.ALIGN_LEFT;
    			}
	     	 	else if (align == StyleConstants.ALIGN_RIGHT) {
	       		    align = StyleConstants.ALIGN_CENTER;
        		}
	    	    }
		    StyleConstants.setAlignment(attr, align);
		    setParagraphAttributes(pane,attr, false);
	  	}
	    }
   	}
    }

    /**
     * Set the Logical Style on the paragraph.
     * It also supports setLogicalStyle which allows setting a new logical style
     */
    static class LogicalStyleAction extends HtmlAction {
	String  styleStr;
      	LogicalStyleAction(String styleStr) {
	    super(LOGICAL_STYLE_ACTION);
	    this.styleStr = styleStr;
      	}
      	public void actionPerformed(ActionEvent e) {

	    JEditorPane pane = (JEditorPane)getFocusedComponent();
	    if (pane != null) {
		StyledDocument doc = (StyledDocument) pane.getDocument();
	    	Style style = doc.getStyle(styleStr);
	    	if (style != null) {
	    	    doc.setLogicalStyle(pane.getCaretPosition(),style);
	    	}
	    }
      	}
	public void setLogicalStyle(Object src ,String styleStr) {
	    this.styleStr = styleStr;
	    actionPerformed(new ActionEvent(src, 0, "html-logical-style"));
        }
      
    }

 /**
 * A method to get all the elements in the range
 */
    static Vector getCharacterElements(JEditorPane editorpane,int start, int end) {
	StyledDocument doc = (StyledDocument)editorpane.getDocument();
  	Vector v = new Vector();
  	int pos = start;
  	start= (start> end?end:start);
  	end = (start > end?pos:end);
  	pos = start;
  	while (pos <= end) {
  	    Element e = doc.getCharacterElement(pos);
            if ( e== null) {
  	    	pos++;
  	    	continue;
  	    }
  	    v.addElement(e);
  	    if (pos < e.getEndOffset()+1)
  	        pos = e.getEndOffset()+1;
  	    else
  	        pos++;
    	}
	return v;
    }


    /**
     * An action to align images top, middle, bottom
     */
    static class ImgAlignAction extends HtmlAction {

	String align;

	/** Create an action to set an image (vertical) alignment.
	    @param  name  The name of this action
	    @param  align  The value of the ALIGN attribute. Correct values
	    	are "top", "middle", "bottom" (see consts in ImageView) */
	public ImgAlignAction(String name, String align) {
	    super(name);
	    this.align = align;
	}

 	public void actionPerformed(ActionEvent ae) {
 	    if(DEBUG)System.out.println("ImgAlignAction: set align to "+align);
	    JTextComponent target = getFocusedComponent();
  	    if (target != null) {
		JEditorPane pane = (JEditorPane) target;
		MutableAttributeSet attr = new SimpleAttributeSet();
		attr.addAttribute(HTMLDefs.ALIGN,align);
	    	setCharacterAttributes(pane, attr, false);

	    	//((HTMLDocument)pane.getDocument()).dump(System.out);//$$$$$ TEST
 	    }
 	}
    }

    /**
     * Toggle the border on an img.
     */
    static class ImgBorderAction extends HtmlAction {

	public ImgBorderAction(String name) {
	    super(name);
	}

 	public void actionPerformed(ActionEvent ae) {
	    JEditorPane pane = (JEditorPane) getFocusedComponent();
	    AttributeSet attr = getImageAttributes(pane);
	    if( attr != null ) {
	    	// Figure out if it currently has a border:
	    	boolean link = attr.isDefined(HTMLDefs.HREF);
	    	int border = link ?2 :0;  // border is default in link
	        String borderStr = (String) attr.getAttribute(HTMLDefs.BORDER);
	        if( borderStr != null ) {
 	    	    try{
 	                border = Integer.parseInt(borderStr);
 	    	    }catch( NumberFormatException x ) {
 	    	    }
 	    	}
 	    	
 	    	// Work out the new 'border' attribute value:
		MutableAttributeSet newAttr = new SimpleAttributeSet();
		boolean replace = false;
 	    	if( border == 0  && !link ) {
 	    	    if(DEBUG)System.out.println("ImgBorderAction: set border=2");
 	    	    newAttr.addAttribute(HTMLDefs.BORDER,"2");
 	    	} else if( border != 0 && link ) {
 	    	    if(DEBUG)System.out.println("ImgBorderAction: set border=0");
 	    	    newAttr.addAttribute(HTMLDefs.BORDER,"0");
		} else {
		    // need to remove 'border' attribute entirely:
 	    	    if(DEBUG)System.out.println("ImgBorderAction: reset border");
		    newAttr.addAttributes(attr);
		    newAttr.removeAttribute(HTMLDefs.BORDER);
		    replace = true;
 	    	}
 	    	
 	    	// Update/replace attributes:
 	    	setCharacterAttributes(pane, newAttr, replace);
	    }
 	}
 	
 	/** If an image is selected, return its AttributeSet. */
 	private AttributeSet getImageAttributes( JEditorPane editorpane ) {
  	    if (editorpane != null) {
        	int start = editorpane.getSelectionStart();
        	int end = editorpane.getSelectionEnd();
        	if( Math.abs(end-start)==1 ) {
	            StyledDocument htmldoc = (StyledDocument) editorpane.getDocument();
	            Element e = htmldoc.getCharacterElement(Math.min(start,end));
	            if( e != null && e.getName().equals(HTMLDefs.IMG) )
	    	        return e.getAttributes();
	    	}
	    }
	    return null;
 	}
    }
    
    static final boolean DEBUG = false;
}
