/*
 * @(#)Short.java	1.11 98/03/18
 *
 * Copyright 1996, 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.lang;

/**
 * The Short class is the standard wrapper for short values.
 *
 * @author  Nakul Saraiya
 * @version 1.11, 03/18/98
 * @see     java.lang.Number
 * @since   JDK1.1
 */
public final
class Short extends Number implements Comparable {

    /**
     * The minimum value a Short can have.
     */
    public static final short   MIN_VALUE = -32768;

    /**
     * The maximum value a Short can have.
     */
    public static final short   MAX_VALUE = 32767;

    /**
     * The Class object representing the primitive type short.
     */
    public static final Class	TYPE = Class.getPrimitiveClass("short");

    /**
     * Returns a new String object representing the specified
     * Short. The radix is assumed to be 10.
     *
     * @param s the short to be converted
     */
    public static String toString(short s) {
	return Integer.toString((int)s, 10);
    }

    /**
     * Assuming the specified String represents a short, returns
     * that short's value. Throws an exception if the String cannot
     * be parsed as a short.  The radix is assumed to be 10.
     *
     * @param s		the String containing the short
     * @exception	NumberFormatException If the string does not
     *			contain a parsable short.
     */
    public static short parseShort(String s) throws NumberFormatException {
	return parseShort(s, 10);
    }

    /**
     * Assuming the specified String represents a short, returns
     * that short's value. Throws an exception if the String cannot
     * be parsed as a short.
     *
     * @param s		the String containing the short
     * @param radix	the radix to be used
     * @exception	NumberFormatException If the String does not
     *			contain a parsable short.
     */
    public static short parseShort(String s, int radix)
	throws NumberFormatException {
	int i = Integer.parseInt(s, radix);
	if (i < MIN_VALUE || i > MAX_VALUE)
	    throw new NumberFormatException();
	return (short)i;
    }

    /**
     * Assuming the specified String represents a short, returns a
     * new Short object initialized to that value.  Throws an
     * exception if the String cannot be parsed as a short.
     *
     * @param s		the String containing the integer
     * @param radix 	the radix to be used
     * @exception	NumberFormatException If the String does not
     *			contain a parsable short.
     */
    public static Short valueOf(String s, int radix)
	throws NumberFormatException {
	return new Short(parseShort(s, radix));
    }

    /**
     * Assuming the specified String represents a short, returns a
     * new Short object initialized to that value.  Throws an
     * exception if the String cannot be parsed as a short.
     *
     * @param s		the String containing the integer
     * @exception	NumberFormatException If the String does not
     *			contain a parsable short.
     */
    public static Short valueOf(String s) throws NumberFormatException {
	return valueOf(s, 10);
    }

    /**
     * Decodes a String into a Short.  The String may represent
     * decimal, hexadecimal, and octal numbers.
     *
     * @param nm the string to decode
     */
    public static Short decode(String nm) throws NumberFormatException {
	if (nm.startsWith("0x")) {
	    return Short.valueOf(nm.substring(2), 16);
	}
	if (nm.startsWith("#")) {
	    return Short.valueOf(nm.substring(1), 16);
	}
	if (nm.startsWith("0") && nm.length() > 1) {
	    return Short.valueOf(nm.substring(1), 8);
	}

	return Short.valueOf(nm);
    }

    /**
     * The value of the Short.
     */
    private short value;

    /**
     * Constructs a Short object initialized to the specified short value.
     *
     * @param value	the initial value of the Short
     */
    public Short(short value) {
	this.value = value;
    }

    /**
     * Constructs a Short object initialized to the value specified by the
     * String parameter.  The radix is assumed to be 10.
     *
     * @param s		the String to be converted to a Short
     * @exception	NumberFormatException If the String does not
     *			contain a parsable short.
     */
    public Short(String s) throws NumberFormatException {
	this.value = parseShort(s);
    }

    /**
     * Returns the value of this Short as a byte.
     */
    public byte byteValue() {
	return (byte)value;
    }

    /**
     * Returns the value of this Short as a short.
     */
    public short shortValue() {
	return value;
    }

    /**
     * Returns the value of this Short as an int.
     */
    public int intValue() {
	return (int)value;
    }

    /**
     * Returns the value of this Short as a long.
     */
    public long longValue() {
	return (long)value;
    }

    /**
     * Returns the value of this Short as a float.
     */
    public float floatValue() {
	return (float)value;
    }

    /**
     * Returns the value of this Short as a double.
     */
    public double doubleValue() {
	return (double)value;
    }

    /**
     * Returns a String object representing this Short's value.
     */
    public String toString() {
	return String.valueOf((int)value);
    }

    /**
     * Returns a hashcode for this Short.
     */
    public int hashCode() {
	return (int)value;
    }

    /**
     * Compares this object to the specified object.
     *
     * @param obj	the object to compare with
     * @return 		true if the objects are the same; false otherwise.
     */
    public boolean equals(Object obj) {
	if ((obj != null) && (obj instanceof Short)) {
	    return value == ((Short)obj).shortValue();
	}
	return false;
    }

    /**
     * Compares two Shorts numerically.
     *
     * @param   anotherShort   the <code>Short</code> to be compared.
     * @return  the value <code>0</code> if the argument Short is equal to
     *          this Short; a value less than <code>0</code> if this Short
     *          is numerically less than the Short argument; and a
     *          value greater than <code>0</code> if this Short is
     *          numerically greater than the Short argument
     *		(signed comparison).
     * @since   JDK1.2
     */
    public int compareTo(Short anotherShort) {
	return this.value - anotherShort.value;
    }

    /**
     * Compares this Short to another Object.  If the Object is a Short,
     * this function behaves like <code>compareTo(Short)</code>.  Otherwise,
     * it throws a <code>ClassCastException</code> (as Shorts are comparable
     * only to other Shorts).
     *
     * @param   o the <code>Object</code> to be compared.
     * @return  the value <code>0</code> if the argument is a Short
     *		numerically equal to this Short; a value less than
     *		<code>0</code> if the argument is a Short numerically
     *		greater than this Short; and a value greater than
     *		<code>0</code> if the argument is a Short numerically
     *		less than this Short.
     * @exception <code>ClassCastException</code> if the argument is not a
     *		  <code>Short</code>. 
     * @see     java.lang.Comparable
     * @since   JDK1.2
     */
    public int compareTo(Object o) {
	return compareTo((Short)o);
    }

    /** use serialVersionUID from JDK 1.1. for interoperability */
    private static final long serialVersionUID = 7515723908773894738L;
}
