/*
 * @(#)SocketType.java	1.4 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.rmi.server;

import java.io.*;
import java.net.*;

/**
 * The SocketType class is a type that can be passed to an
 * RMISocketFactory object on creation of a socket or a server socket
 * to choose desired socket type.
 */
public final class SocketType {

    private String protocol;

    private byte[] refData;

    private Object serverData;

    /**
     * Create new SocketType object with given properties.
     * @param protocol string identifying socket protocol
     * @param refData optional array of bytes with additional protocol data
     * @param serverData server-specific protocol data
     */
    public SocketType(String protocol, byte[] refData, Object serverData)
    {
	this.protocol   = protocol;
	this.refData    = refData;
	this.serverData = serverData;
    }

    /**
     * Return the protocol string for this descriptor.
     */
    public String getProtocol() {
	return protocol;
    }

    /**
     * Return the optional client data for this descriptor.
     */
    public byte[] getRefData() {
	return refData;
    }

    /**
     * Return the server-specific data for this descriptor.
     */
    public Object getServerData() {
	return serverData;
    }

    /**
     * Write descriptor to output stream.
     * @param out the output object to write to
     */
    public void write(DataOutput out)
	throws java.io.IOException
    {
	out.writeUTF(protocol);
	if (refData == null) {
	    out.writeInt(-1);
	} else {
	    out.writeInt(refData.length);
	    out.write(refData);
	}
    }

    /**
     * Read a descriptor from input stream.
     * @param in the input object to read from
     */
    public static SocketType read(DataInput in)
	throws java.io.IOException
    {
	String protocol = in.readUTF();
	byte refData[];
	int len = in.readInt();
	if (len < 0) {
	    refData = null;
	} else {
	    refData = new byte[len];
	    in.readFully(refData);
	}

	return new SocketType(protocol, refData, null);
    }

    public int hashCode() {
	return protocol.hashCode();
    }

    public boolean equals(Object obj) {
	if (obj != null && obj instanceof SocketType) {
	    SocketType desc = (SocketType) obj;
	    if (!protocol.equals(desc.protocol))
		return false;
	    if (refData != null ^ desc.refData != null)
		return false;
	    if (refData != null) {
		if (refData.length != desc.refData.length)
		    return false;
		for (int i = 0; i < refData.length; i++)
		    if (refData[i] != desc.refData[i])
			return false;
	    }
	    return true;
	} else {
	    return false;
	}
    }

    public String toString() {
	return "[" + protocol + "]";
    }
}
