/*
 * @(#)ORB.java	1.71 98/03/18
 *
 * Copyright 1995-1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package org.omg.CORBA;

import org.omg.CORBA.portable.*;
import org.omg.CORBA.ORBPackage.InvalidName;
import java.util.Properties;
import java.applet.Applet;

/**
 * Class ORB represents an implementation of a CORBA Object Request
 * Broker. The ORB class encapsulates generic CORBA functionality,
 * including:
 * <ul>
 * <li> initializing the ORB implementation by supplying values for
 *      predefined properties and environmental parameters.
 * <li> obtaining initial object references to the NameService, etc
 *      using the resolve_initial_references() method.
 * <li> converting object references to strings and back.
 * <li> connecting and disconnecting a servant (an instance of a CORBA Object
        implementation) to the ORB
 * <li> creating pseudo-objects such as TypeCode, Any, NamedValue, Context, 
 *      Environment and lists containing these pseudo-objects.
 * <li> sending multiple messages in the Dynamic Invocation Interface.
 * </ul>
 *
 * The ORB class can be used to obtain references to objects
 * implemented anywhere on the network. Specific ORB implementations
 * can extend this class to add new features.<P>
 *
 * <TABLE BORDER>
 * <TR><TH>Property Name</TH>   <TH>Property Value</TH></TR>
 * <CAPTION>Standard Java CORBA Properties:</CAPTION>
 *     <TR><TD>org.omg.CORBA.ORBClass</TD>
 *     <TD>class name of an ORB implementation</TD></TR>
 *     <TR><TD>org.omg.CORBA.ORBSingletonClass</TD>
 *     <TD>class name of the ORB returned by init()</TD></TR>
 * </TABLE>
 *
 * When creating an ORB instance, these values are located using
 * a standard search order:<P>
 *
 * <OL>
 *     <LI>check in Applet parameter or application string array, if any</LI>
 *
 *     <LI>check in properties parameter, if any</LI>
 *
 *     <LI>check in the System properties (currently applications only)</LI>
 *
 *     <LI>fallback on a hardcoded default behavior</LI>
 * </OL>
 *     
 * @version 1.71, 03/18/98
 * @since   JDK1.2
 */
abstract public class ORB {

    //
    // This is the ORB implementation used when nothing else is specified.
    // This string is customized by whoever provides this class to
    // point at their ORB implementation.
    //
    private static final String ORBClassKey = "org.omg.CORBA.ORBClass";
    private static final String ORBSingletonClassKey = "org.omg.CORBA.ORBSingletonClass";

    //
    // The last resort fallback ORB implementation classes in case
    // no ORB implementation class is dynamically configured through
    // properties or applet parameters. Change these values to
    // vendor-specific class names.
    //
    private static final String defaultORB = "com.sun.CORBA.iiop.ORB";
    private static final String defaultORBSingleton = "com.sun.CORBA.iiop.ORBSingleton";

    //
    // The global instance of the singleton ORB implementation which
    // acts as a factory for typecodes for generated Helper classes.
    // TypeCodes should be immutable since they may be shared across
    // different security contexts (applets). There should be no way to
    // use an TypeCode as a storage depot for illicitly passing
    // information or Java objects between different security contexts.
    //
    static private ORB singleton;

    private static Class thisClass = ORB.class;

    static {
        String className;

	className = getSystemProperty(ORBSingletonClassKey);
        if (className == null)
          className = defaultORBSingleton;
	singleton = create_impl(className, null);
	singleton.set_parameters((String [])null, (Properties)null);
    }


    //
    // A hack. If we are loaded as a system class (no class loader)
    // then try to access the property in privileged mode.
    // Otherwise don't bother because we will probably just get
    // a SecurityException anyway and this will show up in
    // browser Java consoles when we are downloaded and it will
    // confuse or scare people.
    //
    private static String getSystemProperty(String name) {
        String className = null;

        if (thisClass.getClassLoader() == null) {
            try {
                try {
                    // java.security.AccessController.beginPrivileged();
                    className = System.getProperty(name);
                } finally {
                    // java.security.AccessController.endPrivileged();
                }
            } catch (Throwable ex) {
                //
                // Eat any exceptions related to being downloaded into
                // a pre-JDK1.2 browser that doesn't have AccessController.
                //
                if (ex instanceof ThreadDeath)
                    throw (ThreadDeath)ex;
            }
        }

	return className;
    }

    /**
     * Return the ORB singleton object. This method always returns the
     * same ORB instance which is an instance of the class described by the
     * org.omg.CORBA.ORBSingletonClass system property.
     * When called in an applet environment, the returned ORB
     * can only be used as a factory for TypeCodes. The resulting
     * TypeCodes can be safely shared between untrusted applets.
     * @return the singleton ORB
     */
    public static ORB init() {
	return singleton;
    }
  
    private static ORB create_impl(String className, ClassLoader cl) {
        try {
            return (ORB)Class.forName(className).newInstance();
	} catch (ClassNotFoundException ex) {
	    // Eat the exception and try again below...
        } catch (Exception ex) {
	    throw new INITIALIZE(
              "can't instantiate default ORB implementation " + className);
        }

        try {
            return (ORB)cl.loadClass(className).newInstance();
        } catch (Exception ex) {
	    throw new INITIALIZE(
              "can't instantiate default ORB implementation " + className);
        }
    }

    /**
     * Create a new ORB instance for an application.
     * @param args application strings from main() method
     * @param props application-specific properties
     * @return the newly created ORB instance
     */
    public static ORB init(String[] args, Properties props) {
	//
	// Note that there is no standard command-line argument for
	// specifying the default ORB implementation. For an
	// application you can choose an implementation either by
	// setting the CLASSPATH to pick a different org.omg.CORBA
	// and it's baked-in ORB implementation default or by
	// setting an entry in the properties object or in the
	// system properties.
	//
        String className = null;
	ORB orb;

	if (props != null)
	    className = props.getProperty(ORBClassKey);
	if (className == null)
	    className = getSystemProperty(ORBClassKey);
        if (className == null)
          className = defaultORB;
	orb = create_impl(className, null);
	orb.set_parameters(args, props);
	return orb;
    }


    /**
     * Create a new ORB instance for an applet.
     * @param app the applet
     * @param props applet-specific properties
     * @return the newly created ORB instance
     */
    public static ORB init(Applet app, Properties props) {
        String className;
	ORB orb;
	ClassLoader cl = null;

        className = app.getParameter(ORBClassKey);
	if (className == null && props != null)
	    className = props.getProperty(ORBClassKey);
	// if (className == null)
	//     className = getSystemProperty(ORBClassKey);
        if (className == null)
          className = defaultORB;
	if (app != null)
	    cl = app.getClass().getClassLoader();
        orb = create_impl(className, cl);
        orb.set_parameters(app, props);
	return orb;
    }

    /** set_parameters must be implemented by subclass ORB implementations.
     *  It allows the ORB implementation to be initialized with the given
     *  parameters and properties. This method is used in applications only.
     * @param args application strings from main() method
     * @param props application-specific properties
     */
    abstract protected void set_parameters(String[] args, Properties props);
  
    /** set_parameters must be implemented by subclass ORB implementations.
     *  It allows the ORB implementation to be initialized with the given
     *  parameters and properties. This method is used in applets only.
     * @param app the applet
     * @param props applet-specific properties
     */
    abstract protected void set_parameters(Applet app, Properties props);

    /**
     * Connect a servant (a Java object which is an instance of the server
     * implementation class) to the ORB. The servant class must 
     * extend the "ImplBase" class corresponding to the interface which is
     * supported by the server. The servant must thus be a CORBA object 
     * reference, and inherit from org.omg.CORBA.Object.
     * Servants created by the user can start receiving remote invocations 
     * after connect() has been called. A servant may also be
     * automatically and implicitly connected to the ORB if it is passed as
     * an IDL parameter in an IDL method invocation on a non-local object,
     * i.e. the servant object has to be marshaled and sent outside of the
     * process address space. Calling connect() when an object is already
     * connected has no effect.
     * 
     * @param obj The servant object reference.
     */
    abstract public void connect(org.omg.CORBA.Object obj);

    /**
     * Disconnect a servant from the ORB. After this method returns, 
     * incoming remote requests for that servant will be rejected by the ORB
     * by sending the org.omg.CORBA.OBJECT_NOT_EXIST exception back to the
     * remote client. Thus the object appears to be destroyed from the 
     * point of view of remote clients. Note however, that local requests issued
     * using the servant (i.e the Java object reference) directly do not 
     * pass through the ORB, hence will continue to be processed by the
     * servant. disconnect() has no effect if the servant is not connected.
     */
    abstract public void disconnect(org.omg.CORBA.Object obj);

    //
    // ORB method implementations.
    //
    // We are trying to accomplish 2 things at once in this class.
    // It can act as a default ORB implementation front-end
    // creating an actual ORB implementation object which is a
    // subclass of this ORB class and then delegating the method
    // implementations. 
    //
    // To accomplish the delegation model, the 'delegate' private instance
    // variable is set if an instance of this class is created directly.
    //
  
    /**
     * Return a list of the initially available CORBA object references,
     * such as the NameService and InterfaceRepository. 
     * @return the list of object service names
     */
    abstract public String[] list_initial_services();
  
    /**
     * Resolve a specific object reference from the set of available
     * initial service names.
     * @param name the name of the initial service
     * @return An object reference for the named object.
     * @exception InvalidName The given name is not associated with a
     *                         known service.
     */
    abstract public org.omg.CORBA.Object resolve_initial_references(String name) 
      throws InvalidName;
  
    /**
     * Convert a CORBA object reference to a string.
     * Note that the format of this string is predefined by IIOP, allowing 
     * strings generated by a different ORB to be converted into an object
     * reference. 
     * @param obj the object reference to stringify.
     * @return the string representing the object reference.
     */
    abstract public String object_to_string(org.omg.CORBA.Object obj);
  
    /**
     * Convert a string to a CORBA object reference.
     * 
     * @param str the string
     * @return the object reference.
     * @exception SystemException One of a fixed set of CORBA system exceptions.
     */
    abstract public org.omg.CORBA.Object string_to_object(String str);

    /**
     * Create an NVList with a given number of elements.
     *
     * @param count	size of list to create
     * @result		NVList created
     *
     * @see NVList
     */
    abstract public NVList create_list(int count);

    /**
     * Create an NVList for the arguments of the IDL method corresponding 
     * to an OperationDef (obtained from
     * an Interface Repository).
     *
     * @param oper	operation def to use to create list
     * @result		NVList created
     *
     * @see NVList
     */
    abstract public NVList create_operation_list(OperationDef oper);

    /**
     * Create a NamedValue using the given name, value and flags.
     *
     * @param s The name of the NamedValue object
     * @param any The Any value to be inserted in the NamedValue
     * @param flags The flags (e.g. argument mode flags) for the NamedValue
     * @result		NamedValue created
     */
    abstract public NamedValue create_named_value(String s, Any any, int flags);

    /**
     * Create an ExceptionList
     *
     * @result		ExceptionList created
     */
    abstract public ExceptionList create_exception_list();

    /**
     * Create a ContextList
     *
     * @result		ContextList created
     */
    abstract public ContextList create_context_list();

    /**
     * Get the default Context object
     *
     * @result		the default Context object
     */
    abstract public Context get_default_context();

    /**
     * Create an Environment
     *
     * @result		Environment created
     */
    abstract public Environment create_environment();

    /**
     * Create a new org.omg.CORBA.portable.OutputStream into which 
     * IDL method parameters can be marshalled during method invocation.
     * @result		the new org.omg.CORBA.portable.OutputStream
     */
    abstract public org.omg.CORBA.portable.OutputStream create_output_stream();

    /**
     * Send multiple dynamic (DII) requests asynchronously without expecting
     * any responses. Note that oneway invocations are not guaranteed to
     * reach the server.
     *
     * @param req		an array of request objects.
     */
    abstract public void send_multiple_requests_oneway(Request[] req);

    /**
     * Send multiple dynamic (DII) requests asynchronously.
     *
     * @param req		an array of request objects.
     */
    abstract public void send_multiple_requests_deferred(Request[] req);

    /**
     * Find out if any of the deferred (asynchronous) invocations have 
     * a response yet. 
     * @result a boolean value which is true if there is a response available.
     */
    abstract public boolean poll_next_response();

    /**
     * Get the next request instance for which a response has been received.
     *
     * @result		the next request instance ready with a response.
     * @exception WrongTransaction This exception may be raised if 
     * get_next_response is called from a different transaction's scope
     * than the one from which the original request was sent. See the
     * OMG Transaction Service specification for details.
     */
    abstract public Request get_next_response() throws WrongTransaction;

    /**
     * Get the TypeCode for a primitive IDL type.
     *
     * @param tcKind	the integer kind for the primitive type
     * @return		the requested TypeCode
     */
    abstract public TypeCode get_primitive_tc(TCKind tcKind);

    /**
     * Create a TypeCode for an IDL struct.
     *
     * @param id	the logical (repository) id for the struct.
     * @param name	the name of the struct.
     * @param members	an array describing the members of the struct.
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_struct_tc(String id, String name,
					  StructMember[] members);
  
    /**
     * Create a TypeCode for an IDL union.
     *
     * @param id	the logical id of the union.
     * @param name	the name of the union.
     * @param discriminator_type
     *			the type of the union discriminator.
     * @param members	an array describing the members of the union.
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_union_tc(String id, String name,
					 TypeCode discriminator_type,
					 UnionMember[] members);
  
    /**
     * Create a TypeCode for an IDL enum.
     *
     * @param id	the logical id for the enum.
     * @param name	the name for the enum.
     * @param members	an array describing the members of the enum.
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_enum_tc(String id, String name, String[] members);
  
    /**
     * Create a TypeCode for an alias (IDL typedef).
     *
     * @param id	the logical id for the alias.
     * @param name	the name for the alias.
     * @param original_type
     * 			the TypeCode for the original type for which this 
		        is an alias.
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_alias_tc(String id, String name,
					 TypeCode original_type);
  
    /**
     * Create a TypeCode for an IDL exception.
     *
     * @param id	the logical id for the exception.
     * @param name	the name for the exception.
     * @param members	an array describing the members of the exception.
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_exception_tc(String id, String name,
					     StructMember[] members);
  
    /**
     * Create a TypeCode for an IDL interface.
     *
     * @param id	the logical id for the interface.
     * @param name	the name for the interface.
     * @return		the requested TypeCode.
     */

    abstract public TypeCode create_interface_tc(String id, String name);
  
    /**
     * Create a TypeCode for a string.
     *
     * @param bound	the bound for the string.
     * @return		the requested TypeCode.
     */

    abstract public TypeCode create_string_tc(int bound);
  
    /**
     * Create a TypeCode for a wide string.
     *
     * @param bound	the bound for the string.
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_wstring_tc(int bound);
  
    /**
     * Create a TypeCode for a sequence.
     *
     * @param bound	the bound for the sequence.
     * @param element_type
     *			the type of elements of the sequence.
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_sequence_tc(int bound, TypeCode element_type);
  
    /**
     * Create a recursive TypeCode in a sequence.
     *
     * @param bound	the bound for the sequence.
     * @param offset	the index to the enclosing TypeCode that is
     *			being referenced.  
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_recursive_sequence_tc(int bound, int offset);
  
    /**
     * Create a TypeCode for an array.
     *
     * @param length	the length of the array.
     * @param element_type
     *			the type of elements of the array.
     * @return		the requested TypeCode.
     */
    abstract public TypeCode create_array_tc(int length, TypeCode element_type);

    /**
     * Create an instance of Any. The returned Any object is initialized to
     * contain a typecode of type TCKind.tc_null.
     *
     * @return		the requested Any.
     */
    abstract public Any create_any();

    /**
     * Get a Current object. 
     * The Current interface is used to manage thread-specific
     * information for use by the transactions, security and other
     * services.
     *
     * @return		a Current object.
     */
    abstract public Current get_current();
}
  
